@@ -15,27 +15,45 @@
 #include "src/buildtool/build_engine/target_map/built_in_rules.hpp"
 
 #include <algorithm>
+#include <compare>
 #include <cstddef>
 #include <filesystem>
 #include <functional>
+#include <initializer_list>
 #include <iterator>
 #include <memory>
+#include <optional>
+#include <set>
 #include <sstream>
 #include <string>
+#include <type_traits>
 #include <unordered_map>
 #include <unordered_set>
 #include <utility>  // std::move
+#include <vector>
 
 #include "fmt/core.h"
+#include "src/buildtool/build_engine/analysed_target/analysed_target.hpp"
+#include "src/buildtool/build_engine/analysed_target/target_graph_information.hpp"
+#include "src/buildtool/build_engine/base_maps/entity_name.hpp"
+#include "src/buildtool/build_engine/base_maps/entity_name_data.hpp"
 #include "src/buildtool/build_engine/base_maps/field_reader.hpp"
+#include "src/buildtool/build_engine/expression/configuration.hpp"
 #include "src/buildtool/build_engine/expression/evaluator.hpp"
 #include "src/buildtool/build_engine/expression/expression.hpp"
 #include "src/buildtool/build_engine/expression/expression_ptr.hpp"
+#include "src/buildtool/build_engine/expression/function_map.hpp"
+#include "src/buildtool/build_engine/expression/linked_map.hpp"
+#include "src/buildtool/build_engine/expression/target_result.hpp"
 #include "src/buildtool/build_engine/target_map/export.hpp"
 #include "src/buildtool/build_engine/target_map/utils.hpp"
+#include "src/buildtool/common/action_description.hpp"
 #include "src/buildtool/common/artifact_description.hpp"
 #include "src/buildtool/common/artifact_digest_factory.hpp"
 #include "src/buildtool/common/repository_config.hpp"
+#include "src/buildtool/common/tree.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
+#include "src/buildtool/storage/storage.hpp"
 #include "src/utils/cpp/path.hpp"
 #include "src/utils/cpp/vector.hpp"
 
@@ -63,7 +81,6 @@ auto const kBlobGenRuleFields =
                                     "tainted",
                                     "type"};
 auto const kTreeRuleFields = std::unordered_set<std::string>{"arguments_config",
-                                                             "data",
                                                              "deps",
                                                              "name",
                                                              "tainted",
@@ -1351,9 +1368,18 @@ void GenericRuleWithDeps(
     for (auto const& dep : dependency_values) {
         inputs = ExpressionPtr{Expression::map_t{inputs, (*dep)->RunFiles()}};
     }
+    auto inputs_conflict = BuildMaps::Target::Utils::tree_conflict(inputs);
     for (auto const& dep : dependency_values) {
         inputs = ExpressionPtr{Expression::map_t{inputs, (*dep)->Artifacts()}};
     }
+    // While syntactical conflicts are resolved in a latest wins (with artifacts
+    // after runfiles), semantic path conclicts are an error.
+    if (inputs_conflict) {
+        (*logger)(fmt::format("Input artifacts have staging conflict on {}",
+                              nlohmann::json(*inputs_conflict).dump()),
+                  /*fatal=*/true);
+        return;
+    }
     std::vector<Tree::Ptr> trees{};
     inputs = BuildMaps::Target::Utils::add_dir_for(
         cwd_value->String(), inputs, &trees);
