@@ -16,12 +16,19 @@
 #define INCLUDED_SRC_BUILDTOOL_BUILD_ENGINE_TARGET_MAP_RESULT_MAP_HPP
 
 #include <algorithm>
+#include <compare>
 #include <cstddef>
 #include <fstream>
+#include <functional>
+#include <iomanip>
+#include <iterator>
+#include <memory>
 #include <mutex>
+#include <numeric>
 #include <optional>
 #include <string>
 #include <thread>
+#include <unordered_map>
 #include <unordered_set>
 #include <utility>  // std::move
 #include <vector>
@@ -29,19 +36,19 @@
 #include "gsl/gsl"
 #include "nlohmann/json.hpp"
 #include "src/buildtool/build_engine/analysed_target/analysed_target.hpp"
-#include "src/buildtool/build_engine/base_maps/entity_name.hpp"
-#include "src/buildtool/build_engine/expression/expression.hpp"
+#include "src/buildtool/build_engine/analysed_target/target_graph_information.hpp"
+#include "src/buildtool/build_engine/base_maps/entity_name_data.hpp"
+#include "src/buildtool/build_engine/expression/configuration.hpp"
 #include "src/buildtool/build_engine/target_map/configured_target.hpp"
+#include "src/buildtool/common/action_description.hpp"
+#include "src/buildtool/common/identifier.hpp"
 #include "src/buildtool/common/statistics.hpp"
 #include "src/buildtool/common/tree.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
-#include "src/buildtool/multithreading/task.hpp"
 #include "src/buildtool/multithreading/task_system.hpp"
 #include "src/buildtool/progress_reporting/progress.hpp"
-#include "src/buildtool/storage/target_cache.hpp"
 #include "src/buildtool/storage/target_cache_key.hpp"
-#include "src/utils/cpp/hash_combine.hpp"
 
 namespace BuildMaps::Target {
 
@@ -221,8 +228,8 @@ class ResultTargetMap {
                 std::sort(origin_map[i.first].begin(),
                           origin_map[i.first].end(),
                           [](auto const& left, auto const& right) {
-                              auto left_target = left.first.ToString();
-                              auto right_target = right.first.ToString();
+                              auto const& left_target = left.first;
+                              auto const& right_target = right.first;
                               return (left_target < right_target) or
                                      (left_target == right_target and
                                       left.second < right.second);
@@ -302,6 +309,31 @@ class ResultTargetMap {
                             }
                         });
         result.actions.erase(lastaction, result.actions.end());
+        auto& output_map = progress->OutputMap();
+        output_map.clear();
+        output_map.reserve(na);
+        if constexpr (kIncludeOrigins) {
+            for (auto const& action : result.actions) {
+                if (not action.desc->OutputFiles().empty()) {
+                    output_map[action.desc->Id()] =
+                        action.desc->OutputFiles()[0];
+                }
+                else if (not action.desc->OutputDirs().empty()) {
+                    output_map[action.desc->Id()] =
+                        action.desc->OutputDirs()[0];
+                }
+            }
+        }
+        else {
+            for (auto const& action : result.actions) {
+                if (not action->OutputFiles().empty()) {
+                    output_map[action->Id()] = action->OutputFiles()[0];
+                }
+                else if (not action->OutputDirs().empty()) {
+                    output_map[action->Id()] = action->OutputDirs()[0];
+                }
+            }
+        }
 
         int trees_traversed = stats->TreesAnalysedCounter();
         if (trees_traversed > 0) {
