@@ -15,11 +15,13 @@
 #ifndef INCLUDED_SRC_BUILDTOOL_COMMON_CLI_HPP
 #define INCLUDED_SRC_BUILDTOOL_COMMON_CLI_HPP
 
+#include <algorithm>
 #include <chrono>
 #include <cstddef>
 #include <cstdint>
-#include <cstdlib>
+#include <exception>
 #include <filesystem>
+#include <functional>
 #include <optional>
 #include <string>
 #include <thread>
@@ -32,9 +34,10 @@
 #include "nlohmann/json.hpp"
 #include "src/buildtool/build_engine/expression/evaluator.hpp"
 #include "src/buildtool/common/clidefaults.hpp"
-#include "src/buildtool/common/retry_cli.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
+#include "src/buildtool/file_system/symlinks/resolve_special.hpp"
 #include "src/buildtool/logging/log_level.hpp"
+#include "src/buildtool/logging/logger.hpp"
 #include "src/buildtool/main/build_utils.hpp"
 #include "src/utils/cpp/path.hpp"
 
@@ -112,6 +115,7 @@ struct BuildArguments {
     std::size_t build_jobs{};
     std::optional<std::string> dump_artifacts{std::nullopt};
     std::optional<std::string> print_to_stdout{std::nullopt};
+    bool print_unique{false};
     bool show_runfiles{false};
 };
 
@@ -182,6 +186,8 @@ struct ServeArguments {
     std::optional<std::string> remote_serve_address;
     // repositories populated from just-serve config file
     std::vector<std::filesystem::path> repositories;
+    // remote execution endpoint as seen by the client
+    std::optional<std::string> client_remote_address;
 };
 
 struct GcArguments {
@@ -191,6 +197,7 @@ struct GcArguments {
 struct ToAddArguments {
     std::filesystem::path location;
     bool follow_symlinks{};
+    std::optional<ResolveSpecial> resolve_special{std::nullopt};
 };
 
 struct ProtocolArguments final {
@@ -512,6 +519,12 @@ static inline auto SetupBuildArguments(
            clargs->build_jobs,
            "Number of jobs to run during build phase (Default: same as jobs).")
         ->type_name("NUM");
+}
+
+static inline auto SetupExtendedBuildArguments(
+    gsl::not_null<CLI::App*> const& app,
+    gsl::not_null<BuildArguments*> const& clargs) {
+
     app->add_option("--dump-artifacts",
                     clargs->dump_artifacts,
                     "Dump artifacts to file (use - for stdout).")
@@ -525,6 +538,10 @@ static inline auto SetupBuildArguments(
                     clargs->print_to_stdout,
                     "After building, print the specified artifact to stdout.")
         ->type_name("LOGICAL_PATH");
+
+    app->add_flag("-p,--print-unique-artifact",
+                  clargs->print_unique,
+                  "Print the unique artifact, if any, to stdout.");
 }
 
 static inline auto SetupTCArguments(gsl::not_null<CLI::App*> const& app,
@@ -666,6 +683,22 @@ static inline auto SetupToAddArguments(
                   clargs->follow_symlinks,
                   "Resolve the positional argument to not be a symbolic link "
                   "before adding it to CAS.");
+    app->add_option_function<std::string>(
+        "--resolve-special",
+        [clargs](auto const& raw_value) {
+            if (kResolveSpecialMap.contains(raw_value)) {
+                clargs->resolve_special = kResolveSpecialMap.at(raw_value);
+            }
+            else {
+                Logger::Log(LogLevel::Warning,
+                            "Ignoring unknown --resolve-special strategy {}.",
+                            nlohmann::json(raw_value).dump());
+            }
+        },
+        "Optional strategy for handling special entries (e.g., symlinks) when "
+        "the content to add is a directory. Default behaviour if missing is to "
+        "only allow non-upward symlinks. Currently supported values: ignore, "
+        "tree-upwards, tree-all, all.");
 }
 
 static inline auto SetupGraphArguments(
