@@ -15,17 +15,21 @@
 #include "src/buildtool/execution_api/bazel_msg/bazel_msg_factory.hpp"
 
 #include <algorithm>
+#include <compare>
+#include <cstddef>
 #include <exception>
 #include <filesystem>
 #include <functional>
 #include <memory>
 #include <optional>
-#include <sstream>
 #include <string>
+#include <unordered_map>
 #include <utility>  // std::move
 #include <vector>
 
 #include "fmt/core.h"
+#include "google/protobuf/duration.pb.h"
+#include "google/protobuf/repeated_ptr_field.h"
 #include "src/buildtool/common/artifact_digest_factory.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
 #include "src/buildtool/file_system/git_repo.hpp"
@@ -120,15 +124,26 @@ template <class T>
     std::vector<std::string> const& symlink_names,
     std::vector<ArtifactDigest> const& symlink_digests,
     BazelMsgFactory::LinkDigestResolveFunc const& resolve_links)
-    -> std::vector<bazel_re::SymlinkNode> {
+    -> std::optional<std::vector<bazel_re::SymlinkNode>> {
+    if (symlink_names.size() != symlink_digests.size()) {
+        return std::nullopt;
+    }
+
     std::vector<std::string> symlink_targets;
+    symlink_targets.reserve(symlink_digests.size());
     resolve_links(symlink_digests, &symlink_targets);
-    auto it_name = symlink_names.begin();
-    auto it_target = symlink_targets.begin();
+
+    // Fail if the number of resolved symlinks does not match the requested
+    // number.
+    if (symlink_targets.size() != symlink_names.size()) {
+        return std::nullopt;
+    }
+
     std::vector<bazel_re::SymlinkNode> symlink_nodes;
-    // both loops have same length
-    for (; it_name != symlink_names.end(); ++it_name, ++it_target) {
-        symlink_nodes.emplace_back(CreateSymlinkNode(*it_name, *it_target));
+    symlink_nodes.reserve(symlink_targets.size());
+    for (std::size_t i = 0; i < symlink_targets.size(); ++i) {
+        symlink_nodes.emplace_back(
+            CreateSymlinkNode(symlink_names[i], symlink_targets[i]));
     }
     return symlink_nodes;
 }
@@ -149,13 +164,14 @@ struct DirectoryNodeBundle final {
 
     // SHA256 is used since bazel types are processed here.
     HashFunction const hash_function{HashFunction::Type::PlainSHA256};
-    auto digest = ArtifactDigestFactory::HashDataAs<ObjectType::File>(
-        hash_function, *content);
-
-    return DirectoryNodeBundle{
-        .message = CreateDirectoryNode(dir_name, digest),
-        .blob = ArtifactBlob{
-            std::move(digest), std::move(*content), /*is_exec=*/false}};
+    auto blob = ArtifactBlob::FromMemory(
+        hash_function, ObjectType::File, *std::move(content));
+    if (not blob.has_value()) {
+        return std::nullopt;
+    }
+    auto const digest = blob->GetDigest();
+    return DirectoryNodeBundle{.message = CreateDirectoryNode(dir_name, digest),
+                               .blob = *std::move(blob)};
 }
 
 /// \brief Create bundle for protobuf message Command from args strings.
@@ -186,11 +202,12 @@ struct DirectoryNodeBundle final {
     if (not content) {
         return std::nullopt;
     }
-    auto digest = ArtifactDigestFactory::HashDataAs<ObjectType::File>(
-        request.hash_function, *content);
-    return ArtifactBlob{std::move(digest),
-                        std::move(*content),
-                        /*is_exec=*/false};
+    auto blob = ArtifactBlob::FromMemory(
+        request.hash_function, ObjectType::File, *std::move(content));
+    if (not blob.has_value()) {
+        return std::nullopt;
+    }
+    return *std::move(blob);
 }
 
 /// \brief Create bundle for protobuf message Action from Command.
@@ -225,11 +242,12 @@ struct DirectoryNodeBundle final {
     if (not content) {
         return std::nullopt;
     }
-    auto digest = ArtifactDigestFactory::HashDataAs<ObjectType::File>(
-        request.hash_function, *content);
-    return ArtifactBlob{std::move(digest),
-                        std::move(*content),
-                        /*is_exec=*/false};
+    auto blob = ArtifactBlob::FromMemory(
+        request.hash_function, ObjectType::File, *std::move(content));
+    if (not blob.has_value()) {
+        return std::nullopt;
+    }
+    return *std::move(blob);
 }
 
 /// \brief Convert `DirectoryTree` to `DirectoryNodeBundle`.
@@ -280,13 +298,15 @@ struct DirectoryNodeBundle final {
                 }
             }
         }
+
+        auto symlink_nodes = CreateSymlinkNodesFromDigests(
+            symlink_names, symlink_digests, resolve_links);
+        if (not symlink_nodes.has_value()) {
+            return std::nullopt;
+        }
         return CreateDirectoryNodeBundle(
             root_name,
-            CreateDirectory(
-                file_nodes,
-                dir_nodes,
-                CreateSymlinkNodesFromDigests(
-                    symlink_names, symlink_digests, resolve_links)));
+            CreateDirectory(file_nodes, dir_nodes, *std::move(symlink_nodes)));
     } catch (...) {
         return std::nullopt;
     }
@@ -330,7 +350,7 @@ auto BazelMsgFactory::CreateDirectoryDigestFromTree(
         return std::nullopt;
     }
 
-    auto digest = bundle->blob.digest;
+    auto digest = bundle->blob.GetDigest();
     try {
         if (not process_blob(std::move(bundle->blob))) {
             return std::nullopt;
@@ -518,6 +538,223 @@ auto BazelMsgFactory::CreateDirectoryDigestFromGitTree(
     }
 }
 
+// NOLINTNEXTLINE(misc-no-recursion)
+auto BazelMsgFactory::CreateGitTreeDigestFromDirectory(
+    ArtifactDigest const& digest,
+    PathReadFunc const& read_path,
+    FileStoreFunc const& store_file,
+    TreeStoreFunc const& store_tree,
+    SymlinkStoreFunc const& store_symlink,
+    RehashedDigestReadFunc const& read_rehashed,
+    RehashedDigestStoreFunc const& store_rehashed) noexcept
+    -> expected<ArtifactDigest, std::string> {
+    GitRepo::tree_entries_t entries{};
+
+    try {
+        // read directory object
+        auto const tree_path = read_path(digest, ObjectType::Tree);
+        if (not tree_path) {
+            return unexpected{
+                fmt::format("failed reading CAS entry {}", digest.hash())};
+        }
+        auto const tree_content = FileSystemManager::ReadFile(*tree_path);
+        if (not tree_content) {
+            return unexpected{fmt::format("failed reading content of tree {}",
+                                          digest.hash())};
+        }
+        auto dir = BazelMsgFactory::MessageFromString<bazel_re::Directory>(
+            *tree_content);
+
+        // process subdirectories
+        for (auto const& subdir : dir->directories()) {
+            // get digest
+            auto const subdir_digest = ArtifactDigestFactory::FromBazel(
+                HashFunction::Type::PlainSHA256, subdir.digest());
+            if (not subdir_digest) {
+                return unexpected{subdir_digest.error()};
+            }
+            // get any cached digest mapping, to avoid unnecessary work
+            auto const cached_obj = read_rehashed(*subdir_digest);
+            if (not cached_obj) {
+                return unexpected{cached_obj.error()};
+            }
+            if (cached_obj.value()) {
+                // no work to be done, just add to map
+                if (auto raw_id =
+                        FromHexString(cached_obj.value()->digest.hash())) {
+                    entries[std::move(*raw_id)].emplace_back(subdir.name(),
+                                                             ObjectType::Tree);
+                }
+                else {
+                    return unexpected{fmt::format(
+                        "failed to get raw id for cached dir digest {}",
+                        cached_obj.value()->digest.hash())};
+                }
+            }
+            else {
+                // recursively get the subdirectory digest
+                auto const tree_digest =
+                    CreateGitTreeDigestFromDirectory(*subdir_digest,
+                                                     read_path,
+                                                     store_file,
+                                                     store_tree,
+                                                     store_symlink,
+                                                     read_rehashed,
+                                                     store_rehashed);
+                if (not tree_digest) {
+                    return unexpected{tree_digest.error()};
+                }
+                if (auto raw_id = FromHexString(tree_digest->hash())) {
+                    entries[std::move(*raw_id)].emplace_back(subdir.name(),
+                                                             ObjectType::Tree);
+                }
+                else {
+                    return unexpected{
+                        fmt::format("failed to get raw id for tree digest {}",
+                                    tree_digest->hash())};
+                }
+                // no need to cache the digest mapping, as this was done in the
+                // recursive call
+            }
+        }
+
+        // process symlinks
+        for (auto const& sym : dir->symlinks()) {
+            // get digest
+            auto const sym_digest =
+                ArtifactDigestFactory::HashDataAs<ObjectType::File>(
+                    HashFunction{HashFunction::Type::PlainSHA256},
+                    sym.target());
+            // get any cached digest mapping, to avoid unnecessary work
+            auto const cached_obj = read_rehashed(sym_digest);
+            if (not cached_obj) {
+                return unexpected{cached_obj.error()};
+            }
+            if (cached_obj.value()) {
+                // no work to be done, just add to map
+                if (auto raw_id =
+                        FromHexString(cached_obj.value()->digest.hash())) {
+                    entries[std::move(*raw_id)].emplace_back(
+                        sym.name(), ObjectType::Symlink);
+                }
+                else {
+                    return unexpected{fmt::format(
+                        "failed to get raw id for cached symlink digest {}",
+                        cached_obj.value()->digest.hash())};
+                }
+            }
+            else {
+                // check validity of symlink
+                if (not PathIsNonUpwards(sym.target())) {
+                    return unexpected{fmt::format(
+                        "found non-upwards symlink {}", sym_digest.hash())};
+                }
+                // rehash symlink
+                auto const blob_digest = store_symlink(sym.target());
+                if (not blob_digest) {
+                    return unexpected{
+                        fmt::format("failed rehashing as blob symlink {}",
+                                    sym_digest.hash())};
+                }
+                if (auto raw_id = FromHexString(blob_digest->hash())) {
+                    entries[std::move(*raw_id)].emplace_back(
+                        sym.name(), ObjectType::Symlink);
+                }
+                else {
+                    return unexpected{fmt::format(
+                        "failed to get raw id for symlink blob digest {}",
+                        blob_digest->hash())};
+                }
+                // while useless for future symlinks, cache digest mapping for
+                // file-type blobs with same content
+                if (auto error_msg = store_rehashed(
+                        sym_digest, *blob_digest, ObjectType::Symlink)) {
+                    return unexpected{*std::move(error_msg)};
+                }
+            }
+        }
+
+        // process files
+        for (auto const& file : dir->files()) {
+            // get digest
+            auto const file_digest = ArtifactDigestFactory::FromBazel(
+                HashFunction::Type::PlainSHA256, file.digest());
+            if (not file_digest) {
+                return unexpected{file_digest.error()};
+            }
+            auto const file_type = file.is_executable() ? ObjectType::Executable
+                                                        : ObjectType::File;
+            // get any cached digest mapping, to avoid unnecessary work
+            auto const cached_obj = read_rehashed(*file_digest);
+            if (not cached_obj) {
+                return unexpected{cached_obj.error()};
+            }
+            if (cached_obj.value()) {
+                // no work to be done, just add to map
+                if (auto raw_id =
+                        FromHexString(cached_obj.value()->digest.hash())) {
+                    entries[std::move(*raw_id)].emplace_back(file.name(),
+                                                             file_type);
+                }
+                else {
+                    return unexpected{fmt::format(
+                        "failed to get raw id for cached file digest {}",
+                        cached_obj.value()->digest.hash())};
+                }
+            }
+            else {
+                // rehash file
+                auto const file_path = read_path(*file_digest, file_type);
+                if (not file_path) {
+                    return unexpected{fmt::format("failed reading CAS entry {}",
+                                                  file_digest->hash())};
+                }
+                auto const blob_digest =
+                    store_file(*file_path, file.is_executable());
+                if (not blob_digest) {
+                    return unexpected{
+                        fmt::format("failed rehashing as blob file {}",
+                                    file_digest->hash())};
+                }
+                if (auto raw_id = FromHexString(blob_digest->hash())) {
+                    entries[std::move(*raw_id)].emplace_back(file.name(),
+                                                             file_type);
+                }
+                else {
+                    return unexpected{fmt::format(
+                        "failed to get raw id for file blob digest {}",
+                        blob_digest->hash())};
+                }
+                // cache digest mapping
+                if (auto error_msg =
+                        store_rehashed(*file_digest, *blob_digest, file_type)) {
+                    return unexpected{*std::move(error_msg)};
+                }
+            }
+        }
+
+        // create and store Git tree
+        auto const git_tree = GitRepo::CreateShallowTree(entries);
+        if (not git_tree) {
+            return unexpected{
+                fmt::format("failed creating Git tree for bazel Directory {}",
+                            digest.hash())};
+        }
+        auto const tree_digest = store_tree(git_tree->second);
+        // cache digest mapping
+        if (auto error_msg =
+                store_rehashed(digest, *tree_digest, ObjectType::Tree)) {
+            return unexpected{*std::move(error_msg)};
+        }
+        // return digest
+        return *tree_digest;
+    } catch (std::exception const& ex) {
+        return unexpected{fmt::format(
+            "creating Git tree digest unexpectedly failed with:\n{}",
+            ex.what())};
+    }
+}
+
 auto BazelMsgFactory::CreateDirectoryDigestFromLocalTree(
     std::filesystem::path const& root,
     FileStoreFunc const& store_file,
@@ -687,20 +924,15 @@ auto BazelMsgFactory::CreateActionDigestFromCommandLine(
         return std::nullopt;
     }
 
-    auto action = CreateActionBundle(cmd->digest, request);
+    auto action = CreateActionBundle(cmd->GetDigest(), request);
     if (not action) {
         return std::nullopt;
     }
 
+    auto result = action->GetDigest();
     if (request.store_blob) {
-        std::invoke(*request.store_blob,
-                    BazelBlob{ArtifactDigestFactory::ToBazel(cmd->digest),
-                              cmd->data,
-                              cmd->is_exec});
-        std::invoke(*request.store_blob,
-                    BazelBlob{ArtifactDigestFactory::ToBazel(action->digest),
-                              action->data,
-                              action->is_exec});
+        std::invoke(*request.store_blob, *std::move(cmd));
+        std::invoke(*request.store_blob, *std::move(action));
     }
-    return action->digest;
+    return result;
 }
