@@ -12,11 +12,14 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+#ifndef INCLUDED_SRC_BUILDTOOL_EXECUTION_API_COMMON_MESSAGE_LIMITS_HPP
+#define INCLUDED_SRC_BUILDTOOL_EXECUTION_API_COMMON_MESSAGE_LIMITS_HPP
+
 #include <cstddef>
 
-#include <grpc/grpc.h>
+struct MessageLimits final {
+    // Maximum length of a gprc message.
+    static constexpr std::size_t kMaxGrpcLength = 3UL * 1024 * 1024;
+};
 
-// Max size for batch transfers
-static constexpr std::size_t kMaxBatchTransferSize = 3UL * 1024 * 1024;
-static_assert(kMaxBatchTransferSize < GRPC_DEFAULT_MAX_RECV_MESSAGE_LENGTH,
-              "Max batch transfer size too large.");
+#endif  // INCLUDED_SRC_BUILDTOOL_EXECUTION_API_COMMON_MESSAGE_LIMITS_HPP
