@@ -14,18 +14,32 @@
 
 #include "src/buildtool/execution_api/execution_service/execution_server.hpp"
 
-#include <algorithm>
+#include <chrono>
+#include <filesystem>
 #include <fstream>
+#include <functional>
+#include <map>
+#include <memory>
 #include <string>
+#include <thread>
+#include <unordered_map>
 #include <utility>
+#include <vector>
 
 #include "fmt/core.h"
+#include "google/protobuf/any.pb.h"
+#include "google/protobuf/repeated_ptr_field.h"
+#include "google/protobuf/timestamp.pb.h"
+#include "google/rpc/status.pb.h"
+#include "src/buildtool/common/artifact.hpp"
 #include "src/buildtool/common/artifact_digest.hpp"
 #include "src/buildtool/common/artifact_digest_factory.hpp"
 #include "src/buildtool/common/protocol_traits.hpp"
+#include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/execution_api/execution_service/operation_cache.hpp"
 #include "src/buildtool/execution_api/local/local_cas_reader.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/storage/garbage_collector.hpp"
 #include "src/utils/cpp/hex_string.hpp"
@@ -90,7 +104,7 @@ auto ExecutionServiceImpl::ToIExecutionAction(
     execution_action->SetCacheFlag(
         action.do_not_cache() ? IExecutionAction::CacheFlag::DoNotCacheOutput
                               : IExecutionAction::CacheFlag::CacheOutput);
-    return std::move(execution_action);
+    return execution_action;
 }
 
 auto ExecutionServiceImpl::ToBazelExecuteResponse(
@@ -146,7 +160,7 @@ auto ExecutionServiceImpl::ToBazelExecuteResponse(
 
     // we run the action locally, so no communication issues should happen
     bazel_response.mutable_status()->set_code(grpc::StatusCode::OK);
-    return std::move(bazel_response);
+    return bazel_response;
 }
 
 void ExecutionServiceImpl::WriteResponse(
@@ -318,7 +332,7 @@ namespace {
         (*out_dir.mutable_tree_digest()) =
             ArtifactDigestFactory::ToBazel(*cas_digest);
     }
-    return std::move(out_dir);
+    return out_dir;
 }
 
 [[nodiscard]] auto ToBazelOutputSymlink(std::string path,
@@ -349,7 +363,7 @@ namespace {
     }
 
     *(out_link.mutable_target()) = *std::move(content);
-    return std::move(out_link);
+    return out_link;
 }
 
 [[nodiscard]] auto ToBazelOutputFile(std::string path,
@@ -405,7 +419,7 @@ namespace {
             result_files.Add(ToBazelOutputFile(path, info));
         }
     }
-    return std::move(result);
+    return result;
 }
 
 [[nodiscard]] auto ToBazelAction(ArtifactDigest const& action_digest,
@@ -440,7 +454,7 @@ namespace {
             fmt::format("could not retrieve input root {} from cas",
                         input_root_digest->hash())};
     }
-    return std::move(action);
+    return action;
 }
 
 [[nodiscard]] auto ToBazelCommand(bazel_re::Action const& action,
@@ -463,6 +477,6 @@ namespace {
         return unexpected{fmt::format("Failed to parse command from blob {}",
                                       command_digest->hash())};
     }
-    return std::move(c);
+    return c;
 }
 }  // namespace
