@@ -14,43 +14,36 @@
 
 #include "src/buildtool/execution_api/execution_service/server_implementation.hpp"
 
-#include <iostream>
-#include <memory>
-
 #ifdef __unix__
-#include <sys/types.h>
+#include <unistd.h>
 #else
 #error "Non-unix is not supported yet"
 #endif
 
+#include <memory>
+#include <utility>
+#include <variant>
+#include <vector>
+
+#include <grpcpp/grpcpp.h>
+
 #include "fmt/core.h"
-#include "grpcpp/grpcpp.h"
 #include "nlohmann/json.hpp"
+#include "src/buildtool/auth/authentication.hpp"
 #include "src/buildtool/common/protocol_traits.hpp"
 #include "src/buildtool/common/remote/port.hpp"
+#include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/execution_api/execution_service/ac_server.hpp"
 #include "src/buildtool/execution_api/execution_service/bytestream_server.hpp"
 #include "src/buildtool/execution_api/execution_service/capabilities_server.hpp"
 #include "src/buildtool/execution_api/execution_service/cas_server.hpp"
 #include "src/buildtool/execution_api/execution_service/execution_server.hpp"
 #include "src/buildtool/execution_api/execution_service/operations_server.hpp"
+#include "src/buildtool/file_system/atomic.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
-
-namespace {
-template <typename T>
-auto TryWrite(std::string const& file, T const& content) noexcept -> bool {
-    std::ofstream of{file};
-    if (not of.good()) {
-        Logger::Log(LogLevel::Error,
-                    "Could not open {}. Make sure to have write permissions",
-                    file);
-        return false;
-    }
-    of << content;
-    return true;
-}
-}  // namespace
+#include "src/buildtool/storage/config.hpp"
+#include "src/utils/cpp/type_safe_arithmetic.hpp"
 
 auto ServerImpl::Create(std::optional<std::string> interface,
                         std::optional<int> port,
@@ -77,7 +70,7 @@ auto ServerImpl::Create(std::optional<std::string> interface,
     if (pid_file) {
         server.pid_file_ = std::move(*pid_file);
     }
-    return std::move(server);
+    return server;
 }
 
 auto ServerImpl::Run(gsl::not_null<LocalContext const*> const& local_context,
@@ -136,7 +129,8 @@ auto ServerImpl::Run(gsl::not_null<LocalContext const*> const& local_context,
         {"interface", interface_}, {"port", port_}, {"pid", pid}};
 
     if (not pid_file_.empty()) {
-        if (not TryWrite(pid_file_, pid)) {
+        if (not FileSystemAtomic::WriteFile(pid_file_,
+                                            fmt::format("{}", pid))) {
             server->Shutdown();
             return false;
         }
@@ -149,7 +143,7 @@ auto ServerImpl::Run(gsl::not_null<LocalContext const*> const& local_context,
                 info_str);
 
     if (not info_file_.empty()) {
-        if (not TryWrite(info_file_, info_str)) {
+        if (not FileSystemAtomic::WriteFile(info_file_, info_str)) {
             server->Shutdown();
             return false;
         }
