@@ -18,6 +18,7 @@
 #include <algorithm>
 #include <atomic>
 #include <cstddef>
+#include <functional>
 #include <map>
 #include <memory>
 #include <optional>
@@ -25,7 +26,6 @@
 #include <unordered_map>
 #include <unordered_set>
 #include <utility>  // std::move
-#include <variant>
 #include <vector>
 
 #include "gsl/gsl"
@@ -33,11 +33,10 @@
 #include "src/buildtool/common/action_description.hpp"
 #include "src/buildtool/common/artifact.hpp"
 #include "src/buildtool/common/artifact_description.hpp"
-#include "src/buildtool/file_system/object_type.hpp"
+#include "src/buildtool/common/identifier.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
 #include "src/utils/cpp/hex_string.hpp"
-#include "src/utils/cpp/transformed_range.hpp"
 #include "src/utils/cpp/type_safe_arithmetic.hpp"
 
 /// \brief Plain DirectedAcyclicGraph.
@@ -263,9 +262,6 @@ class DependencyGraph : DirectedAcyclicGraph {
             Action::LocalPath path;
             base::OtherNodePtr node;
         };
-        using LocalPaths =
-            TransformedRange<std::vector<NamedOtherNodePtr>::const_iterator,
-                             Action::LocalPath>;
 
         [[nodiscard]] static auto Create(Action const& content) noexcept
             -> Ptr {
@@ -361,16 +357,19 @@ class DependencyGraph : DirectedAcyclicGraph {
             return Content().NoCache();
         }
 
-        [[nodiscard]] auto OutputFilePaths() const& noexcept -> LocalPaths {
-            return NodePaths(&output_files_);
+        [[nodiscard]] auto OutputFilePaths() const& noexcept
+            -> std::vector<Action::LocalPath> {
+            return NodePaths(output_files_);
         }
 
-        [[nodiscard]] auto OutputDirPaths() const& noexcept -> LocalPaths {
-            return NodePaths(&output_dirs_);
+        [[nodiscard]] auto OutputDirPaths() const& noexcept
+            -> std::vector<Action::LocalPath> {
+            return NodePaths(output_dirs_);
         }
 
-        [[nodiscard]] auto DependencyPaths() const& noexcept -> LocalPaths {
-            return NodePaths(&dependencies_);
+        [[nodiscard]] auto DependencyPaths() const& noexcept
+            -> std::vector<Action::LocalPath> {
+            return NodePaths(dependencies_);
         }
 
         // To initialise the action traversal specific data before traversing
@@ -392,12 +391,15 @@ class DependencyGraph : DirectedAcyclicGraph {
             std::make_unique<ActionNodeTraversalState>()};
 
         [[nodiscard]] static auto NodePaths(
-            gsl::not_null<std::vector<NamedOtherNodePtr> const*> const& nodes)
-            -> LocalPaths {
-            return TransformedRange{
-                nodes->begin(),
-                nodes->end(),
-                [](NamedOtherNodePtr const& node) { return node.path; }};
+            std::vector<NamedOtherNodePtr> const& nodes)
+            -> std::vector<Action::LocalPath> {
+            std::vector<Action::LocalPath> paths{nodes.size()};
+            std::transform(
+                nodes.cbegin(),
+                nodes.cend(),
+                paths.begin(),
+                [](auto const& named_node) { return named_node.path; });
+            return paths;
         }
     };
 
