@@ -17,41 +17,60 @@
 
 #include <algorithm>
 #include <chrono>
+#include <cmath>
+#include <exception>
+#include <filesystem>
 #include <functional>
-#include <iostream>
 #include <map>
+#include <memory>
+#include <mutex>
 #include <optional>
 #include <sstream>
-#include <type_traits>
+#include <string>
 #include <unordered_map>
+#include <unordered_set>
 #include <utility>
 #include <vector>
 
+#include "fmt/core.h"
 #include "gsl/gsl"
+#include "nlohmann/json.hpp"
 #include "src/buildtool/build_engine/expression/evaluator.hpp"
+#include "src/buildtool/build_engine/target_map/configured_target.hpp"
+#include "src/buildtool/common/action.hpp"
+#include "src/buildtool/common/artifact.hpp"
+#include "src/buildtool/common/artifact_blob.hpp"
 #include "src/buildtool/common/artifact_digest.hpp"
 #include "src/buildtool/common/artifact_digest_factory.hpp"
 #include "src/buildtool/common/git_hashes_converter.hpp"
 #include "src/buildtool/common/protocol_traits.hpp"
+#include "src/buildtool/common/remote/remote_common.hpp"
+#include "src/buildtool/common/repository_config.hpp"
 #include "src/buildtool/common/statistics.hpp"
-#include "src/buildtool/common/tree.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
-#include "src/buildtool/execution_api/common/artifact_blob_container.hpp"
+#include "src/buildtool/execution_api/bazel_msg/execution_config.hpp"
+#include "src/buildtool/execution_api/common/api_bundle.hpp"
 #include "src/buildtool/execution_api/common/common_api.hpp"
+#include "src/buildtool/execution_api/common/execution_action.hpp"
 #include "src/buildtool/execution_api/common/execution_api.hpp"
+#include "src/buildtool/execution_api/common/execution_response.hpp"
 #include "src/buildtool/execution_api/remote/bazel/bazel_api.hpp"
 #include "src/buildtool/execution_api/remote/config.hpp"
 #include "src/buildtool/execution_api/remote/context.hpp"
 #include "src/buildtool/execution_engine/dag/dag.hpp"
 #include "src/buildtool/execution_engine/executor/context.hpp"
-#include "src/buildtool/file_system/file_system_manager.hpp"
+#include "src/buildtool/file_system/file_root.hpp"
+#include "src/buildtool/file_system/git_tree.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
 #include "src/buildtool/progress_reporting/progress.hpp"
+#include "src/buildtool/progress_reporting/task_tracker.hpp"
+#include "src/utils/cpp/back_map.hpp"
+#include "src/utils/cpp/expected.hpp"
 #include "src/utils/cpp/hex_string.hpp"
 #include "src/utils/cpp/path_rebase.hpp"
 #include "src/utils/cpp/prefix.hpp"
-#include "src/utils/cpp/transformed_range.hpp"
 
 /// \brief Implementations for executing actions and uploading artifacts.
 class ExecutorImpl {
@@ -65,7 +84,6 @@ class ExecutorImpl {
         IExecutionApi const& api,
         ExecutionProperties const& merged_properties,
         gsl::not_null<RemoteContext const*> const& remote_context,
-        gsl::not_null<HashFunction const*> const& hash_function,
         std::chrono::milliseconds const& timeout,
         IExecutionAction::CacheFlag cache_flag,
         gsl::not_null<Statistics*> const& stats,
@@ -115,7 +133,7 @@ class ExecutorImpl {
 
         // get the alternative endpoint
         auto alternative_api = GetAlternativeEndpoint(
-            merged_properties, remote_context, hash_function);
+            merged_properties, remote_context, api.GetHashType());
         if (alternative_api) {
             if (not api.ParallelRetrieveToCas(
                     std::vector<Artifact::ObjectInfo>{Artifact::ObjectInfo{
@@ -131,10 +149,10 @@ class ExecutorImpl {
         }
 
         auto base = action->Content().Cwd();
-        auto cwd_relative_output_files = RebasePathStringsRelativeTo(
-            base, action->OutputFilePaths().ToVector());
-        auto cwd_relative_output_dirs = RebasePathStringsRelativeTo(
-            base, action->OutputDirPaths().ToVector());
+        auto cwd_relative_output_files =
+            RebasePathStringsRelativeTo(base, action->OutputFilePaths());
+        auto cwd_relative_output_dirs =
+            RebasePathStringsRelativeTo(base, action->OutputDirPaths());
         auto remote_action = (alternative_api ? *alternative_api : api)
                                  .CreateAction(*root_digest,
                                                action->Command(),
@@ -219,7 +237,6 @@ class ExecutorImpl {
                 }
 
                 if (not VerifyOrUploadKnownArtifact(
-                        apis.hash_function.GetType(),
                         *apis.remote,
                         artifact->Content().Repository(),
                         repo_config,
@@ -247,11 +264,8 @@ class ExecutorImpl {
             return oss.str();
         });
         auto repo = artifact->Content().Repository();
-        auto new_info = UploadFile(*apis.remote,
-                                   apis.hash_function,
-                                   repo,
-                                   repo_config,
-                                   *file_path_opt);
+        auto new_info =
+            UploadFile(*apis.remote, repo, repo_config, *file_path_opt);
         if (not new_info) {
             Logger::Log(LogLevel::Error,
                         "artifact in {} could not be uploaded to CAS.",
@@ -276,26 +290,17 @@ class ExecutorImpl {
                                                  GitTree const& tree) noexcept
         -> bool {
         // create list of digests for batch check of CAS availability
-        std::vector<ArtifactDigest> digests;
-        std::unordered_map<ArtifactDigest, gsl::not_null<GitTreeEntryPtr>>
-            entry_map;
-        for (auto const& [path, entry] : tree) {
-            // Since GitTrees are processed here, HashFunction::Type::GitSHA1 is
-            // used
-            auto digest =
-                ArtifactDigestFactory::Create(HashFunction::Type::GitSHA1,
-                                              entry->Hash(),
-                                              *entry->Size(),
-                                              entry->IsTree());
-            if (not digest) {
-                return false;
-            }
-            digests.emplace_back(*digest);
-            try {
-                entry_map.emplace(*std::move(digest), entry);
-            } catch (...) {
-                return false;
-            }
+        using ElementType = typename GitTree::entries_t::value_type;
+        auto const back_map = BackMap<ArtifactDigest, ElementType>::Make(
+            &tree, [](ElementType const& entry) {
+                return ArtifactDigestFactory::Create(
+                    HashFunction::Type::GitSHA1,
+                    entry.second->Hash(),
+                    *entry.second->Size(),
+                    entry.second->IsTree());
+            });
+        if (back_map == nullptr) {
+            return false;
         }
 
         Logger::Log(LogLevel::Trace, [&tree]() {
@@ -310,44 +315,46 @@ class ExecutorImpl {
         });
 
         // find missing digests
-        auto missing_digests = api.IsAvailable(digests);
+        auto const missing_digests = api.GetMissingDigests(back_map->GetKeys());
+        auto const missing_entries =
+            back_map->IterateReferences(&missing_digests);
 
         // process missing trees
-        for (auto const& digest : missing_digests) {
-            if (auto it = entry_map.find(digest); it != entry_map.end()) {
-                auto const& entry = it->second;
-                if (entry->IsTree()) {
-                    auto const& tree = entry->Tree();
-                    if (not tree or not VerifyOrUploadTree(api, *tree)) {
-                        return false;
-                    }
+        for (auto const& [_, value] : missing_entries) {
+            auto const entry = value->second;
+            if (entry->IsTree()) {
+                auto const& tree = entry->Tree();
+                if (not tree or not VerifyOrUploadTree(api, *tree)) {
+                    return false;
                 }
             }
         }
 
         // upload missing entries (blobs or trees)
-        ArtifactBlobContainer container;
-        for (auto const& digest : missing_digests) {
-            if (auto it = entry_map.find(digest); it != entry_map.end()) {
-                auto const& entry = it->second;
-                auto content = entry->RawData();
-                if (not content) {
-                    return false;
-                }
-                // store and/or upload blob, taking into account the maximum
-                // transfer size
-                if (not UpdateContainerAndUpload<ArtifactDigest>(
-                        &container,
-                        ArtifactBlob{digest,
-                                     std::move(*content),
-                                     IsExecutableObject(entry->Type())},
-                        /*exception_is_fatal=*/true,
-                        [&api](ArtifactBlobContainer&& blobs) {
-                            return api.Upload(std::move(blobs),
-                                              /*skip_find_missing=*/true);
-                        })) {
-                    return false;
-                }
+        HashFunction const hash_function{api.GetHashType()};
+        std::unordered_set<ArtifactBlob> container;
+        for (auto const& [digest, value] : missing_entries) {
+            auto const entry = value->second;
+            auto content = entry->RawData();
+            if (not content.has_value()) {
+                return false;
+            }
+            auto blob = ArtifactBlob::FromMemory(
+                hash_function, entry->Type(), *std::move(content));
+            if (not blob.has_value()) {
+                return false;
+            }
+            // store and/or upload blob, taking into account the maximum
+            // transfer size
+            if (not UpdateContainerAndUpload(
+                    &container,
+                    *std::move(blob),
+                    /*exception_is_fatal=*/true,
+                    [&api](std::unordered_set<ArtifactBlob>&& blobs) {
+                        return api.Upload(std::move(blobs),
+                                          /*skip_find_missing=*/true);
+                    })) {
+                return false;
             }
         }
         // upload remaining blobs
@@ -393,10 +400,14 @@ class ExecutorImpl {
             return false;
         }
 
-        return api.Upload(ArtifactBlobContainer{{ArtifactBlob{
-                              info.digest,
-                              std::move(*content),
-                              IsExecutableObject(info.type)}}},
+        auto blob = ArtifactBlob::FromMemory(
+            HashFunction{api.GetHashType()}, info.type, *std::move(content));
+        if (not blob.has_value()) {
+            Logger::Log(LogLevel::Error, "failed to create ArtifactBlob");
+            return false;
+        }
+
+        return api.Upload({*std::move(blob)},
                           /*skip_find_missing=*/true);
     }
 
@@ -439,12 +450,11 @@ class ExecutorImpl {
     /// \param info         The info of the object
     /// \returns true on success
     [[nodiscard]] static auto VerifyOrUploadKnownArtifact(
-        HashFunction::Type hash_type,
         IExecutionApi const& api,
         std::string const& repo,
         gsl::not_null<const RepositoryConfig*> const& repo_config,
         Artifact::ObjectInfo const& info) noexcept -> bool {
-        if (not ProtocolTraits::IsNative(hash_type)) {
+        if (not ProtocolTraits::IsNative(api.GetHashType())) {
             auto opt =
                 GitHashesConverter::Instance().GetGitEntry(info.digest.hash());
             if (opt) {
@@ -466,7 +476,6 @@ class ExecutorImpl {
     /// \returns The computed object info on success
     [[nodiscard]] static auto UploadFile(
         IExecutionApi const& api,
-        HashFunction hash_function,
         std::string const& repo,
         gsl::not_null<const RepositoryConfig*> const& repo_config,
         std::filesystem::path const& file_path) noexcept
@@ -483,12 +492,14 @@ class ExecutorImpl {
         if (not content.has_value()) {
             return std::nullopt;
         }
-        auto digest = ArtifactDigestFactory::HashDataAs<ObjectType::File>(
-            hash_function, *content);
-        if (not api.Upload(ArtifactBlobContainer{
-                {ArtifactBlob{digest,
-                              std::move(*content),
-                              IsExecutableObject(*object_type)}}})) {
+
+        auto blob = ArtifactBlob::FromMemory(
+            HashFunction{api.GetHashType()}, *object_type, *std::move(content));
+        if (not blob.has_value()) {
+            return std::nullopt;
+        }
+        auto digest = blob->GetDigest();
+        if (not api.Upload({*std::move(blob)})) {
             return std::nullopt;
         }
         return Artifact::ObjectInfo{.digest = std::move(digest),
@@ -536,7 +547,7 @@ class ExecutorImpl {
     /// are present in the artifacts map
     [[nodiscard]] static auto CheckOutputsExist(
         IExecutionResponse::ArtifactInfos const& artifacts,
-        DependencyGraph::ActionNode::LocalPaths const& outputs,
+        std::vector<Action::LocalPath> const& outputs,
         std::string base) noexcept -> bool {
         return std::all_of(
             outputs.begin(),
@@ -560,6 +571,7 @@ class ExecutorImpl {
 
         if (not response) {
             logger.Emit(LogLevel::Trace, "response is empty");
+            PrintError(logger, action, progress);
             return false;
         }
 
@@ -573,17 +585,15 @@ class ExecutorImpl {
         progress->TaskTracker().Stop(action->Content().Id());
 
         PrintInfo(logger, action, response);
+        bool action_failed = false;
         bool should_fail_outputs = false;
         for (auto const& [local_path, node] : action->Dependencies()) {
             should_fail_outputs |= node->Content().Info()->failed;
         }
         if (response->ExitCode() != 0) {
             if (action->MayFail()) {
-                logger.Emit(LogLevel::Warning,
-                            "{} (exit code {})",
-                            *(action->MayFail()),
-                            response->ExitCode());
                 should_fail_outputs = true;
+                action_failed = true;
             }
             else {
                 logger.Emit(LogLevel::Error,
@@ -600,30 +610,56 @@ class ExecutorImpl {
             return false;
         }
 
+        auto const output_files = action->OutputFilePaths();
+        auto const output_dirs = action->OutputDirPaths();
+
         if (artifacts.value()->empty() or
-            not CheckOutputsExist(*artifacts.value(),
-                                  action->OutputFilePaths(),
-                                  action->Content().Cwd()) or
-            not CheckOutputsExist(*artifacts.value(),
-                                  action->OutputDirPaths(),
-                                  action->Content().Cwd())) {
-            logger.Emit(LogLevel::Error, [&] {
-                std::string message{
-                    "action executed with missing outputs.\n"
-                    " Action outputs should be the following artifacts:"};
-                for (auto const& output : action->OutputFilePaths()) {
-                    message += "\n  - file: " + output;
+            not CheckOutputsExist(
+                *artifacts.value(), output_files, action->Content().Cwd()) or
+            not CheckOutputsExist(
+                *artifacts.value(), output_dirs, action->Content().Cwd())) {
+            logger.Emit(LogLevel::Error, [&]() {
+                std::ostringstream message{};
+                if (action_failed) {
+                    message << *(action->MayFail()) << " (exit code "
+                            << response->ExitCode() << ")\nMoreover ";
                 }
-                for (auto const& output : action->OutputDirPaths()) {
-                    message += "\n  - dir: " + output;
+                message << "action executed with missing outputs.\nAction "
+                           "outputs should be the following artifacts:";
+                for (auto const& output : output_files) {
+                    message << "\n  - file: " << output;
                 }
-                return message;
+                for (auto const& output : output_dirs) {
+                    message << "\n  - dir: " << output;
+                }
+                return message.str();
             });
             PrintError(logger, action, progress);
             return false;
         }
 
         SaveObjectInfo(*artifacts.value(), action, should_fail_outputs);
+        if (action_failed) {
+            logger.Emit(LogLevel::Warning, [&]() {
+                std::ostringstream message{};
+                auto base = action->Content().Cwd();
+                message << *(action->MayFail()) << " (exit code "
+                        << response->ExitCode() << "); outputs:";
+                for (auto const& [name, node] : action->OutputFiles()) {
+                    message << "\n - " << nlohmann::json(name).dump() << " "
+                            << artifacts.value()
+                                   ->at(RebasePathStringRelativeTo(base, name))
+                                   .ToString();
+                }
+                for (auto const& [name, node] : action->OutputDirs()) {
+                    message << "\n - " << nlohmann::json(name).dump() << " "
+                            << artifacts.value()
+                                   ->at(RebasePathStringRelativeTo(base, name))
+                                   .ToString();
+                }
+                return message.str();
+            });
+        }
 
         return true;
     }
@@ -640,7 +676,7 @@ class ExecutorImpl {
         }
         auto const has_err = response->HasStdErr();
         auto const has_out = response->HasStdOut();
-        auto build_message = [has_err, has_out, &logger, &action, &response]() {
+        auto build_message = [has_err, has_out, &action, &response]() {
             using namespace std::string_literals;
             auto message = ""s;
             bool has_both = has_err and has_out;
@@ -721,8 +757,7 @@ class ExecutorImpl {
     [[nodiscard]] static auto GetAlternativeEndpoint(
         const ExecutionProperties& properties,
         const gsl::not_null<RemoteContext const*>& remote_context,
-        const gsl::not_null<HashFunction const*>& hash_function)
-        -> std::unique_ptr<BazelApi> {
+        HashFunction::Type hash_type) -> std::unique_ptr<BazelApi> {
         for (auto const& [pred, endpoint] :
              remote_context->exec_config->dispatch) {
             bool match = true;
@@ -745,7 +780,7 @@ class ExecutorImpl {
                     remote_context->auth,
                     remote_context->retry_config,
                     config,
-                    hash_function);
+                    HashFunction{hash_type});
             }
         }
         return nullptr;
@@ -788,7 +823,6 @@ class Executor {
                     context_.remote_context->exec_config->platform_properties,
                     action->ExecutionProperties()),
                 context_.remote_context,
-                &context_.apis->hash_function,
                 Impl::ScaleTime(timeout_, action->TimeoutScale()),
                 action->NoCache() ? CF::DoNotCacheOutput : CF::CacheOutput,
                 context_.statistics,
@@ -811,7 +845,6 @@ class Executor {
                 context_.remote_context->exec_config->platform_properties,
                 action->ExecutionProperties()),
             context_.remote_context,
-            &context_.apis->hash_function,
             Impl::ScaleTime(timeout_, action->TimeoutScale()),
             action->NoCache() ? CF::DoNotCacheOutput : CF::CacheOutput,
             context_.statistics,
@@ -885,7 +918,6 @@ class Rebuilder {
                 context_.remote_context->exec_config->platform_properties,
                 action->ExecutionProperties()),
             context_.remote_context,
-            &context_.apis->hash_function,
             Impl::ScaleTime(timeout_, action->TimeoutScale()),
             CF::PretendCached,
             context_.statistics,
@@ -904,7 +936,6 @@ class Rebuilder {
                 context_.remote_context->exec_config->platform_properties,
                 action->ExecutionProperties()),
             context_.remote_context,
-            &context_.apis->hash_function,
             Impl::ScaleTime(timeout_, action->TimeoutScale()),
             CF::FromCacheOnly,
             context_.statistics,
