@@ -15,14 +15,20 @@
 #ifndef INCLUDED_SRC_BUILDTOOL_FILE_SYSTEM_FILE_ROOT_HPP
 #define INCLUDED_SRC_BUILDTOOL_FILE_SYSTEM_FILE_ROOT_HPP
 
+#include <algorithm>
+#include <cstddef>
+#include <exception>
 #include <filesystem>
+#include <functional>
 #include <iterator>
 #include <memory>
+#include <optional>
 #include <string>
 #include <unordered_map>
 #include <utility>
 #include <variant>
 
+#include "fmt/core.h"
 #include "gsl/gsl"
 #include "nlohmann/json.hpp"
 #include "src/buildtool/common/artifact_description.hpp"
@@ -32,11 +38,16 @@
 #include "src/buildtool/common/protocol_traits.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
+#include "src/buildtool/file_system/git_cas.hpp"
 #include "src/buildtool/file_system/git_tree.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
+#include "src/buildtool/file_system/precomputed_root.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
 #include "src/utils/cpp/concepts.hpp"
-#include "src/utils/cpp/json.hpp"
+#include "src/utils/cpp/expected.hpp"
+// Keep it to ensure fmt::format works on JSON objects
+#include "src/utils/cpp/json.hpp"  // IWYU pragma: keep
 
 /// FilteredIterator is an helper class to allow for iteration over
 /// directory-only or file-only entries stored inside the class
@@ -100,15 +111,18 @@ class FileRoot {
         gsl::not_null<GitCASPtr> cas;
         gsl::not_null<GitTreePtr> tree;
     };
+
     // absent roots are defined by a tree hash with no witnessing repository
     using absent_root_t = std::string;
-    using root_t = std::variant<fs_root_t, RootGit, absent_root_t>;
+    using root_t =
+        std::variant<fs_root_t, RootGit, absent_root_t, PrecomputedRoot>;
 
   public:
     static constexpr auto kGitTreeMarker = "git tree";
     static constexpr auto kGitTreeIgnoreSpecialMarker =
         "git tree ignore-special";
     static constexpr auto kFileIgnoreSpecialMarker = "file ignore-special";
+    static constexpr auto kComputedMarker = "computed";
 
     class DirectoryEntries {
         friend class FileRoot;
@@ -325,6 +339,8 @@ class FileRoot {
              gsl::not_null<GitTreePtr> const& tree,
              bool ignore_special = false) noexcept
         : root_{RootGit{cas, tree}}, ignore_special_{ignore_special} {}
+    explicit FileRoot(PrecomputedRoot precomputed)
+        : root_{std::move(precomputed)} {}
 
     [[nodiscard]] static auto FromGit(std::filesystem::path const& repo_path,
                                       std::string const& git_tree_id,
@@ -377,6 +393,25 @@ class FileRoot {
         return std::nullopt;
     }
 
+    [[nodiscard]] auto GetTreeHash() const noexcept
+        -> std::optional<std::string> {
+        if (auto const* root_git = std::get_if<RootGit>(&root_)) {
+            return root_git->tree->FileRootHash();
+        }
+        if (auto const* absent_root = std::get_if<absent_root_t>(&root_)) {
+            return *absent_root;
+        }
+        return std::nullopt;
+    }
+
+    [[nodiscard]] auto GetGitCasRoot() const noexcept
+        -> std::optional<std::filesystem::path> {
+        if (auto const* git_root = std::get_if<RootGit>(&root_)) {
+            return git_root->cas->GetPath();
+        }
+        return std::nullopt;
+    }
+
     // Indicates that subsequent calls to `Exists()`, `IsFile()`,
     // `IsDirectory()`, and `BlobType()` on contents of the same directory will
     // be served without any additional file system lookups.
@@ -504,7 +539,8 @@ class FileRoot {
                             return true;
                         },
                         /*allow_upwards=*/false,
-                        ignore_special_)) {
+                        ignore_special_,
+                        /*log_failure_at=*/LogLevel::Warning)) {
                     return DirectoryEntries{std::move(map)};
                 }
             }
@@ -625,6 +661,18 @@ class FileRoot {
         return std::nullopt;
     }
 
+    [[nodiscard]] auto IsPrecomputed() const noexcept -> bool {
+        return std::holds_alternative<PrecomputedRoot>(root_);
+    }
+
+    [[nodiscard]] auto GetPrecomputedDescription() const noexcept
+        -> std::optional<PrecomputedRoot> {
+        if (auto const* precomputed = std::get_if<PrecomputedRoot>(&root_)) {
+            return *precomputed;
+        }
+        return std::nullopt;
+    }
+
     [[nodiscard]] auto IgnoreSpecial() const noexcept -> bool {
         return ignore_special_;
     }
@@ -634,31 +682,30 @@ class FileRoot {
     /// nullopt on errors.
     [[nodiscard]] static auto ParseRoot(std::string const& repo,
                                         std::string const& keyword,
-                                        nlohmann::json const& root,
-                                        gsl::not_null<std::string*> error_msg)
-        -> std::optional<
-            std::pair<FileRoot, std::optional<std::filesystem::path>>> {
+                                        nlohmann::json const& root)
+        -> expected<std::pair<FileRoot, std::optional<std::filesystem::path>>,
+                    std::string> {
+        using ResultType =
+            std::pair<FileRoot, std::optional<std::filesystem::path>>;
         if ((not root.is_array()) or root.empty()) {
-            *error_msg = fmt::format(
+            return unexpected{fmt::format(
                 "Expected {} for {} to be of the form [<scheme>, ...], but "
                 "found {}",
                 keyword,
                 repo,
-                root.dump());
-            return std::nullopt;
+                root.dump())};
         }
         if (root[0] == "file") {
             if (root.size() != 2 or (not root[1].is_string())) {
-                *error_msg = fmt::format(
+                return unexpected{fmt::format(
                     "\"file\" scheme expects precisely one string argument, "
                     "but found {} for {} of repository {}",
                     root.dump(),
                     keyword,
-                    repo);
-                return std::nullopt;
+                    repo)};
             }
             auto path = std::filesystem::path{root[1].get<std::string>()};
-            return std::pair(FileRoot{path}, std::move(path));
+            return ResultType{FileRoot{path}, std::move(path)};
         }
         if (root[0] == FileRoot::kGitTreeMarker) {
             bool const has_one_arg = root.size() == 2 and root[1].is_string();
@@ -666,43 +713,40 @@ class FileRoot {
                                       root[1].is_string() and
                                       root[2].is_string();
             if (not has_one_arg and not has_two_args) {
-                *error_msg = fmt::format(
+                return unexpected{fmt::format(
                     "\"git tree\" scheme expects one or two string "
                     "arguments, but found {} for {} of repository {}",
                     root.dump(),
                     keyword,
-                    repo);
-                return std::nullopt;
+                    repo)};
             }
             if (root.size() == 3) {
                 if (auto git_root = FileRoot::FromGit(root[2], root[1])) {
-                    return std::pair(std::move(*git_root), std::nullopt);
+                    return ResultType{std::move(*git_root), std::nullopt};
                 }
-                *error_msg = fmt::format(
+                return unexpected{fmt::format(
                     "Could not create file root for {}tree id {}",
                     root.size() == 3
                         ? fmt::format("git repository {} and ", root[2])
                         : "",
-                    root[1]);
-                return std::nullopt;
+                    root[1])};
             }
             // return absent root
-            return std::pair(FileRoot{std::string{root[1]}}, std::nullopt);
+            return ResultType{FileRoot{std::string{root[1]}}, std::nullopt};
         }
         if (root[0] == FileRoot::kFileIgnoreSpecialMarker) {
             if (root.size() != 2 or (not root[1].is_string())) {
-                *error_msg = fmt::format(
+                return unexpected{fmt::format(
                     "\"file ignore-special\" scheme expects precisely "
                     "one string "
                     "argument, but found {} for {} of repository {}",
                     root.dump(),
                     keyword,
-                    repo);
-                return std::nullopt;
+                    repo)};
             }
             auto path = std::filesystem::path{root[1].get<std::string>()};
-            return std::pair(FileRoot{path, /*ignore_special=*/true},
-                             std::move(path));
+            return ResultType{FileRoot{path, /*ignore_special=*/true},
+                              std::move(path)};
         }
         if (root[0] == FileRoot::kGitTreeIgnoreSpecialMarker) {
             bool const has_one_arg = root.size() == 2 and root[1].is_string();
@@ -710,39 +754,48 @@ class FileRoot {
                                       root[1].is_string() and
                                       root[2].is_string();
             if (not has_one_arg and not has_two_args) {
-                *error_msg = fmt::format(
+                return unexpected{fmt::format(
                     "\"git tree ignore-special\" scheme expects one or two "
                     "string arguments, but found {} for {} of repository {}",
                     root.dump(),
                     keyword,
-                    repo);
-                return std::nullopt;
+                    repo)};
             }
             if (root.size() == 3) {
                 if (auto git_root = FileRoot::FromGit(
                         root[2], root[1], /*ignore_special=*/true)) {
-                    return std::pair(std::move(*git_root), std::nullopt);
+                    return ResultType{std::move(*git_root), std::nullopt};
                 }
-                *error_msg = fmt::format(
+                return unexpected{fmt::format(
                     "Could not create ignore-special file root for {}tree id "
                     "{}",
                     root.size() == 3
                         ? fmt::format("git repository {} and ", root[2])
                         : "",
-                    root[1]);
-                return std::nullopt;
+                    root[1])};
             }
             // return absent root
-            return std::pair(
+            return ResultType{
                 FileRoot{std::string{root[1]}, /*ignore_special=*/true},
-                std::nullopt);
+                std::nullopt};
         }
-        *error_msg = fmt::format(
+        if (PrecomputedRoot::IsPrecomputedMarker(root[0])) {
+            auto precomputed = PrecomputedRoot::Parse(root);
+            if (not precomputed) {
+                return unexpected{
+                    fmt::format("While parsing {} for {} of repository{}:\n{}",
+                                root.dump(),
+                                keyword,
+                                repo,
+                                std::move(precomputed).error())};
+            }
+            return ResultType{FileRoot{*std::move(precomputed)}, std::nullopt};
+        }
+        return unexpected{fmt::format(
             "Unknown scheme in the specification {} of {} of repository {}",
             root.dump(),
             keyword,
-            repo);
-        return std::nullopt;
+            repo)};
     }
 
   private:
