@@ -15,22 +15,6 @@
 #ifndef INCLUDED_SRC_BUILDTOOL_FILE_SYSTEM_FILE_SYSTEM_MANAGER_HPP
 #define INCLUDED_SRC_BUILDTOOL_FILE_SYSTEM_FILE_SYSTEM_MANAGER_HPP
 
-#include <array>
-#include <cerrno>  // for errno
-#include <chrono>
-#include <cstddef>
-#include <cstdint>
-#include <cstdio>   // for std::fopen
-#include <cstdlib>  // std::exit, std::getenv
-#include <cstring>
-#include <exception>
-#include <filesystem>
-#include <fstream>
-#include <optional>
-#include <system_error>
-#include <unordered_set>
-#include <variant>
-
 #ifdef __unix__
 #include <fcntl.h>
 #include <pwd.h>
@@ -42,6 +26,29 @@
 #error "Non-unix is not supported yet"
 #endif
 
+#include <algorithm>
+#include <array>
+#include <cerrno>  // for errno
+#include <chrono>
+#include <cstddef>
+#include <cstdint>
+#include <cstdio>   // for std::fopen
+#include <cstdlib>  // std::exit, std::getenv
+#include <cstring>
+#include <ctime>
+#include <exception>
+#include <filesystem>
+#include <fstream>
+#include <functional>
+#include <optional>
+#include <string>
+#include <string_view>
+#include <system_error>
+#include <unordered_set>
+#include <utility>
+#include <variant>
+
+#include "fmt/core.h"
 #include "gsl/gsl"
 #include "nlohmann/json.hpp"
 #include "src/buildtool/file_system/object_type.hpp"
@@ -49,6 +56,7 @@
 #include "src/buildtool/logging/logger.hpp"
 #include "src/buildtool/system/system.hpp"
 #include "src/utils/cpp/expected.hpp"
+#include "src/utils/cpp/incremental_reader.hpp"
 #include "src/utils/cpp/path.hpp"
 
 namespace detail {
@@ -696,28 +704,30 @@ class FileSystemManager {
                         file.string());
             return std::nullopt;
         }
-        try {
-            std::string chunk{};
-            std::string content{};
-            chunk.resize(kChunkSize);
-            std::ifstream file_reader(file.string(), std::ios::binary);
-            if (file_reader.is_open()) {
-                auto ssize = gsl::narrow<std::streamsize>(chunk.size());
-                while (file_reader.good()) {
-                    file_reader.read(chunk.data(), ssize);
-                    auto count = file_reader.gcount();
-                    if (count == ssize) {
-                        content += chunk;
-                    }
-                    else {
-                        content +=
-                            chunk.substr(0, gsl::narrow<std::size_t>(count));
-                    }
-                }
-                file_reader.close();
-                return content;
-            }
+
+        auto const to_read = IncrementalReader::FromFile(kChunkSize, file);
+        if (not to_read.has_value()) {
+            Logger::Log(LogLevel::Debug,
+                        "FileSystemManager: failed to create reader for {}\n{}",
+                        file.string(),
+                        to_read.error());
             return std::nullopt;
+        }
+
+        try {
+            std::string content{};
+            content.reserve(to_read->GetContentSize());
+            for (auto chunk : *to_read) {
+                if (not chunk.has_value()) {
+                    Logger::Log(LogLevel::Error,
+                                "reading file failed {}:\n{}",
+                                file.string(),
+                                chunk.error());
+                    return std::nullopt;
+                }
+                content.append(*chunk);
+            }
+            return content;
         } catch (std::exception const& e) {
             Logger::Log(LogLevel::Error,
                         "reading file {}:\n{}",
@@ -735,7 +745,8 @@ class FileSystemManager {
         std::filesystem::path const& dir,
         ReadDirEntryFunc const& read_entry,
         bool allow_upwards = false,
-        bool ignore_special = false) noexcept -> bool {
+        bool ignore_special = false,
+        LogLevel log_failure_at = LogLevel::Debug) noexcept -> bool {
         try {
             for (auto const& entry : std::filesystem::directory_iterator{dir}) {
                 ObjectType type{};
@@ -765,7 +776,7 @@ class FileSystemManager {
                         }
                         else {
                             Logger::Log(
-                                LogLevel::Error,
+                                log_failure_at,
                                 "unsupported upwards symlink dir entry {}",
                                 entry.path().string());
                             return false;
@@ -776,7 +787,7 @@ class FileSystemManager {
                     }
                 }
                 else {
-                    Logger::Log(LogLevel::Error,
+                    Logger::Log(log_failure_at,
                                 "unsupported type for dir entry {}",
                                 entry.path().string());
                     return false;
@@ -787,7 +798,7 @@ class FileSystemManager {
             }
         } catch (std::exception const& ex) {
             Logger::Log(
-                LogLevel::Error, "reading directory {} failed", dir.string());
+                log_failure_at, "reading directory {} failed", dir.string());
             return false;
         }
         return true;
@@ -1223,9 +1234,10 @@ class FileSystemManager {
                 ssize_t wlen{};
                 ssize_t written_len{};
                 while (written_len < len and
-                       (wlen = write(out.fd,
-                                     buf.data() + written_len,  // NOLINT
-                                     len - written_len)) > 0) {
+                       (wlen = write(
+                            out.fd,
+                            buf.data() + written_len,  // NOLINT
+                            static_cast<std::size_t>(len - written_len))) > 0) {
                     written_len += wlen;
                 }
                 if (wlen < 0) {
@@ -1254,7 +1266,7 @@ class FileSystemManager {
                 if (len < 0) {
                     return PackError(ERROR_WRITE_OUTPUT, errno);
                 }
-                pos += len;
+                pos += static_cast<std::size_t>(len);
             }
             return 0;
         }
