@@ -19,13 +19,13 @@
 #include <filesystem>
 #include <memory>
 #include <optional>
-#include <shared_mutex>
-#include <unordered_map>
+#include <string>
 #include <utility>
-#include <vector>
 
-#include "src/buildtool/file_system/git_context.hpp"
+#include "gsl/gsl"
 #include "src/buildtool/file_system/git_utils.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
+#include "src/buildtool/logging/log_level.hpp"
 
 class GitCAS;
 using GitCASPtr = std::shared_ptr<GitCAS const>;
@@ -33,8 +33,11 @@ using GitCASPtr = std::shared_ptr<GitCAS const>;
 /// \brief Git CAS that maintains its Git context.
 class GitCAS {
   public:
-    static auto Open(std::filesystem::path const& repo_path) noexcept
-        -> GitCASPtr;
+    [[nodiscard]] static auto Open(
+        std::filesystem::path const& repo_path,
+        LogLevel log_failure = LogLevel::Warning) noexcept -> GitCASPtr;
+
+    [[nodiscard]] static auto CreateEmpty() noexcept -> GitCASPtr;
 
     GitCAS() noexcept;
     ~GitCAS() noexcept = default;
@@ -45,10 +48,28 @@ class GitCAS {
     auto operator=(GitCAS const&) = delete;
     auto operator=(GitCAS&& other) = delete;
 
+    [[nodiscard]] auto GetODB() const noexcept -> gsl::not_null<git_odb*> {
+        return odb_.get();
+    }
+
+    [[nodiscard]] auto GetRepository() const noexcept
+        -> gsl::not_null<git_repository*> {
+        return repo_.get();
+    }
+
+    [[nodiscard]] auto GetPath() const noexcept
+        -> std::filesystem::path const& {
+        return git_path_;
+    }
+
     /// \brief Read object from CAS.
-    /// \param id         The object id.
-    /// \param is_hex_id  Specify whether `id` is hex string or raw.
-    [[nodiscard]] auto ReadObject(std::string const& id, bool is_hex_id = false)
+    /// \param id          The object id.
+    /// \param is_hex_id   Specify whether `id` is hex string or raw.
+    /// \param log_failure Log level at which to log failures accessing the
+    /// object.
+    [[nodiscard]] auto ReadObject(std::string const& id,
+                                  bool is_hex_id = false,
+                                  LogLevel log_failure = LogLevel::Warning)
         const noexcept -> std::optional<std::string>;
 
     /// \brief Read object header from CAS.
@@ -63,18 +84,11 @@ class GitCAS {
 
   private:
     std::unique_ptr<git_odb, decltype(&odb_closer)> odb_{nullptr, odb_closer};
+    std::unique_ptr<git_repository, decltype(&repository_closer)> repo_{
+        nullptr,
+        repository_closer};
     // git folder path of repo
     std::filesystem::path git_path_;
-
-    // mutex to guard odb while setting up a "fake" repository; it needs to be
-    // uniquely owned while wrapping the odb, but then git operations are free
-    // to share it.
-    mutable std::shared_mutex mutex_;
-
-    [[nodiscard]] auto OpenODB(std::filesystem::path const& repo_path) noexcept
-        -> bool;
-
-    friend class GitRepo;  // allow access to ODB
 };
 
 #endif  // INCLUDED_SRC_BUILDTOOL_FILE_SYSTEM_GIT_CAS_HPP
