@@ -17,6 +17,8 @@
 
 #include <filesystem>
 #include <functional>
+#include <memory>
+#include <mutex>
 #include <optional>
 #include <string>
 #include <unordered_map>
@@ -27,6 +29,8 @@
 #include "src/buildtool/common/artifact_digest.hpp"
 #include "src/buildtool/file_system/git_cas.hpp"
 #include "src/buildtool/file_system/git_types.hpp"
+#include "src/buildtool/file_system/git_utils.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
 #include "src/buildtool/storage/config.hpp"
 #include "src/utils/cpp/expected.hpp"
 
@@ -312,38 +316,31 @@ class GitRepo {
     [[nodiscard]] auto GetConfigSnapshot() const noexcept
         -> std::shared_ptr<git_config>;
 
+    /// \brief Import source directory to target git repository
+    /// \param storage_config   Storage where the source must be imported to
+    /// \param source_dir       Directory to import
+    /// \param commit_message   Message of the commit
+    /// \param tagging_lock     Mutex to protect critical git operations
+    /// \return The tree id of the commited directory on success or an
+    /// unexpected error as string on failure.
+    [[nodiscard]] static auto ImportToGit(
+        StorageConfig const& storage_config,
+        std::filesystem::path const& source_dir,
+        std::string const& commit_message,
+        gsl::not_null<std::mutex*> const& tagging_lock) noexcept
+        -> expected<std::string, std::string>;
+
+    /// \brief Check that the given repository contains the given tree
+    /// \returns Flag reflecting whether the tree is present in the repository
+    /// or an error message on failure.
+    [[nodiscard]] static auto IsTreeInRepo(std::filesystem::path const& repo,
+                                           std::string const& tree_id) noexcept
+        -> expected<bool, std::string>;
+
   private:
-    /// \brief Wrapped git_repository with guarded destructor.
-    /// Kept privately nested to avoid misuse of its raw pointer members.
-    class GuardedRepo {
-      public:
-        GuardedRepo() noexcept = delete;
-        explicit GuardedRepo(std::shared_mutex* mutex) noexcept;
-        ~GuardedRepo() noexcept;
-
-        // prohibit moves and copies
-        GuardedRepo(GuardedRepo const&) = delete;
-        GuardedRepo(GuardedRepo&& other) = delete;
-        auto operator=(GuardedRepo const&) = delete;
-        auto operator=(GuardedRepo&& other) = delete;
-
-        // get the bare pointer
-        [[nodiscard]] auto Ptr() -> git_repository*;
-        [[nodiscard]] auto PtrRef() -> git_repository**;
-
-      private:
-        std::shared_mutex* mutex_;
-        git_repository* repo_{nullptr};
-    };
-
-    using GuardedRepoPtr = std::shared_ptr<GuardedRepo>;
-
-    // IMPORTANT! The GitCAS object must be defined before the repo object to
-    // keep the GitContext alive until cleanup ends.
-    GitCASPtr git_cas_{nullptr};
-    GuardedRepoPtr repo_{nullptr};
+    GitCASPtr git_cas_;
     // default to real repo, as that is non-thread-safe
-    bool is_repo_fake_{false};
+    bool is_repo_fake_;
 
   protected:
     /// \brief Open "fake" repository wrapper for existing CAS.
@@ -351,14 +348,6 @@ class GitRepo {
     /// \brief Open real repository at given location.
     explicit GitRepo(std::filesystem::path const& repo_path) noexcept;
 
-    [[nodiscard]] auto GetRepoRef() const noexcept -> GuardedRepoPtr;
-
-    [[nodiscard]] auto GetGitPath() const noexcept
-        -> std::filesystem::path const&;
-
-    [[nodiscard]] auto GetGitOdb() const noexcept
-        -> std::unique_ptr<git_odb, decltype(&odb_closer)> const&;
-
     using StoreDirEntryFunc =
         std::function<bool(std::filesystem::path const&, ObjectType type)>;
 
