@@ -15,7 +15,7 @@
 #include "src/buildtool/file_system/git_tree.hpp"
 
 #include <algorithm>
-#include <sstream>
+#include <functional>
 #include <vector>
 
 #include "src/buildtool/common/artifact_digest.hpp"
@@ -23,10 +23,6 @@
 #include "src/buildtool/logging/logger.hpp"
 #include "src/utils/cpp/path.hpp"
 
-extern "C" {
-#include <git2.h>
-}
-
 namespace {
 
 // resolve '.' and '..' in path.
