@@ -15,32 +15,52 @@
 #ifndef INCLUDED_SRC_BUILDTOOL_GRAPH_TRAVERSER_GRAPH_TRAVERSER_HPP
 #define INCLUDED_SRC_BUILDTOOL_GRAPH_TRAVERSER_GRAPH_TRAVERSER_HPP
 
+#ifdef __unix__
+#include <unistd.h>
+#else
+#error "Non-unix is not supported yet"
+#endif
+
 #include <algorithm>
+#include <atomic>
+#include <condition_variable>
 #include <cstddef>
 #include <cstdint>
-#include <cstdlib>
+#include <cstdio>
 #include <filesystem>
+#include <fstream>
 #include <functional>
+#include <iomanip>
+#include <iostream>
+#include <iterator>
 #include <map>
+#include <memory>
 #include <optional>
 #include <sstream>
 #include <string>
 #include <thread>
+#include <tuple>
 #include <unordered_map>
+#include <unordered_set>
 #include <utility>
 #include <vector>
 
 #include "fmt/core.h"
 #include "gsl/gsl"
+#include "nlohmann/json.hpp"
+#include "src/buildtool/common/action_description.hpp"
+#include "src/buildtool/common/artifact.hpp"
+#include "src/buildtool/common/artifact_blob.hpp"
+#include "src/buildtool/common/artifact_description.hpp"
 #include "src/buildtool/common/artifact_digest.hpp"
-#include "src/buildtool/common/artifact_digest_factory.hpp"
 #include "src/buildtool/common/cli.hpp"
+#include "src/buildtool/common/identifier.hpp"
+#include "src/buildtool/common/statistics.hpp"
 #include "src/buildtool/common/tree.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
-#include "src/buildtool/execution_api/common/artifact_blob_container.hpp"
+#include "src/buildtool/execution_api/common/api_bundle.hpp"
 #include "src/buildtool/execution_api/common/common_api.hpp"
 #include "src/buildtool/execution_api/common/execution_api.hpp"
-#include "src/buildtool/execution_api/remote/config.hpp"
 #include "src/buildtool/execution_api/utils/subobject.hpp"
 #include "src/buildtool/execution_engine/dag/dag.hpp"
 #include "src/buildtool/execution_engine/executor/context.hpp"
@@ -50,11 +70,11 @@
 #include "src/buildtool/file_system/jsonfs.hpp"
 #include "src/buildtool/file_system/object_type.hpp"
 #include "src/buildtool/logging/log_level.hpp"
-#include "src/buildtool/logging/log_sink_cmdline.hpp"
-#include "src/buildtool/logging/log_sink_file.hpp"
 #include "src/buildtool/logging/logger.hpp"
 #include "src/buildtool/progress_reporting/base_progress_reporter.hpp"
+#include "src/utils/cpp/expected.hpp"
 #include "src/utils/cpp/json.hpp"
+#include "src/utils/cpp/path.hpp"
 
 class GraphTraverser {
   public:
@@ -94,18 +114,18 @@ class GraphTraverser {
     [[nodiscard]] auto BuildAndStage(
         std::map<std::string, ArtifactDescription> const& artifact_descriptions,
         std::map<std::string, ArtifactDescription> const& runfile_descriptions,
-        std::vector<ActionDescription::Ptr> const& action_descriptions,
-        std::vector<std::string> const& blobs,
-        std::vector<Tree::Ptr> const& trees,
+        std::vector<ActionDescription::Ptr>&& action_descriptions,
+        std::vector<std::string>&& blobs,
+        std::vector<Tree::Ptr>&& trees,
         std::vector<ArtifactDescription>&& extra_artifacts = {}) const
         -> std::optional<BuildResult> {
         DependencyGraph graph;  // must outlive artifact_nodes
         auto artifacts = BuildArtifacts(&graph,
                                         artifact_descriptions,
                                         runfile_descriptions,
-                                        action_descriptions,
-                                        trees,
-                                        blobs,
+                                        std::move(action_descriptions),
+                                        std::move(trees),
+                                        std::move(blobs),
                                         extra_artifacts);
         if (not artifacts) {
             return std::nullopt;
@@ -138,7 +158,13 @@ class GraphTraverser {
                          rel_paths,
                          artifact_nodes,
                          runfile_descriptions);
-            MaybePrintToStdout(rel_paths, artifact_nodes);
+            MaybePrintToStdout(
+                rel_paths,
+                artifact_nodes,
+                artifact_descriptions.size() == 1
+                    ? std::optional<std::string>{artifact_descriptions.begin()
+                                                     ->first}
+                    : std::nullopt);
             return BuildResult{.output_paths = std::move(rel_paths),
                                .extra_infos = std::move(infos),
                                .failed_artifacts = failed_artifacts};
@@ -162,7 +188,13 @@ class GraphTraverser {
                      artifact_nodes,
                      runfile_descriptions);
 
-        MaybePrintToStdout(rel_paths, artifact_nodes);
+        MaybePrintToStdout(
+            rel_paths,
+            artifact_nodes,
+            artifact_descriptions.size() == 1
+                ? std::optional<std::string>{artifact_descriptions.begin()
+                                                 ->first}
+                : std::nullopt);
 
         return BuildResult{.output_paths = *output_paths,
                            .extra_infos = std::move(infos),
@@ -179,10 +211,10 @@ class GraphTraverser {
         if (not desc) {
             return std::nullopt;
         }
-        auto const [blobs, tree_descs, actions] = *desc;
+        auto [blobs, tree_descs, actions] = *std::move(desc);
 
         HashFunction::Type const hash_type =
-            context_.apis->hash_function.GetType();
+            context_.apis->local->GetHashType();
         std::vector<ActionDescription::Ptr> action_descriptions{};
         action_descriptions.reserve(actions.size());
         for (auto const& [id, description] : actions.items()) {
@@ -213,8 +245,11 @@ class GraphTraverser {
             artifact_descriptions.emplace(rel_path, std::move(*artifact));
         }
 
-        return BuildAndStage(
-            artifact_descriptions, {}, action_descriptions, blobs, trees);
+        return BuildAndStage(artifact_descriptions,
+                             {},
+                             std::move(action_descriptions),
+                             std::move(blobs),
+                             std::move(trees));
     }
 
   private:
@@ -278,26 +313,30 @@ class GraphTraverser {
     /// of the uploads fails, execution is terminated
     /// \param[in]  blobs   blobs to be uploaded
     [[nodiscard]] auto UploadBlobs(
-        std::vector<std::string> const& blobs) const noexcept -> bool {
-        ArtifactBlobContainer container;
-        for (auto const& blob : blobs) {
-            auto digest = ArtifactDigestFactory::HashDataAs<ObjectType::File>(
-                context_.apis->hash_function, blob);
+        std::vector<std::string>&& blobs) const noexcept -> bool {
+        std::unordered_set<ArtifactBlob> container;
+        HashFunction const hash_function{context_.apis->remote->GetHashType()};
+        for (auto& content : blobs) {
+            auto blob = ArtifactBlob::FromMemory(
+                hash_function, ObjectType::File, std::move(content));
+            if (not blob.has_value()) {
+                logger_->Emit(LogLevel::Trace, "Failed to create ArtifactBlob");
+                return false;
+            }
             Logger::Log(logger_, LogLevel::Trace, [&]() {
                 return fmt::format(
-                    "Will upload blob {}, its digest has id {} and size {}.",
-                    nlohmann::json(blob).dump(),
-                    digest.hash(),
-                    digest.size());
+                    "Will upload blob, its digest has id {} and size {}.",
+                    blob->GetDigest().hash(),
+                    blob->GetDigest().size());
             });
             // Store and/or upload blob, taking into account the maximum
             // transfer size.
-            if (not UpdateContainerAndUpload<ArtifactDigest>(
+            if (not UpdateContainerAndUpload(
                     &container,
-                    ArtifactBlob{std::move(digest), blob, /*is_exec=*/false},
+                    *std::move(blob),
                     /*exception_is_fatal=*/true,
-                    [&api =
-                         context_.apis->remote](ArtifactBlobContainer&& blobs) {
+                    [&api = context_.apis->remote](
+                        std::unordered_set<ArtifactBlob>&& blobs) {
                         return api->Upload(std::move(blobs));
                     },
                     logger_)) {
@@ -457,15 +496,15 @@ class GraphTraverser {
         gsl::not_null<DependencyGraph*> const& graph,
         std::map<std::string, ArtifactDescription> const& artifacts,
         std::map<std::string, ArtifactDescription> const& runfiles,
-        std::vector<ActionDescription::Ptr> const& actions,
-        std::vector<Tree::Ptr> const& trees,
-        std::vector<std::string> const& blobs,
+        std::vector<ActionDescription::Ptr>&& actions,
+        std::vector<Tree::Ptr>&& trees,
+        std::vector<std::string>&& blobs,
         std::vector<ArtifactDescription> const& extra_artifacts = {}) const
         -> std::optional<
             std::tuple<std::vector<std::filesystem::path>,
                        std::vector<DependencyGraph::ArtifactNode const*>,
                        std::vector<DependencyGraph::ArtifactNode const*>>> {
-        if (not UploadBlobs(blobs)) {
+        if (not UploadBlobs(std::move(blobs))) {
             return std::nullopt;
         }
 
@@ -651,8 +690,8 @@ class GraphTraverser {
 
     void MaybePrintToStdout(
         std::vector<std::filesystem::path> const& paths,
-        std::vector<DependencyGraph::ArtifactNode const*> const& artifacts)
-        const {
+        std::vector<DependencyGraph::ArtifactNode const*> const& artifacts,
+        std::optional<std::string> const& unique_artifact) const {
         if (clargs_.build.print_to_stdout) {
             auto const& remote = *context_.apis->remote;
             for (std::size_t i = 0; i < paths.size(); i++) {
@@ -661,7 +700,8 @@ class GraphTraverser {
                     if (info) {
                         if (not remote.RetrieveToFds({*info},
                                                      {dup(fileno(stdout))},
-                                                     /*raw_tree=*/false)) {
+                                                     /*raw_tree=*/false,
+                                                     &*context_.apis->local)) {
                             Logger::Log(logger_,
                                         LogLevel::Error,
                                         "Failed to retrieve {}",
@@ -700,9 +740,11 @@ class GraphTraverser {
                         auto new_info =
                             RetrieveSubPathId(*info, *context_.apis, relpath);
                         if (new_info) {
-                            if (not remote.RetrieveToFds({*new_info},
-                                                         {dup(fileno(stdout))},
-                                                         /*raw_tree=*/false)) {
+                            if (not remote.RetrieveToFds(
+                                    {*new_info},
+                                    {dup(fileno(stdout))},
+                                    /*raw_tree=*/false,
+                                    &*context_.apis->local)) {
                                 Logger::Log(logger_,
                                             LogLevel::Error,
                                             "Failed to retrieve artifact {} at "
@@ -728,6 +770,38 @@ class GraphTraverser {
                         "{} not a logical path of the specified target",
                         *(clargs_.build.print_to_stdout));
         }
+        else if (clargs_.build.print_unique) {
+            if (unique_artifact) {
+                auto const& remote = *context_.apis->remote;
+                std::optional<Artifact::ObjectInfo> info = std::nullopt;
+                for (std::size_t i = 0; i < paths.size(); i++) {
+                    if (paths[i] == unique_artifact) {
+                        info = artifacts[i]->Content().Info();
+                    }
+                }
+                if (info) {
+                    if (not remote.RetrieveToFds({*info},
+                                                 {dup(fileno(stdout))},
+                                                 /*raw_tree=*/false,
+                                                 &*context_.apis->local)) {
+                        Logger::Log(logger_,
+                                    LogLevel::Error,
+                                    "Failed to retrieve {}",
+                                    *unique_artifact);
+                    }
+                }
+                else {
+                    Logger::Log(logger_,
+                                LogLevel::Error,
+                                "Failed to obtain object information for {}",
+                                *unique_artifact);
+                }
+                return;
+            }
+            Logger::Log(logger_,
+                        LogLevel::Info,
+                        "Target does not have precisely one artifact.");
+        }
     }
 };
 
