@@ -12,34 +12,46 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-#include <algorithm>
+#include <cmath>
+#include <cstddef>
 #include <cstdlib>
+#include <exception>
 #include <filesystem>
 #include <fstream>
+#include <functional>
 #include <iostream>
 #include <map>
+#include <memory>
 #include <mutex>
 #include <optional>
+#include <set>
 #include <string>
-#include <unordered_map>
-#include <unordered_set>
 #include <utility>
 #include <variant>
 #include <vector>
 
 #include "gsl/gsl"
 #include "nlohmann/json.hpp"
+#include "src/buildtool/build_engine/analysed_target/analysed_target.hpp"
 #include "src/buildtool/build_engine/base_maps/entity_name.hpp"
+#include "src/buildtool/build_engine/base_maps/entity_name_data.hpp"
+#include "src/buildtool/build_engine/expression/configuration.hpp"
 #include "src/buildtool/build_engine/expression/evaluator.hpp"
 #include "src/buildtool/build_engine/expression/expression.hpp"
-#include "src/buildtool/build_engine/target_map/target_map.hpp"
+#include "src/buildtool/build_engine/expression/expression_ptr.hpp"
+#include "src/buildtool/build_engine/target_map/absent_target_map.hpp"
+#include "src/buildtool/build_engine/target_map/configured_target.hpp"
+#include "src/buildtool/build_engine/target_map/result_map.hpp"
 #include "src/buildtool/common/artifact_description.hpp"
+#include "src/buildtool/common/cli.hpp"
+#include "src/buildtool/common/clidefaults.hpp"
 #include "src/buildtool/common/protocol_traits.hpp"
 #include "src/buildtool/common/remote/remote_common.hpp"
 #include "src/buildtool/common/repository_config.hpp"
 #include "src/buildtool/common/statistics.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/file_system/file_root.hpp"
+#include "src/buildtool/file_system/file_system_manager.hpp"
 #include "src/buildtool/logging/log_config.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/log_sink_cmdline.hpp"
@@ -55,20 +67,19 @@
 #include "src/buildtool/main/exit_codes.hpp"
 #include "src/buildtool/main/install_cas.hpp"
 #include "src/buildtool/main/version.hpp"
-#include "src/buildtool/multithreading/async_map_consumer.hpp"
 #include "src/buildtool/multithreading/task_system.hpp"
 #include "src/buildtool/progress_reporting/progress.hpp"
 #include "src/buildtool/serve_api/remote/serve_api.hpp"
 #include "src/buildtool/storage/config.hpp"
 #include "src/buildtool/storage/file_chunker.hpp"
 #include "src/buildtool/storage/storage.hpp"
-#include "src/buildtool/storage/target_cache.hpp"
-#include "src/utils/cpp/concepts.hpp"
+#include "src/utils/cpp/expected.hpp"
 #include "src/utils/cpp/json.hpp"
 #ifndef BOOTSTRAP_BUILD_TOOL
 #include "fmt/core.h"
 #include "src/buildtool/auth/authentication.hpp"
 #include "src/buildtool/common/remote/retry_config.hpp"
+#include "src/buildtool/computed_roots/evaluate.hpp"
 #include "src/buildtool/execution_api/common/api_bundle.hpp"
 #include "src/buildtool/execution_api/execution_service/server_implementation.hpp"
 #include "src/buildtool/execution_api/local/config.hpp"
@@ -76,6 +87,7 @@
 #include "src/buildtool/execution_api/remote/config.hpp"
 #include "src/buildtool/execution_api/remote/context.hpp"
 #include "src/buildtool/execution_engine/executor/context.hpp"
+#include "src/buildtool/file_system/git_context.hpp"
 #include "src/buildtool/graph_traverser/graph_traverser.hpp"
 #include "src/buildtool/main/describe.hpp"
 #include "src/buildtool/main/retry.hpp"
@@ -85,7 +97,6 @@
 #include "src/buildtool/serve_api/serve_service/serve_server_implementation.hpp"
 #include "src/buildtool/storage/backend_description.hpp"
 #include "src/buildtool/storage/garbage_collector.hpp"
-#include "src/utils/cpp/gsl.hpp"
 #endif  // BOOTSTRAP_BUILD_TOOL
 
 namespace {
@@ -122,16 +133,22 @@ void SetupLogging(LogArguments const& clargs) {
         builder.SetBuildRoot(*eargs.local_root);
     }
 
+    auto backend_description = BackendDescription::Describe(
+        remote_address, remote_platform_properties, remote_dispatch);
+    if (not backend_description) {
+        Logger::Log(LogLevel::Error, std::move(backend_description).error());
+        return std::nullopt;
+    }
+
     auto config =
         builder.SetHashType(hash_type)
-            .SetRemoteExecutionArgs(
-                remote_address, remote_platform_properties, remote_dispatch)
+            .SetBackendDescription(std::move(backend_description).value())
             .Build();
-    if (config) {
-        return *std::move(config);
+    if (not config) {
+        Logger::Log(LogLevel::Error, std::move(config).error());
+        return std::nullopt;
     }
-    Logger::Log(LogLevel::Error, config.error());
-    return std::nullopt;
+    return *std::move(config);
 }
 
 #ifndef BOOTSTRAP_BUILD_TOOL
@@ -175,6 +192,7 @@ void SetupLogging(LogArguments const& clargs) {
     -> std::optional<RemoteServeConfig> {
     RemoteServeConfig::Builder builder;
     builder.SetRemoteAddress(srvargs.remote_serve_address)
+        .SetClientExecutionAddress(srvargs.client_remote_address)
         .SetKnownRepositories(srvargs.repositories)
         .SetJobs(cargs.jobs)
         .SetActionTimeout(bargs.timeout)
@@ -230,24 +248,6 @@ void SetupFileChunker() {
     FileChunker::Initialize();
 }
 
-/// \brief Write backend description (which determines the target cache shard)
-/// to CAS.
-void StoreTargetCacheShard(
-    StorageConfig const& storage_config,  // NOLINT(misc-unused-parameters)
-    Storage const& storage,
-    RemoteExecutionConfig const& remote_exec_config) noexcept {
-    auto backend_description =
-        DescribeBackend(remote_exec_config.remote_address,
-                        remote_exec_config.platform_properties,
-                        remote_exec_config.dispatch);
-    if (not backend_description) {
-        Logger::Log(LogLevel::Error, backend_description.error());
-        std::exit(kExitFailure);
-    }
-    [[maybe_unused]] auto id = storage.CAS().StoreBlob(*backend_description);
-    EnsuresAudit(id and id->hash() == storage_config.backend_description_id);
-}
-
 #endif  // BOOTSTRAP_BUILD_TOOL
 
 // returns path relative to `root`.
@@ -539,21 +539,21 @@ auto DetermineRoots(gsl::not_null<RepositoryConfig*> const& repository_config,
         repos[main_repo] = nlohmann::json::object();
     }
 
-    std::string error_msg;
     for (auto const& [repo, desc] : repos.items()) {
         std::optional<FileRoot> ws_root{};
         bool const is_main_repo{repo == main_repo};
         auto it_ws = desc.find("workspace_root");
         if (it_ws != desc.end()) {
-            if (auto parsed_root = FileRoot::ParseRoot(
-                    repo, "workspace_root", *it_ws, &error_msg)) {
-                ws_root = std::move(parsed_root->first);
-                if (is_main_repo and parsed_root->second.has_value()) {
-                    main_ws_root = std::move(parsed_root->second);
+            if (auto parsed_root =
+                    FileRoot::ParseRoot(repo, "workspace_root", *it_ws)) {
+                auto result = *std::move(parsed_root);
+                ws_root = std::move(result.first);
+                if (is_main_repo and result.second.has_value()) {
+                    main_ws_root = std::move(result.second);
                 }
             }
             else {
-                Logger::Log(LogLevel::Error, error_msg);
+                Logger::Log(LogLevel::Error, parsed_root.error());
                 std::exit(kExitFailure);
             }
         }
@@ -577,20 +577,18 @@ auto DetermineRoots(gsl::not_null<RepositoryConfig*> const& repository_config,
             std::exit(kExitFailure);
         }
         auto info = RepositoryConfig::RepositoryInfo{std::move(*ws_root)};
-        auto parse_keyword_root = [&desc = desc,
-                                   &repo = repo,
-                                   &error_msg = error_msg,
-                                   is_main_repo](FileRoot* keyword_root,
-                                                 std::string const& keyword,
-                                                 auto const& keyword_carg) {
+        auto parse_keyword_root = [&desc = desc, &repo = repo, is_main_repo](
+                                      FileRoot* keyword_root,
+                                      std::string const& keyword,
+                                      auto const& keyword_carg) {
             auto it = desc.find(keyword);
             if (it != desc.end()) {
                 if (auto parsed_root =
-                        FileRoot::ParseRoot(repo, keyword, *it, &error_msg)) {
-                    (*keyword_root) = parsed_root->first;
+                        FileRoot::ParseRoot(repo, keyword, *it)) {
+                    (*keyword_root) = std::move(parsed_root)->first;
                 }
                 else {
-                    Logger::Log(LogLevel::Error, error_msg);
+                    Logger::Log(LogLevel::Error, parsed_root.error());
                     std::exit(kExitFailure);
                 }
             }
@@ -805,8 +803,6 @@ auto main(int argc, char* argv[]) -> int {
                     return kExitFailure;
                 }
                 auto const storage = Storage::Create(&*storage_config);
-                StoreTargetCacheShard(
-                    *storage_config, storage, remote_exec_config);
 
                 // pack the local context instances to be passed as needed
                 LocalContext const local_context{
@@ -848,11 +844,13 @@ auto main(int argc, char* argv[]) -> int {
         }
 
         if (arguments.cmd == SubCommand::kServe) {
+            std::mutex lock{};
             auto serve_server =
                 ServeServerImpl::Create(arguments.service.interface,
                                         arguments.service.port,
                                         arguments.service.info_file,
-                                        arguments.service.pid_file);
+                                        arguments.service.pid_file,
+                                        &lock);
             if (serve_server) {
                 // Set up remote execution config.
                 auto remote_exec_config = CreateRemoteExecutionConfig(
@@ -872,8 +870,6 @@ auto main(int argc, char* argv[]) -> int {
                     return kExitFailure;
                 }
                 auto const storage = Storage::Create(&*storage_config);
-                StoreTargetCacheShard(
-                    *storage_config, storage, *remote_exec_config);
 
                 // pack the local context instances to be passed as needed
                 LocalContext const local_context{
@@ -954,10 +950,6 @@ auto main(int argc, char* argv[]) -> int {
         }
         auto const storage = Storage::Create(&*storage_config);
 
-#ifndef BOOTSTRAP_BUILD_TOOL
-        StoreTargetCacheShard(*storage_config, storage, *remote_exec_config);
-#endif  // BOOTSTRAP_BUILD_TOOL
-
         auto jobs = arguments.build.build_jobs > 0 ? arguments.build.build_jobs
                                                    : arguments.common.jobs;
 
@@ -994,16 +986,20 @@ auto main(int argc, char* argv[]) -> int {
                                             .remote_context = &remote_context,
                                             .statistics = &stats,
                                             .progress = &progress};
+        const GraphTraverser::CommandLineArguments traverse_args{
+            jobs,
+            std::move(arguments.build),
+            std::move(stage_args),
+            std::move(rebuild_args)};
+
         GraphTraverser const traverser{
-            {jobs,
-             std::move(arguments.build),
-             std::move(stage_args),
-             std::move(rebuild_args)},
+            traverse_args,
             &exec_context,
             ProgressReporter::Reporter(&stats, &progress)};
 
         if (arguments.cmd == SubCommand::kInstallCas) {
-            if (not repo_config.SetGitCAS(storage_config->GitRoot())) {
+            if (not repo_config.SetGitCAS(storage_config->GitRoot(),
+                                          LogLevel::Trace)) {
                 Logger::Log(LogLevel::Debug,
                             "Failed set Git CAS {}.",
                             storage_config->GitRoot().string());
@@ -1023,9 +1019,20 @@ auto main(int argc, char* argv[]) -> int {
         auto [main_repo, main_ws_root] =
             DetermineRoots(&repo_config, arguments.common, arguments.analysis);
 
+        std::size_t eval_root_jobs =
+            std::lround(std::ceil(std::sqrt(arguments.common.jobs)));
 #ifndef BOOTSTRAP_BUILD_TOOL
         std::optional<ServeApi> serve = ServeApi::Create(
             *serve_config, &local_context, &remote_context, &main_apis);
+        if (not EvaluatePrecomputedRoots(&repo_config,
+                                         main_repo,
+                                         serve ? &*serve : nullptr,
+                                         *storage_config,
+                                         traverse_args,
+                                         &exec_context,
+                                         eval_root_jobs)) {
+            return kExitFailure;
+        }
 #else
         std::optional<ServeApi> serve;
 #endif  // BOOTSTRAP_BUILD_TOOL
@@ -1047,7 +1054,8 @@ auto main(int argc, char* argv[]) -> int {
                                 "--compatible");
                     return kExitFailure;
                 }
-                if (not repo_config.SetGitCAS(*arguments.graph.git_cas)) {
+                if (not repo_config.SetGitCAS(*arguments.graph.git_cas,
+                                              LogLevel::Debug)) {
                     Logger::Log(LogLevel::Warning,
                                 "Failed set Git CAS {}.",
                                 arguments.graph.git_cas->string());
@@ -1081,7 +1089,6 @@ auto main(int argc, char* argv[]) -> int {
         }
 
 #endif  // BOOTSTRAP_BUILD_TOOL
-        BuildMaps::Target::ResultTargetMap result_map{arguments.common.jobs};
         auto id = ReadConfiguredTarget(
             main_repo, main_ws_root, &repo_config, arguments.analysis);
         auto serve_errors = nlohmann::json::array();
@@ -1106,13 +1113,13 @@ auto main(int argc, char* argv[]) -> int {
                                    .progress = &exports_progress,
                                    .serve = serve ? &*serve : nullptr};
 
-        auto result = AnalyseTarget(&analyse_ctx,
-                                    id,
-                                    &result_map,
-                                    arguments.common.jobs,
-                                    arguments.analysis.request_action_input,
-                                    /*logger=*/nullptr,
-                                    &collect_serve_errors);
+        auto analyse_result =
+            AnalyseTarget(&analyse_ctx,
+                          id,
+                          arguments.common.jobs,
+                          arguments.analysis.request_action_input,
+                          /*logger=*/nullptr,
+                          &collect_serve_errors);
         if (arguments.analysis.serve_errors_file) {
             Logger::Log(serve_errors.empty() ? LogLevel::Debug : LogLevel::Info,
                         "Dumping serve-error information to {}",
@@ -1120,11 +1127,11 @@ auto main(int argc, char* argv[]) -> int {
             std::ofstream os(*arguments.analysis.serve_errors_file);
             os << serve_errors.dump() << std::endl;
         }
-        if (result) {
-            Logger::Log(
-                LogLevel::Info,
-                "Analysed target {}",
-                result->id.ToShortString(Evaluator::GetExpressionLogLimit()));
+        if (analyse_result) {
+            Logger::Log(LogLevel::Info,
+                        "Analysed target {}",
+                        analyse_result->id.ToShortString(
+                            Evaluator::GetExpressionLogLimit()));
 
             {
                 auto cached = stats.ExportsCachedCounter();
@@ -1144,15 +1151,15 @@ auto main(int argc, char* argv[]) -> int {
             }
 
             if (arguments.analysis.graph_file) {
-                result_map.ToFile(
+                analyse_result->result_map.ToFile(
                     *arguments.analysis.graph_file, &stats, &progress);
             }
             if (arguments.analysis.graph_file_plain) {
-                result_map.ToFile</*kIncludeOrigins=*/false>(
+                analyse_result->result_map.ToFile</*kIncludeOrigins=*/false>(
                     *arguments.analysis.graph_file_plain, &stats, &progress);
             }
             auto const [artifacts, runfiles] =
-                ReadOutputArtifacts(result->target);
+                ReadOutputArtifacts(analyse_result->target);
             if (arguments.analysis.artifacts_to_build_file) {
                 DumpArtifactsToBuild(
                     artifacts,
@@ -1160,45 +1167,48 @@ auto main(int argc, char* argv[]) -> int {
                     *arguments.analysis.artifacts_to_build_file);
             }
             if (arguments.cmd == SubCommand::kAnalyse) {
-                DiagnoseResults(*result, result_map, arguments.diagnose);
-                ReportTaintedness(*result);
+                DiagnoseResults(*analyse_result, arguments.diagnose);
+                ReportTaintedness(*analyse_result);
                 // Clean up in parallel
                 {
                     TaskSystem ts{arguments.common.jobs};
-                    result_map.Clear(&ts);
+                    analyse_result->result_map.Clear(&ts);
                 }
                 return kExitSuccess;
             }
 #ifndef BOOTSTRAP_BUILD_TOOL
-            ReportTaintedness(*result);
-            auto const& [actions, blobs, trees] =
-                result_map.ToResult(&stats, &progress);
+            ReportTaintedness(*analyse_result);
+            auto [actions, blobs, trees] =
+                analyse_result->result_map.ToResult(&stats, &progress);
 
             // collect cache targets and artifacts for target-level caching
-            auto const cache_targets = result_map.CacheTargets();
+            auto const cache_targets =
+                analyse_result->result_map.CacheTargets();
             auto cache_artifacts = CollectNonKnownArtifacts(cache_targets);
 
-            // Clean up result map, now that it is no longer needed
+            // Clean up analyse_result map, now that it is no longer needed
             {
                 TaskSystem ts{arguments.common.jobs};
-                result_map.Clear(&ts);
+                analyse_result->result_map.Clear(&ts);
             }
 
             Logger::Log(
                 LogLevel::Info,
                 "{}ing{} {}.",
                 arguments.cmd == SubCommand::kRebuild ? "Rebuild" : "Build",
-                result->modified
-                    ? fmt::format(" input of action {} of", *(result->modified))
+                analyse_result->modified
+                    ? fmt::format(" input of action {} of",
+                                  *(analyse_result->modified))
                     : "",
-                result->id.ToShortString(Evaluator::GetExpressionLogLimit()));
+                analyse_result->id.ToShortString(
+                    Evaluator::GetExpressionLogLimit()));
 
             auto build_result =
                 traverser.BuildAndStage(artifacts,
                                         runfiles,
-                                        actions,
-                                        blobs,
-                                        trees,
+                                        std::move(actions),
+                                        std::move(blobs),
+                                        std::move(trees),
                                         std::move(cache_artifacts));
             if (build_result) {
                 WriteTargetCacheEntries(
@@ -1214,7 +1224,7 @@ auto main(int argc, char* argv[]) -> int {
 
                 // Repeat taintedness message to make the user aware that
                 // the artifacts are not for production use.
-                ReportTaintedness(*result);
+                ReportTaintedness(*analyse_result);
                 if (build_result->failed_artifacts) {
                     Logger::Log(LogLevel::Warning,
                                 "Build result contains failed artifacts.");
