@@ -17,9 +17,6 @@
 
 #include <memory>
 #include <optional>
-#include <string>
-#include <utility>
-#include <vector>
 
 #include "gsl/gsl"
 #include "justbuild/just_serve/just_serve.grpc.pb.h"
