@@ -16,21 +16,27 @@
 
 #include "src/buildtool/serve_api/serve_service/serve_server_implementation.hpp"
 
-#include <iostream>
-#include <memory>
-#include <variant>
-
 #ifdef __unix__
-#include <sys/types.h>
+#include <unistd.h>
 #else
 #error "Non-unix is not supported yet"
 #endif
 
+#include <filesystem>
+#include <memory>
+#include <utility>
+#include <variant>
+#include <vector>
+
+#include <grpcpp/grpcpp.h>
+
 #include "fmt/core.h"
-#include "grpcpp/grpcpp.h"
 #include "nlohmann/json.hpp"
+#include "src/buildtool/auth/authentication.hpp"
 #include "src/buildtool/common/protocol_traits.hpp"
 #include "src/buildtool/common/remote/port.hpp"
+#include "src/buildtool/crypto/hash_function.hpp"
+#include "src/buildtool/execution_api/common/execution_api.hpp"
 #include "src/buildtool/execution_api/execution_service/ac_server.hpp"
 #include "src/buildtool/execution_api/execution_service/bytestream_server.hpp"
 #include "src/buildtool/execution_api/execution_service/capabilities_server.hpp"
@@ -39,36 +45,26 @@
 #include "src/buildtool/execution_api/execution_service/operations_server.hpp"
 #include "src/buildtool/execution_api/local/local_api.hpp"
 #include "src/buildtool/execution_api/serve/mr_local_api.hpp"
+#include "src/buildtool/file_system/atomic.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
 #include "src/buildtool/file_system/git_repo.hpp"
 #include "src/buildtool/logging/log_level.hpp"
+#include "src/buildtool/logging/logger.hpp"
 #include "src/buildtool/serve_api/serve_service/configuration.hpp"
 #include "src/buildtool/serve_api/serve_service/source_tree.hpp"
 #include "src/buildtool/serve_api/serve_service/target.hpp"
 #include "src/buildtool/storage/config.hpp"
 #include "src/buildtool/storage/storage.hpp"
-
-namespace {
-template <typename T>
-auto TryWrite(std::string const& file, T const& content) noexcept -> bool {
-    std::ofstream of{file};
-    if (not of.good()) {
-        Logger::Log(LogLevel::Error,
-                    "Could not open {}. Make sure to have write permissions",
-                    file);
-        return false;
-    }
-    of << content;
-    return true;
-}
-}  // namespace
+#include "src/utils/cpp/expected.hpp"
+#include "src/utils/cpp/type_safe_arithmetic.hpp"
 
 auto ServeServerImpl::Create(std::optional<std::string> interface,
                              std::optional<int> port,
                              std::optional<std::string> info_file,
-                             std::optional<std::string> pid_file) noexcept
+                             std::optional<std::string> pid_file,
+                             gsl::not_null<std::mutex*> const& lock) noexcept
     -> std::optional<ServeServerImpl> {
-    ServeServerImpl server;
+    auto server = ServeServerImpl(lock);
     if (interface) {
         server.interface_ = std::move(*interface);
     }
@@ -88,7 +84,7 @@ auto ServeServerImpl::Create(std::optional<std::string> interface,
     if (pid_file) {
         server.pid_file_ = std::move(*pid_file);
     }
-    return std::move(server);
+    return server;
 }
 
 auto ServeServerImpl::Run(
@@ -125,8 +121,10 @@ auto ServeServerImpl::Run(
                      local_context,
                      remote_context,
                      &apis,
+                     lock_,
                      serve ? &*serve : nullptr};
-    ConfigurationService cs{hash_type, remote_context->exec_config};
+    ConfigurationService cs{
+        hash_type, remote_context->exec_config, &serve_config};
 
     // For the SourceTreeService we need to always have access to a native
     // storage. In compatible mode, this requires creating a second local
@@ -138,8 +136,7 @@ auto ServeServerImpl::Run(
     auto const is_compat = not ProtocolTraits::IsNative(hash_type);
     if (is_compat) {
         auto config =
-            StorageConfig::Builder{}
-                .SetBuildRoot(local_context->storage_config->build_root)
+            StorageConfig::Builder::Rebuild(*local_context->storage_config)
                 .SetHashType(HashFunction::Type::GitSHA1)
                 .Build();
         if (not config) {
@@ -165,15 +162,15 @@ auto ServeServerImpl::Run(
         is_compat ? &*local_context : nullptr,
         is_compat ? &*apis.local : nullptr);
     // setup the apis to pass to SourceTreeService
-    auto const mr_apis = ApiBundle{.hash_function = apis.hash_function,
-                                   .local = mr_local_api,
-                                   .remote = apis.remote};
+    auto const mr_apis =
+        ApiBundle{.local = mr_local_api, .remote = apis.remote};
 
     SourceTreeService sts{
         &serve_config,
         &mr_apis,
         is_compat ? &*secondary_local_context
-                  : local_context,             // native_context
+                  : local_context,  // native_context
+        lock_,
         is_compat ? &*local_context : nullptr  // compat_context
     };
 
@@ -236,7 +233,8 @@ auto ServeServerImpl::Run(
         {"interface", interface_}, {"port", port_}, {"pid", pid}};
 
     if (not pid_file_.empty()) {
-        if (not TryWrite(pid_file_, pid)) {
+        if (not FileSystemAtomic::WriteFile(pid_file_,
+                                            fmt::format("{}", pid))) {
             server->Shutdown();
             return false;
         }
@@ -251,7 +249,7 @@ auto ServeServerImpl::Run(
                 info_str);
 
     if (not info_file_.empty()) {
-        if (not TryWrite(info_file_, info_str)) {
+        if (not FileSystemAtomic::WriteFile(info_file_, info_str)) {
             server->Shutdown();
             return false;
         }
