@@ -16,23 +16,34 @@
 
 #include "src/buildtool/serve_api/serve_service/target.hpp"
 
+#include <exception>
+#include <functional>
+#include <optional>
+#include <utility>
+
 #include "fmt/core.h"
+#include "google/protobuf/repeated_ptr_field.h"
+#include "nlohmann/json.hpp"
 #include "src/buildtool/build_engine/base_maps/entity_name.hpp"
 #include "src/buildtool/build_engine/base_maps/entity_name_data.hpp"
 #include "src/buildtool/build_engine/expression/configuration.hpp"
 #include "src/buildtool/build_engine/expression/expression.hpp"
 #include "src/buildtool/build_engine/expression/expression_ptr.hpp"
+#include "src/buildtool/build_engine/expression/target_result.hpp"
 #include "src/buildtool/build_engine/target_map/configured_target.hpp"
 #include "src/buildtool/build_engine/target_map/result_map.hpp"
 #include "src/buildtool/common/artifact.hpp"
 #include "src/buildtool/common/artifact_digest.hpp"
 #include "src/buildtool/common/artifact_digest_factory.hpp"
-#include "src/buildtool/common/remote/retry_config.hpp"
+#include "src/buildtool/common/bazel_types.hpp"
+#include "src/buildtool/common/cli.hpp"
 #include "src/buildtool/common/repository_config.hpp"
 #include "src/buildtool/common/statistics.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
+#include "src/buildtool/execution_api/common/execution_api.hpp"
 #include "src/buildtool/execution_engine/executor/context.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
+#include "src/buildtool/file_system/git_repo.hpp"
 #include "src/buildtool/file_system/object_type.hpp"
 #include "src/buildtool/graph_traverser/graph_traverser.hpp"
 #include "src/buildtool/logging/log_level.hpp"
@@ -45,9 +56,54 @@
 #include "src/buildtool/progress_reporting/progress_reporter.hpp"
 #include "src/buildtool/serve_api/serve_service/target_utils.hpp"
 #include "src/buildtool/storage/backend_description.hpp"
+#include "src/buildtool/storage/config.hpp"
 #include "src/buildtool/storage/garbage_collector.hpp"
 #include "src/buildtool/storage/repository_garbage_collector.hpp"
+#include "src/buildtool/storage/storage.hpp"
+#include "src/buildtool/storage/target_cache.hpp"
+#include "src/buildtool/storage/target_cache_entry.hpp"
 #include "src/buildtool/storage/target_cache_key.hpp"
+#include "src/utils/cpp/tmp_dir.hpp"
+
+namespace {
+// Store the artifact root given by the target-cache entry. Return std::nullopt
+// on success and an error message on failure.
+auto KeepRoot(TargetCacheEntry const& target_entry,
+              LocalContext const& local_context,
+              gsl::not_null<IExecutionApi::Ptr> const& api,
+              gsl::not_null<std::mutex*> const& tagging_lock)
+    -> std::optional<std::string> {
+    auto cache_result = target_entry.ToResult();
+    if (not cache_result) {
+        return "Failed to get analysis result for target cache entry.";
+    }
+    auto tmp_dir =
+        local_context.storage_config->CreateTypedTmpDir("keep-artifacts-stage");
+    if (not tmp_dir) {
+        return "Failed to create a temporary directory for keeping stage.";
+    }
+    std::vector<std::filesystem::path> paths{};
+    std::vector<Artifact::ObjectInfo> object_infos{};
+    for (auto const& [rel_path, artifact] :
+         cache_result->artifact_stage->Map()) {
+        paths.push_back(tmp_dir->GetPath() / rel_path);
+        object_infos.push_back(*artifact->Artifact().ToArtifact().Info());
+    }
+    if (not api->RetrieveToPaths(object_infos, paths)) {
+        return fmt::format("Failed installing {} to temp dir {}.",
+                           cache_result->artifact_stage->ToString(),
+                           tmp_dir->GetPath().string());
+    }
+    auto import_result = GitRepo::ImportToGit(*local_context.storage_config,
+                                              tmp_dir->GetPath(),
+                                              "Keep artifact stage",
+                                              tagging_lock);
+    if (not import_result) {
+        return import_result.error();
+    }
+    return std::nullopt;  // No errors
+}
+}  // namespace
 
 auto TargetService::GetDispatchList(
     ArtifactDigest const& dispatch_digest) noexcept
@@ -188,9 +244,10 @@ auto TargetService::ServeTarget(
     }
 
     // get backend description
-    auto description = DescribeBackend(remote_config->remote_address,
-                                       remote_config->platform_properties,
-                                       remote_config->dispatch);
+    auto description =
+        BackendDescription::Describe(remote_config->remote_address,
+                                     remote_config->platform_properties,
+                                     remote_config->dispatch);
     if (not description) {
         auto err = fmt::format("Failed to create backend description:\n{}",
                                description.error());
@@ -198,22 +255,10 @@ auto TargetService::ServeTarget(
         return ::grpc::Status{::grpc::StatusCode::INTERNAL, err};
     }
 
-    // add backend description to CAS
-    auto execution_backend_dgst =
-        local_context_.storage->CAS().StoreBlob(*description);
-    if (not execution_backend_dgst) {
-        std::string err{
-            "Failed to store execution backend description in local CAS"};
-        logger_->Emit(LogLevel::Error, err);
-        return ::grpc::Status{::grpc::StatusCode::INTERNAL, err};
-    }
-
     // get a target cache instance with the correct computed shard
-    auto shard = remote_config->remote_address
-                     ? std::make_optional(execution_backend_dgst->hash())
-                     : std::nullopt;
-    auto const& tc = local_context_.storage->TargetCache().WithShard(shard);
-    auto const& tc_key =
+    auto const tc =
+        local_context_.storage->TargetCache().WithShard(*description);
+    auto const tc_key =
         TargetCacheKey{{*target_cache_key_digest, ObjectType::File}};
 
     // check if target-level cache entry has already been computed
@@ -238,6 +283,14 @@ auto TargetService::ServeTarget(
             logger_->Emit(LogLevel::Error, "{}", msg);
             return ::grpc::Status{::grpc::StatusCode::UNAVAILABLE, msg};
         }
+        if (request->keep_artifact_root()) {
+            auto keep = KeepRoot(
+                target_entry->first, local_context_, apis_.local, lock_);
+            if (keep) {
+                logger_->Emit(LogLevel::Error, "{}", *keep);
+                return ::grpc::Status{::grpc::StatusCode::INTERNAL, *keep};
+            }
+        }
         // populate response with the target cache value
         (*response->mutable_target_value()) =
             ArtifactDigestFactory::ToBazel(target_entry->second.digest);
@@ -326,7 +379,7 @@ auto TargetService::ServeTarget(
         return ::grpc::Status{::grpc::StatusCode::NOT_FOUND, msg};
     }
     auto const repo_key_dgst =
-        ArtifactDigestFactory::Create(apis_.hash_function.GetType(),
+        ArtifactDigestFactory::Create(apis_.local->GetHashType(),
                                       repo_key->String(),
                                       0,
                                       /*is_tree=*/false);
@@ -436,7 +489,6 @@ auto TargetService::ServeTarget(
                               error_msg};
     }
 
-    BuildMaps::Target::ResultTargetMap result_map{serve_config_.jobs};
     auto configured_target = BuildMaps::Target::ConfiguredTarget{
         .target = std::move(*entity), .config = std::move(config)};
 
@@ -463,38 +515,21 @@ auto TargetService::ServeTarget(
                                .serve = serve_};
 
     // analyse the configured target
-    auto result = AnalyseTarget(&analyse_ctx,
-                                configured_target,
-                                &result_map,
-                                serve_config_.jobs,
-                                std::nullopt /*request_action_input*/,
-                                &logger);
+    auto analyse_result = AnalyseTarget(&analyse_ctx,
+                                        configured_target,
+                                        serve_config_.jobs,
+                                        std::nullopt /*request_action_input*/,
+                                        &logger);
 
-    if (not result) {
+    if (not analyse_result) {
         // report failure locally, to keep track of it...
         auto msg = fmt::format("Failed to analyse target {}",
                                configured_target.target.ToString());
         logger_->Emit(LogLevel::Warning, "{}", msg);
         return HandleFailureLog(tmp_log, "analysis", response);
     }
-    logger_->Emit(LogLevel::Info, "Analysed target {}", result->id.ToString());
-
-    // get the output artifacts
-    auto const [artifacts, runfiles] = ReadOutputArtifacts(result->target);
-
-    // get the result map outputs
-    auto const& [actions, blobs, trees] =
-        result_map.ToResult(&stats, &progress, &logger);
-
-    // collect cache targets and artifacts for target-level caching
-    auto const cache_targets = result_map.CacheTargets();
-    auto cache_artifacts = CollectNonKnownArtifacts(cache_targets);
-
-    // Clean up result map, now that it is no longer needed
-    {
-        TaskSystem ts{serve_config_.jobs};
-        result_map.Clear(&ts);
-    }
+    logger_->Emit(
+        LogLevel::Info, "Analysed target {}", analyse_result->id.ToString());
 
     auto jobs = serve_config_.build_jobs;
     if (jobs == 0) {
@@ -531,12 +566,30 @@ auto TargetService::ServeTarget(
             ProgressReporter::Reporter(&stats, &progress, &logger),
             &logger};
 
+        // get the output artifacts
+        auto const [artifacts, runfiles] =
+            ReadOutputArtifacts(analyse_result->target);
+
+        // get the analyse_result map outputs
+        auto [actions, blobs, trees] =
+            analyse_result->result_map.ToResult(&stats, &progress, &logger);
+
+        // collect cache targets and artifacts for target-level caching
+        auto const cache_targets = analyse_result->result_map.CacheTargets();
+        auto cache_artifacts = CollectNonKnownArtifacts(cache_targets);
+
+        // Clean up analyse_result map, now that it is no longer needed
+        {
+            TaskSystem ts{serve_config_.jobs};
+            analyse_result->result_map.Clear(&ts);
+        }
+
         // perform build
         auto build_result = traverser.BuildAndStage(artifacts,
                                                     runfiles,
-                                                    actions,
-                                                    blobs,
-                                                    trees,
+                                                    std::move(actions),
+                                                    std::move(blobs),
+                                                    std::move(trees),
                                                     std::move(cache_artifacts));
 
         if (not build_result) {
@@ -588,6 +641,15 @@ auto TargetService::ServeTarget(
             logger_->Emit(LogLevel::Error, "{}", msg);
             return ::grpc::Status{::grpc::StatusCode::UNAVAILABLE, msg};
         }
+        if (request->keep_artifact_root()) {
+            auto keep_error = KeepRoot(
+                target_entry->first, local_context_, apis_.local, lock_);
+            if (keep_error) {
+                logger_->Emit(LogLevel::Error, "{}", *keep_error);
+                return ::grpc::Status{::grpc::StatusCode::INTERNAL,
+                                      *keep_error};
+            }
+        }
         // populate response with the target cache value
         (*response->mutable_target_value()) =
             ArtifactDigestFactory::ToBazel(target_entry->second.digest);
