@@ -16,20 +16,19 @@
 #define INCLUDED_SRC_BUILD_SERVE_API_SERVE_SERVICE_TARGET_HPP
 
 #include <filesystem>
-#include <map>
 #include <memory>
-#include <optional>
+#include <mutex>
 #include <string>
-#include <utility>
 #include <vector>
 
+#include <grpcpp/grpcpp.h>
+
 #include "gsl/gsl"
 #include "justbuild/just_serve/just_serve.grpc.pb.h"
-#include "nlohmann/json.hpp"
+#include "justbuild/just_serve/just_serve.pb.h"
 #include "src/buildtool/common/artifact_digest.hpp"
 #include "src/buildtool/common/remote/remote_common.hpp"
 #include "src/buildtool/execution_api/common/api_bundle.hpp"
-#include "src/buildtool/execution_api/common/execution_api.hpp"
 #include "src/buildtool/execution_api/local/context.hpp"
 #include "src/buildtool/execution_api/remote/config.hpp"
 #include "src/buildtool/execution_api/remote/context.hpp"
@@ -46,11 +45,13 @@ class TargetService final : public justbuild::just_serve::Target::Service {
         gsl::not_null<LocalContext const*> const& local_context,
         gsl::not_null<RemoteContext const*> const& remote_context,
         gsl::not_null<ApiBundle const*> const& apis,
+        gsl::not_null<std::mutex*> const& lock,
         ServeApi const* serve = nullptr) noexcept
         : serve_config_{*serve_config},
           local_context_{*local_context},
           remote_context_{*remote_context},
           apis_{*apis},
+          lock_{lock},
           serve_{serve} {}
 
     // Given a target-level caching key, returns the computed value. In doing
@@ -133,6 +134,7 @@ class TargetService final : public justbuild::just_serve::Target::Service {
     LocalContext const& local_context_;
     RemoteContext const& remote_context_;
     ApiBundle const& apis_;
+    gsl::not_null<std::mutex*> lock_;
     ServeApi const* const serve_ = nullptr;
     std::shared_ptr<Logger> logger_{std::make_shared<Logger>("target-service")};
 
