@@ -18,6 +18,7 @@
 #include <filesystem>
 #include <optional>
 #include <string>
+#include <type_traits>
 #include <unordered_set>
 #include <vector>
 
@@ -25,10 +26,10 @@
 #include "src/buildtool/common/artifact_digest.hpp"
 #include "src/buildtool/common/protocol_traits.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
-#include "src/buildtool/file_system/git_repo.hpp"
 #include "src/buildtool/file_system/object_cas.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
 #include "src/buildtool/storage/config.hpp"
-#include "src/buildtool/storage/large_object_cas.hpp"
+#include "src/buildtool/storage/large_object_cas.hpp"  // IWYU pragma: keep
 #include "src/buildtool/storage/uplinker.hpp"
 #include "src/utils/cpp/expected.hpp"
 
@@ -52,13 +53,13 @@ class LocalCAS {
     explicit LocalCAS(
         GenerationConfig const& config,
         gsl::not_null<Uplinker<kDoGlobalUplink> const*> const& uplinker)
-        : cas_file_{&config.storage_config->hash_function,
+        : cas_file_{config.storage_config->hash_function,
                     config.cas_f,
                     MakeUplinker<ObjectType::File>(config, uplinker)},
-          cas_exec_{&config.storage_config->hash_function,
+          cas_exec_{config.storage_config->hash_function,
                     config.cas_x,
                     MakeUplinker<ObjectType::Executable>(config, uplinker)},
-          cas_tree_{&config.storage_config->hash_function,
+          cas_tree_{config.storage_config->hash_function,
                     config.cas_t,
                     MakeUplinker<ObjectType::Tree>(config, uplinker)},
           cas_file_large_{this, config, uplinker},
@@ -284,7 +285,7 @@ class LocalCAS {
     ObjectCAS<ObjectType::Tree> cas_tree_;
     LargeObjectCAS<kDoGlobalUplink, ObjectType::File> cas_file_large_;
     LargeObjectCAS<kDoGlobalUplink, ObjectType::Tree> cas_tree_large_;
-    HashFunction const& hash_function_;
+    HashFunction hash_function_;
 
     /// \brief Provides uplink via "exists callback" for physical object CAS.
     template <ObjectType kType>
@@ -353,7 +354,7 @@ class LocalCAS {
 
 #ifndef BOOTSTRAP_BUILD_TOOL
 // NOLINTNEXTLINE(misc-header-include-cycle)
-#include "src/buildtool/storage/local_cas.tpp"
+#include "src/buildtool/storage/local_cas.tpp"  // IWYU pragma: export
 #else
 template <bool kDoGlobalUplink>
 auto LocalCAS<kDoGlobalUplink>::CheckTreeInvariant(
