@@ -15,9 +15,12 @@
 #ifndef INCLUDED_SRC_OTHER_TOOLS_JUST_MR_CLI_HPP
 #define INCLUDED_SRC_OTHER_TOOLS_JUST_MR_CLI_HPP
 
+#include <algorithm>
 #include <cstddef>
 #include <cstdint>
 #include <filesystem>
+#include <functional>
+#include <memory>
 #include <optional>
 #include <string>
 #include <thread>
@@ -32,7 +35,6 @@
 #include "src/buildtool/common/clidefaults.hpp"
 #include "src/buildtool/common/retry_cli.hpp"
 #include "src/buildtool/common/user_structs.hpp"
-#include "src/buildtool/execution_api/local/config.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/other_tools/just_mr/mirrors.hpp"
 #include "src/other_tools/just_mr/utils.hpp"
@@ -83,6 +85,10 @@ struct MultiRepoUpdateArguments {
     std::vector<std::string> repos_to_update;
 };
 
+struct MultiRepoGcArguments {
+    bool drop_only{false};
+};
+
 struct MultiRepoJustSubCmdsArguments {
     std::optional<std::string> subcmd_name{std::nullopt};
     std::vector<std::string> additional_just_args;
@@ -122,6 +128,7 @@ struct CommandLineArguments {
     MultiRepoSetupArguments setup;
     MultiRepoFetchArguments fetch;
     MultiRepoUpdateArguments update;
+    MultiRepoGcArguments gc;
     MultiRepoJustSubCmdsArguments just_cmd;
     MultiRepoRemoteAuthArguments auth;
     ForwardOnlyArguments launch_fwd;
@@ -336,6 +343,14 @@ static inline void SetupMultiRepoUpdateArguments(
         ->type_name("");
 }
 
+static inline void SetupMultiRepoGcArguments(
+    gsl::not_null<CLI::App*> const& app,
+    gsl::not_null<MultiRepoGcArguments*> const& clargs) {
+    app->add_flag("--drop-only",
+                  clargs->drop_only,
+                  "Only drop old repository generations");
+}
+
 static inline auto SetupMultiRepoRemoteAuthArguments(
     gsl::not_null<CLI::App*> const& app,
     gsl::not_null<MultiRepoRemoteAuthArguments*> const& authargs) {
