@@ -12,22 +12,26 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+#include <cstdlib>
 #include <exception>
 #include <filesystem>
+#include <fstream>
+#include <iostream>
+#include <map>
+#include <memory>
+#include <optional>
+#include <string>
 #include <utility>
-
-#ifdef __unix__
-#include <unistd.h>
-#else
-#error "Non-unix is not supported yet"
-#endif
+#include <vector>
 
 #include "CLI/CLI.hpp"
 #include "gsl/gsl"
 #include "nlohmann/json.hpp"
-#include "src/buildtool/build_engine/expression/configuration.hpp"
+#include "src/buildtool/common/clidefaults.hpp"
 #include "src/buildtool/common/retry_cli.hpp"
+#include "src/buildtool/common/user_structs.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
+#include "src/buildtool/file_system/file_system_manager.hpp"
 #include "src/buildtool/file_system/git_context.hpp"
 #include "src/buildtool/logging/log_config.hpp"
 #include "src/buildtool/logging/log_level.hpp"
@@ -38,14 +42,18 @@
 #include "src/buildtool/storage/config.hpp"
 #include "src/buildtool/storage/garbage_collector.hpp"
 #include "src/buildtool/storage/repository_garbage_collector.hpp"
+#include "src/buildtool/storage/storage.hpp"
 #include "src/other_tools/just_mr/cli.hpp"
 #include "src/other_tools/just_mr/exit_codes.hpp"
 #include "src/other_tools/just_mr/fetch.hpp"
 #include "src/other_tools/just_mr/launch.hpp"
+#include "src/other_tools/just_mr/mirrors.hpp"
 #include "src/other_tools/just_mr/rc.hpp"
 #include "src/other_tools/just_mr/setup.hpp"
 #include "src/other_tools/just_mr/setup_utils.hpp"
 #include "src/other_tools/just_mr/update.hpp"
+#include "src/other_tools/just_mr/utils.hpp"
+#include "src/utils/cpp/expected.hpp"
 
 namespace {
 
@@ -74,6 +82,13 @@ void SetupUpdateCommandArguments(
     SetupMultiRepoUpdateArguments(app, &clargs->update);
 }
 
+/// \brief Setup arguments for subcommand "just-mr gc-repo".
+void SetupUpdateGcArguments(
+    gsl::not_null<CLI::App*> const& app,
+    gsl::not_null<CommandLineArguments*> const& clargs) {
+    SetupMultiRepoGcArguments(app, &clargs->gc);
+}
+
 /// \brief Setup arguments for subcommand "just-mr setup" and
 /// "just-mr setup-env".
 void SetupSetupCommandArguments(
@@ -125,6 +140,7 @@ void SetupSetupCommandArguments(
     SetupSetupCommandArguments(cmd_setup_env, &clargs);
     SetupFetchCommandArguments(cmd_fetch, &clargs);
     SetupUpdateCommandArguments(cmd_update, &clargs);
+    SetupUpdateGcArguments(cmd_gc_repo, &clargs);
 
     // for 'just' calls, allow extra arguments
     cmd_do->allow_extras();
@@ -299,6 +315,9 @@ auto main(int argc, char* argv[]) -> int {
                 arguments.common.alternative_mirrors->preferred_hostnames =
                     checkout_locations_json.value("preferred hostnames",
                                                   nlohmann::json::array());
+                arguments.common.alternative_mirrors->extra_inherit_env =
+                    checkout_locations_json.value("extra inherit env",
+                                                  nlohmann::json::array());
             } catch (std::exception const& e) {
                 Logger::Log(
                     LogLevel::Error,
@@ -328,7 +347,7 @@ auto main(int argc, char* argv[]) -> int {
 
         if (arguments.cmd == SubCommand::kGcRepo) {
             return RepositoryGarbageCollector::TriggerGarbageCollection(
-                       *native_storage_config)
+                       *native_storage_config, arguments.gc.drop_only)
                        ? kExitSuccess
                        : kExitBuiltinCommandFailure;
         }
