@@ -14,38 +14,45 @@
 
 #include "src/other_tools/just_mr/setup.hpp"
 
+#include <algorithm>
 #include <atomic>
+#include <compare>
 #include <condition_variable>
 #include <cstddef>
 #include <exception>
+#include <functional>
 #include <optional>
 #include <thread>
+#include <utility>
+#include <vector>
 
+#include "gsl/gsl"
 #include "nlohmann/json.hpp"
-#include "src/buildtool/auth/authentication.hpp"
-#include "src/buildtool/common/remote/retry_config.hpp"
+#include "src/buildtool/build_engine/expression/expression.hpp"
+#include "src/buildtool/build_engine/expression/expression_ptr.hpp"
+#include "src/buildtool/common/remote/remote_common.hpp"
+#include "src/buildtool/common/user_structs.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
-#include "src/buildtool/execution_api/bazel_msg/bazel_common.hpp"
+#include "src/buildtool/execution_api/bazel_msg/execution_config.hpp"
 #include "src/buildtool/execution_api/common/api_bundle.hpp"
 #include "src/buildtool/execution_api/common/execution_api.hpp"
-#include "src/buildtool/execution_api/local/config.hpp"
 #include "src/buildtool/execution_api/local/context.hpp"
 #include "src/buildtool/execution_api/local/local_api.hpp"
 #include "src/buildtool/execution_api/remote/bazel/bazel_api.hpp"
 #include "src/buildtool/execution_api/remote/config.hpp"
 #include "src/buildtool/execution_api/remote/context.hpp"
 #include "src/buildtool/execution_api/serve/mr_local_api.hpp"
-#include "src/buildtool/file_system/symlinks_map/resolve_symlinks_map.hpp"
+#include "src/buildtool/file_system/file_system_manager.hpp"
+#include "src/buildtool/file_system/symlinks/resolve_symlinks_map.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
 #include "src/buildtool/main/retry.hpp"
 #include "src/buildtool/multithreading/async_map_utils.hpp"
 #include "src/buildtool/multithreading/task_system.hpp"
-#include "src/buildtool/serve_api/remote/config.hpp"
+#include "src/buildtool/progress_reporting/base_progress_reporter.hpp"
 #include "src/buildtool/serve_api/remote/serve_api.hpp"
 #include "src/buildtool/storage/fs_utils.hpp"
 #include "src/buildtool/storage/garbage_collector.hpp"
-#include "src/other_tools/just_mr/exit_codes.hpp"
 #include "src/other_tools/just_mr/progress_reporting/progress.hpp"
 #include "src/other_tools/just_mr/progress_reporting/progress_reporter.hpp"
 #include "src/other_tools/just_mr/progress_reporting/statistics.hpp"
@@ -54,12 +61,15 @@
 #include "src/other_tools/ops_maps/content_cas_map.hpp"
 #include "src/other_tools/ops_maps/critical_git_op_map.hpp"
 #include "src/other_tools/ops_maps/git_tree_fetch_map.hpp"
+#include "src/other_tools/ops_maps/import_to_git_map.hpp"
 #include "src/other_tools/repo_map/repos_to_setup_map.hpp"
 #include "src/other_tools/root_maps/commit_git_map.hpp"
 #include "src/other_tools/root_maps/content_git_map.hpp"
 #include "src/other_tools/root_maps/distdir_git_map.hpp"
+#include "src/other_tools/root_maps/foreign_file_git_map.hpp"
 #include "src/other_tools/root_maps/fpath_git_map.hpp"
 #include "src/other_tools/root_maps/tree_id_git_map.hpp"
+#include "src/utils/cpp/expected.hpp"
 #include "src/utils/cpp/file_locking.hpp"
 
 auto MultiRepoSetup(std::shared_ptr<Configuration> const& config,
@@ -131,7 +141,7 @@ auto MultiRepoSetup(std::shared_ptr<Configuration> const& config,
         JustMR::Utils::ReachableRepositories(repos, *main, setup_repos);
     }
     Logger::Log(LogLevel::Info,
-                "Found {} repositories to set up",
+                "Found {} repositories involved",
                 setup_repos->to_setup.size());
 
     // setup local execution config
@@ -156,8 +166,7 @@ auto MultiRepoSetup(std::shared_ptr<Configuration> const& config,
     std::optional<LockFile> compat_lock = std::nullopt;
     IExecutionApi::Ptr compat_local_api = nullptr;
     if (common_args.compatible) {
-        auto config = StorageConfig::Builder{}
-                          .SetBuildRoot(native_storage_config.build_root)
+        auto config = StorageConfig::Builder::Rebuild(native_storage_config)
                           .SetHashType(HashFunction::Type::PlainSHA256)
                           .Build();
         if (not config) {
@@ -210,12 +219,13 @@ auto MultiRepoSetup(std::shared_ptr<Configuration> const& config,
     }
 
     // create the remote api
-    auto const hash_fct =
-        compat_local_context != nullptr
-            ? compat_local_context->storage_config->hash_function
-            : native_local_context.storage_config->hash_function;
     IExecutionApi::Ptr remote_api = nullptr;
     if (auto const address = remote_exec_config->remote_address) {
+        auto const hash_fct =
+            compat_local_context != nullptr
+                ? compat_local_context->storage_config->hash_function
+                : native_local_context.storage_config->hash_function;
+
         ExecutionConfiguration config;
         config.skip_cache_lookup = false;
         remote_api = std::make_shared<BazelApi>("remote-execution",
@@ -224,7 +234,7 @@ auto MultiRepoSetup(std::shared_ptr<Configuration> const& config,
                                                 &*auth_config,
                                                 &*retry_config,
                                                 config,
-                                                &hash_fct);
+                                                hash_fct);
     }
     bool const has_remote_api = remote_api != nullptr;
 
@@ -240,8 +250,7 @@ auto MultiRepoSetup(std::shared_ptr<Configuration> const& config,
         return std::nullopt;
     }
     auto const apis =
-        ApiBundle{.hash_function = hash_fct,
-                  .local = mr_local_api,
+        ApiBundle{.local = mr_local_api,
                   .remote = has_remote_api ? remote_api : mr_local_api};
     auto serve = ServeApi::Create(
         *serve_config,
@@ -312,10 +321,10 @@ auto MultiRepoSetup(std::shared_ptr<Configuration> const& config,
         &import_to_git_map,
         common_args.git_path->string(),
         *common_args.local_launcher,
+        common_args.alternative_mirrors,
         serve ? &*serve : nullptr,
         &native_storage_config,
         compat_storage_config != nullptr ? &*compat_storage_config : nullptr,
-        compat_storage != nullptr ? &*compat_storage : nullptr,
         &(*apis.local),
         has_remote_api ? &*apis.remote : nullptr,
         false, /* backup_to_remote */
@@ -324,41 +333,32 @@ auto MultiRepoSetup(std::shared_ptr<Configuration> const& config,
 
     auto resolve_symlinks_map = CreateResolveSymlinksMap();
 
-    auto commit_git_map = CreateCommitGitMap(
-        &critical_git_op_map,
-        &import_to_git_map,
-        common_args.just_mr_paths,
-        common_args.alternative_mirrors,
-        common_args.git_path->string(),
-        *common_args.local_launcher,
-        serve ? &*serve : nullptr,
-        &native_storage_config,
-        compat_storage_config != nullptr ? &*compat_storage_config : nullptr,
-        compat_storage != nullptr ? &*compat_storage : nullptr,
-        &(*apis.local),
-        has_remote_api ? &*apis.remote : nullptr,
-        common_args.fetch_absent,
-        &progress,
-        common_args.jobs);
+    auto commit_git_map =
+        CreateCommitGitMap(&critical_git_op_map,
+                           &import_to_git_map,
+                           common_args.just_mr_paths,
+                           common_args.alternative_mirrors,
+                           common_args.git_path->string(),
+                           *common_args.local_launcher,
+                           serve ? &*serve : nullptr,
+                           &native_storage_config,
+                           &(*apis.local),
+                           has_remote_api ? &*apis.remote : nullptr,
+                           common_args.fetch_absent,
+                           &progress,
+                           common_args.jobs);
 
-    auto content_git_map = CreateContentGitMap(
-        &content_cas_map,
-        &import_to_git_map,
-        common_args.just_mr_paths,
-        common_args.alternative_mirrors,
-        common_args.ca_info,
-        &resolve_symlinks_map,
-        &critical_git_op_map,
-        serve ? &*serve : nullptr,
-        &native_storage_config,
-        compat_storage_config != nullptr ? &*compat_storage_config : nullptr,
-        &native_storage,
-        compat_storage != nullptr ? &*compat_storage : nullptr,
-        has_remote_api ? &*apis.local : nullptr,  // only needed if remote given
-        has_remote_api ? &*apis.remote : nullptr,
-        common_args.fetch_absent,
-        &progress,
-        common_args.jobs);
+    auto content_git_map = CreateContentGitMap(&content_cas_map,
+                                               &import_to_git_map,
+                                               common_args.just_mr_paths,
+                                               &resolve_symlinks_map,
+                                               &critical_git_op_map,
+                                               serve ? &*serve : nullptr,
+                                               &native_storage_config,
+                                               &native_storage,
+                                               common_args.fetch_absent,
+                                               &progress,
+                                               common_args.jobs);
 
     auto foreign_file_git_map =
         CreateForeignFileGitMap(&content_cas_map,
@@ -376,40 +376,32 @@ auto MultiRepoSetup(std::shared_ptr<Configuration> const& config,
         &resolve_symlinks_map,
         serve ? &*serve : nullptr,
         &native_storage_config,
-        compat_storage_config != nullptr ? &*compat_storage_config : nullptr,
-        compat_storage != nullptr ? &*compat_storage : nullptr,
-        has_remote_api ? &*apis.local : nullptr,  // only needed if remote given
-        has_remote_api ? &*apis.remote : nullptr,
         common_args.jobs,
         multi_repo_tool_name,
         common_args.just_path ? common_args.just_path->string()
                               : kDefaultJustPath);
 
-    auto distdir_git_map = CreateDistdirGitMap(
-        &content_cas_map,
-        &import_to_git_map,
-        &critical_git_op_map,
-        serve ? &*serve : nullptr,
-        &native_storage_config,
-        compat_storage_config != nullptr ? &*compat_storage_config : nullptr,
-        &native_storage,
-        compat_storage != nullptr ? &*compat_storage : nullptr,
-        &(*apis.local),
-        has_remote_api ? &*apis.remote : nullptr,
-        common_args.jobs);
+    auto distdir_git_map =
+        CreateDistdirGitMap(&content_cas_map,
+                            &import_to_git_map,
+                            &critical_git_op_map,
+                            serve ? &*serve : nullptr,
+                            &native_storage_config,
+                            &native_storage,
+                            &(*apis.local),
+                            has_remote_api ? &*apis.remote : nullptr,
+                            common_args.jobs);
 
-    auto tree_id_git_map = CreateTreeIdGitMap(
-        &git_tree_fetch_map,
-        &critical_git_op_map,
-        &import_to_git_map,
-        common_args.fetch_absent,
-        serve ? &*serve : nullptr,
-        &native_storage_config,
-        compat_storage_config != nullptr ? &*compat_storage_config : nullptr,
-        compat_storage != nullptr ? &*compat_storage : nullptr,
-        &(*apis.local),
-        has_remote_api ? &*apis.remote : nullptr,
-        common_args.jobs);
+    auto tree_id_git_map =
+        CreateTreeIdGitMap(&git_tree_fetch_map,
+                           &critical_git_op_map,
+                           &import_to_git_map,
+                           common_args.fetch_absent,
+                           serve ? &*serve : nullptr,
+                           &native_storage_config,
+                           &(*apis.local),
+                           has_remote_api ? &*apis.remote : nullptr,
+                           common_args.jobs);
 
     auto repos_to_setup_map = CreateReposToSetupMap(config,
                                                     main,
