@@ -15,27 +15,34 @@
 #include "src/other_tools/ops_maps/git_tree_fetch_map.hpp"
 
 #include <cstdlib>
-#include <tuple>    // std::ignore
+#include <filesystem>
+#include <iterator>
+#include <memory>
+#include <optional>
 #include <utility>  // std::move
 
 #include "fmt/core.h"
+#include "nlohmann/json.hpp"
+#include "src/buildtool/common/artifact.hpp"
 #include "src/buildtool/common/artifact_digest.hpp"
-#include "src/buildtool/common/protocol_traits.hpp"
 #include "src/buildtool/common/repository_config.hpp"
-#include "src/buildtool/execution_api/common/execution_common.hpp"
 #include "src/buildtool/execution_api/serve/mr_git_api.hpp"
-#include "src/buildtool/execution_api/serve/utils.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
+#include "src/buildtool/file_system/git_cas.hpp"
+#include "src/buildtool/file_system/git_repo.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
 #include "src/buildtool/multithreading/task_system.hpp"
+#include "src/buildtool/progress_reporting/task_tracker.hpp"
 #include "src/buildtool/system/system_command.hpp"
+#include "src/other_tools/git_operations/git_ops_types.hpp"
 #include "src/other_tools/git_operations/git_repo_remote.hpp"
+#include "src/utils/cpp/tmp_dir.hpp"
 
 namespace {
 
 void BackupToRemote(ArtifactDigest const& digest,
                     StorageConfig const& native_storage_config,
                     StorageConfig const* compat_storage_config,
-                    Storage const* compat_storage,
                     gsl::not_null<IExecutionApi const*> const& local_api,
                     IExecutionApi const& remote_api,
                     GitTreeFetchMap::LoggerPtr const& logger) {
@@ -46,25 +53,18 @@ void BackupToRemote(ArtifactDigest const& digest,
             MRGitApi{&repo,
                      &native_storage_config,
                      compat_storage_config,
-                     compat_storage,
                      compat_storage_config != nullptr ? &*local_api : nullptr};
-        if (not git_api.RetrieveToCas(
+        if (git_api.RetrieveToCas(
                 {Artifact::ObjectInfo{.digest = digest,
                                       .type = ObjectType::Tree}},
                 remote_api)) {
-            // give a warning
-            (*logger)(fmt::format(
-                          "Failed to back up tree {} from local CAS to remote",
-                          digest.hash()),
-                      /*fatal=*/false);
+            return;
         }
     }
-    else {
-        // give a warning
-        (*logger)(fmt::format("Failed to SetGitCAS at {}",
-                              native_storage_config.GitRoot().string()),
-                  /*fatal=*/false);
-    }
+    // give a warning
+    (*logger)(fmt::format("Failed to back up tree {} from local CAS to remote",
+                          digest.hash()),
+              /*fatal=*/false);
 }
 
 /// \brief Moves the root tree from local CAS to the Git cache and sets the
@@ -75,7 +75,6 @@ void MoveCASTreeToGit(
     gsl::not_null<ImportToGitMap*> const& import_to_git_map,
     gsl::not_null<StorageConfig const*> const& native_storage_config,
     StorageConfig const* compat_storage_config,
-    Storage const* compat_storage,
     gsl::not_null<IExecutionApi const*> const& local_api,
     IExecutionApi const* remote_api,
     bool backup_to_remote,
@@ -109,7 +108,6 @@ void MoveCASTreeToGit(
          tree_hash,
          native_storage_config,
          compat_storage_config,
-         compat_storage,
          local_api,
          remote_api,
          backup_to_remote,
@@ -127,7 +125,6 @@ void MoveCASTreeToGit(
                 BackupToRemote(native_digest,
                                *native_storage_config,
                                compat_storage_config,
-                               compat_storage,
                                local_api,
                                *remote_api,
                                logger);
@@ -148,7 +145,6 @@ void TagAndSetRoot(
     ArtifactDigest const& digest,
     gsl::not_null<StorageConfig const*> const& native_storage_config,
     StorageConfig const* compat_storage_config,
-    Storage const* compat_storage,
     gsl::not_null<CriticalGitOpMap*> const& critical_git_op_map,
     gsl::not_null<IExecutionApi const*> const& local_api,
     IExecutionApi const* remote_api,
@@ -171,7 +167,6 @@ void TagAndSetRoot(
          backup_to_remote,
          native_storage_config,
          compat_storage_config,
-         compat_storage,
          local_api,
          remote_api,
          logger,
@@ -187,7 +182,6 @@ void TagAndSetRoot(
                 BackupToRemote(digest,
                                *native_storage_config,
                                compat_storage_config,
-                               compat_storage,
                                local_api,
                                *remote_api,
                                logger);
@@ -209,7 +203,6 @@ void TakeTreeFromOlderGeneration(
     ArtifactDigest const& digest,
     gsl::not_null<StorageConfig const*> const& native_storage_config,
     StorageConfig const* compat_storage_config,
-    Storage const* compat_storage,
     GitCASPtr const& git_cas,
     gsl::not_null<CriticalGitOpMap*> const& critical_git_op_map,
     gsl::not_null<IExecutionApi const*> const& local_api,
@@ -240,8 +233,7 @@ void TakeTreeFromOlderGeneration(
          logger,
          source,
          native_storage_config,
-         compat_storage_config,
-         compat_storage](auto const& values) {
+         compat_storage_config](auto const& values) {
             GitOpValue op_result = *values[0];
             if (not op_result.result) {
                 (*logger)("Tree tagging failed", /*fatal=*/true);
@@ -269,7 +261,6 @@ void TakeTreeFromOlderGeneration(
             TagAndSetRoot(digest,
                           native_storage_config,
                           compat_storage_config,
-                          compat_storage,
                           critical_git_op_map,
                           local_api,
                           remote_api,
@@ -295,10 +286,10 @@ auto CreateGitTreeFetchMap(
     gsl::not_null<ImportToGitMap*> const& import_to_git_map,
     std::string const& git_bin,
     std::vector<std::string> const& launcher,
+    MirrorsPtr const& mirrors,
     ServeApi const* serve,
     gsl::not_null<StorageConfig const*> const& native_storage_config,
     StorageConfig const* compat_storage_config,
-    Storage const* compat_storage,
     gsl::not_null<IExecutionApi const*> const& local_api,
     IExecutionApi const* remote_api,
     bool backup_to_remote,
@@ -308,10 +299,10 @@ auto CreateGitTreeFetchMap(
                           import_to_git_map,
                           git_bin,
                           launcher,
+                          mirrors,
                           serve,
                           native_storage_config,
                           compat_storage_config,
-                          compat_storage,
                           local_api,
                           remote_api,
                           backup_to_remote,
@@ -339,10 +330,10 @@ auto CreateGitTreeFetchMap(
              import_to_git_map,
              git_bin,
              launcher,
+             mirrors,
              serve,
              native_storage_config,
              compat_storage_config,
-             compat_storage,
              local_api,
              remote_api,
              backup_to_remote,
@@ -389,7 +380,6 @@ auto CreateGitTreeFetchMap(
                         BackupToRemote(ArtifactDigest{key.tree_hash, 0},
                                        *native_storage_config,
                                        compat_storage_config,
-                                       compat_storage,
                                        local_api,
                                        *remote_api,
                                        logger);
@@ -419,7 +409,6 @@ auto CreateGitTreeFetchMap(
                                     ArtifactDigest{key.tree_hash, 0},
                                     native_storage_config,
                                     compat_storage_config,
-                                    compat_storage,
                                     op_result.git_cas,
                                     critical_git_op_map,
                                     local_api,
@@ -443,7 +432,6 @@ auto CreateGitTreeFetchMap(
                                      import_to_git_map,
                                      native_storage_config,
                                      compat_storage_config,
-                                     compat_storage,
                                      local_api,
                                      remote_api,
                                      backup_to_remote,
@@ -473,7 +461,6 @@ auto CreateGitTreeFetchMap(
                                          import_to_git_map,
                                          native_storage_config,
                                          compat_storage_config,
-                                         compat_storage,
                                          local_api,
                                          remote_api,
                                          false,  // tree already on remote,
@@ -498,7 +485,6 @@ auto CreateGitTreeFetchMap(
                                      import_to_git_map,
                                      native_storage_config,
                                      compat_storage_config,
-                                     compat_storage,
                                      local_api,
                                      remote_api,
                                      false,  // tree already on remote,
@@ -535,8 +521,10 @@ auto CreateGitTreeFetchMap(
                 std::copy(key.command.begin(),
                           key.command.end(),
                           std::back_inserter(cmdline));
+                auto inherit_env =
+                    MirrorsUtils::GetInheritEnv(mirrors, key.inherit_env);
                 std::map<std::string, std::string> env{key.env_vars};
-                for (auto const& k : key.inherit_env) {
+                for (auto const& k : inherit_env) {
                     const char* v = std::getenv(k.c_str());
                     if (v != nullptr) {
                         env[k] = std::string(v);
@@ -582,9 +570,9 @@ auto CreateGitTreeFetchMap(
                      key,
                      git_bin,
                      launcher,
+                     mirrors,
                      native_storage_config,
                      compat_storage_config,
-                     compat_storage,
                      local_api,
                      remote_api,
                      backup_to_remote,
@@ -668,7 +656,6 @@ auto CreateGitTreeFetchMap(
                         // define temp repo path
                         auto tmp_dir = native_storage_config->CreateTypedTmpDir(
                             "git-tree");
-                        ;
                         if (not tmp_dir) {
                             (*logger)(fmt::format("Could not create unique "
                                                   "path for target {}",
@@ -726,7 +713,6 @@ auto CreateGitTreeFetchMap(
                             [remote_api,
                              native_storage_config,
                              compat_storage_config,
-                             compat_storage,
                              local_api,
                              backup_to_remote,
                              key,
@@ -748,7 +734,6 @@ auto CreateGitTreeFetchMap(
                                         ArtifactDigest{key.tree_hash, 0},
                                         *native_storage_config,
                                         compat_storage_config,
-                                        compat_storage,
                                         local_api,
                                         *remote_api,
                                         logger);
