@@ -14,21 +14,29 @@
 
 #include "src/other_tools/repo_map/repos_to_setup_map.hpp"
 
+#include <filesystem>
+#include <type_traits>
 #include <unordered_map>
 #include <utility>  // std::move
+#include <vector>
 
 #include "fmt/core.h"
+#include "src/buildtool/build_engine/expression/expression.hpp"
+#include "src/buildtool/build_engine/expression/expression_ptr.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/crypto/hash_info.hpp"
 #include "src/buildtool/file_system/file_root.hpp"
-#include "src/buildtool/file_system/symlinks_map/pragma_special.hpp"
-#include "src/buildtool/logging/log_level.hpp"
-#include "src/buildtool/logging/logger.hpp"
+#include "src/buildtool/file_system/precomputed_root.hpp"
+#include "src/buildtool/file_system/symlinks/pragma_special.hpp"
 #include "src/buildtool/multithreading/task_system.hpp"
+#include "src/other_tools/just_mr/utils.hpp"
 #include "src/other_tools/ops_maps/content_cas_map.hpp"
 #include "src/other_tools/ops_maps/git_tree_fetch_map.hpp"
 #include "src/other_tools/utils/parse_archive.hpp"
 #include "src/other_tools/utils/parse_git_tree.hpp"
+#include "src/other_tools/utils/parse_precomputed_root.hpp"
+#include "src/utils/cpp/expected.hpp"
+#include "src/utils/cpp/path.hpp"
 
 namespace {
 
@@ -104,6 +112,13 @@ void GitCheckout(ExpressionPtr const& repo_desc,
                                             ? repo_desc_subdir->String()
                                             : "")
                       .lexically_normal();
+    if (not PathIsNonUpwards(subdir)) {
+        (*logger)(fmt::format("GitCheckout: Expected field \"subdir\" to be a "
+                              "non-upwards path, but found {}",
+                              subdir.string()),
+                  /*fatal=*/true);
+        return;
+    }
     // check optional mirrors
     auto repo_desc_mirrors = repo_desc->Get("mirrors", Expression::list_t{});
     std::vector<std::string> mirrors{};
@@ -629,6 +644,59 @@ void GitTreeCheckout(ExpressionPtr const& repo_desc,
         });
 }
 
+void PrecomputedRootCheckout(ExpressionPtr const& repo_desc,
+                             ExpressionPtr&& repos,
+                             std::string const& repo_name,
+                             ReposToSetupMap::SetterPtr const& setter,
+                             ReposToSetupMap::SubCallerPtr const& subcaller,
+                             ReposToSetupMap::LoggerPtr const& logger) {
+    auto precomputed = ParsePrecomputedRoot(repo_desc);
+    if (not precomputed) {
+        (*logger)(fmt::format("Checkout of precomputed root {} failed:\n{}",
+                              nlohmann::json(repo_name).dump(),
+                              std::move(precomputed).error()),
+                  /*fatal=*/true);
+        return;
+    }
+
+    std::string target_repo = precomputed->GetReferencedRepository();
+    (*subcaller)(
+        {std::move(target_repo)},
+        [setter,
+         repos = std::move(repos),
+         repo_name,
+         result = *std::move(precomputed)](auto const& /*unused*/) {
+            nlohmann::json cfg{};
+            auto& ws_root = cfg["workspace_root"];
+            SetReposTakeOver(&cfg, repos, repo_name);
+
+            bool absent = false;
+            if (auto computed = result.AsComputed()) {
+                ws_root.push_back(ComputedRoot::kMarker);
+                ws_root.push_back(computed->repository);
+                ws_root.push_back(computed->target_module);
+                ws_root.push_back(computed->target_name);
+                ws_root.push_back(computed->config);
+                absent = computed->absent;
+            }
+            else if (auto tree_structure = result.AsTreeStructure()) {
+                ws_root.push_back(TreeStructureRoot::kMarker);
+                ws_root.push_back(tree_structure->repository);
+
+                absent = tree_structure->absent;
+            }
+
+            if (absent) {
+                auto pragma = nlohmann::json::object();
+                pragma["absent"] = true;
+                ws_root.push_back(pragma);
+            }
+
+            std::invoke(*setter, std::move(cfg));
+        },
+        logger);
+}
+
 }  // namespace
 
 auto CreateReposToSetupMap(
@@ -657,7 +725,7 @@ auto CreateReposToSetupMap(
                        stats](auto ts,
                               auto setter,
                               auto logger,
-                              auto /* unused */,
+                              auto subcaller,
                               auto const& key) {
         auto repos = (*config)["repositories"];
         if (main and (key == *main) and interactive) {
@@ -814,6 +882,16 @@ auto CreateReposToSetupMap(
                                     wrapped_logger);
                     break;
                 }
+                case CheckoutType::Precomputed: {
+                    stats->IncrementComputedCounter();
+                    PrecomputedRootCheckout(*resolved_repo_desc,
+                                            std::move(repos),
+                                            key,
+                                            setter,
+                                            subcaller,
+                                            wrapped_logger);
+                    break;
+                }
             }
         }
     };
