@@ -14,16 +14,21 @@
 
 #include "src/other_tools/root_maps/fpath_git_map.hpp"
 
+#include <memory>
 #include <utility>  // std::move
+#include <vector>
 
 #include "fmt/core.h"
-#include "src/buildtool/execution_api/local/config.hpp"
+#include "src/buildtool/common/artifact_digest_factory.hpp"
+#include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/file_system/file_root.hpp"
-#include "src/buildtool/file_system/git_repo.hpp"
+#include "src/buildtool/file_system/file_system_manager.hpp"
+#include "src/buildtool/file_system/git_cas.hpp"
 #include "src/buildtool/multithreading/task_system.hpp"
 #include "src/buildtool/storage/fs_utils.hpp"
+#include "src/other_tools/git_operations/git_ops_types.hpp"
 #include "src/other_tools/git_operations/git_repo_remote.hpp"
-#include "src/other_tools/root_maps/root_utils.hpp"
+#include "src/utils/cpp/expected.hpp"
 #include "src/utils/cpp/tmp_dir.hpp"
 
 namespace {
@@ -31,51 +36,42 @@ namespace {
 /// \brief Does the serve endpoint checks and sets the workspace root.
 /// It guarantees the logger is called exactly once with fatal on failure, and
 /// the setter on success.
-void CheckServeAndSetRoot(
-    std::string const& tree_id,
-    std::string const& repo_root,
-    bool absent,
-    ServeApi const* serve,
-    gsl::not_null<StorageConfig const*> const& native_storage_config,
-    StorageConfig const* compat_storage_config,
-    Storage const* compat_storage,
-    IExecutionApi const* local_api,
-    IExecutionApi const* remote_api,
-    FilePathGitMap::SetterPtr const& ws_setter,
-    FilePathGitMap::LoggerPtr const& logger) {
+void CheckServeAndSetRoot(std::string const& tree_id,
+                          std::string const& repo_root,
+                          bool absent,
+                          ServeApi const* serve,
+                          FilePathGitMap::SetterPtr const& ws_setter,
+                          FilePathGitMap::LoggerPtr const& logger) {
     // if serve endpoint is given, try to ensure it has this tree available to
     // be able to build against it. If root is not absent, do not fail if we
     // don't have a suitable remote endpoint, but warn user nonetheless.
     if (serve != nullptr) {
-        auto has_tree = CheckServeHasAbsentRoot(*serve, tree_id, logger);
+        auto const has_tree = serve->CheckRootTree(tree_id);
         if (not has_tree) {
-            return;  // fatal
+            (*logger)(fmt::format("Checking that the serve endpoint knows tree "
+                                  "{} failed.",
+                                  tree_id),
+                      /*fatal=*/true);
+            return;
         }
         if (not *has_tree) {
-            // only enforce root setup on the serve endpoint if root is absent
-            if (remote_api == nullptr) {
-                (*logger)(
-                    fmt::format("Missing or incompatible remote-execution "
-                                "endpoint needed to sync workspace root {} "
-                                "with the serve endpoint.",
-                                tree_id),
-                    /*fatal=*/absent);
-                if (absent) {
-                    return;
-                }
+            auto digest =
+                ArtifactDigestFactory::Create(HashFunction::Type::GitSHA1,
+                                              tree_id,
+                                              /*size_unknown=*/0,
+                                              /*is_tree=*/true);
+            if (not digest) {
+                (*logger)(std::move(digest).error(), /*fatal=*/true);
+                return;
             }
-            else {
-                if (not EnsureAbsentRootOnServe(*serve,
-                                                tree_id,
-                                                repo_root,
-                                                native_storage_config,
-                                                compat_storage_config,
-                                                compat_storage,
-                                                local_api,
-                                                remote_api,
-                                                logger,
-                                                /*no_sync_is_fatal=*/absent)) {
-                    return;  // fatal
+
+            auto uploaded = serve->UploadTree(*digest, repo_root);
+            // only enforce root setup on the serve endpoint if root is absent
+            if (not uploaded.has_value()) {
+                bool const fatal = absent or not uploaded.error().IsSyncError();
+                (*logger)(std::move(uploaded).error().Message(), fatal);
+                if (fatal) {
+                    return;
                 }
             }
         }
@@ -109,10 +105,6 @@ void ResolveFilePathTree(
     gsl::not_null<ResolveSymlinksMap*> const& resolve_symlinks_map,
     ServeApi const* serve,
     gsl::not_null<StorageConfig const*> const& native_storage_config,
-    StorageConfig const* compat_storage_config,
-    Storage const* compat_storage,
-    IExecutionApi const* local_api,
-    IExecutionApi const* remote_api,
     gsl::not_null<TaskSystem*> const& ts,
     FilePathGitMap::SetterPtr const& ws_setter,
     FilePathGitMap::LoggerPtr const& logger) {
@@ -137,11 +129,6 @@ void ResolveFilePathTree(
                                  native_storage_config->GitRoot().string(),
                                  absent,
                                  serve,
-                                 native_storage_config,
-                                 compat_storage_config,
-                                 compat_storage,
-                                 local_api,
-                                 remote_api,
                                  ws_setter,
                                  logger);
         }
@@ -155,17 +142,12 @@ void ResolveFilePathTree(
                                     /*known_info=*/std::nullopt,
                                     source_cas,
                                     target_cas)},
-                [resolve_symlinks_map,
-                 critical_git_op_map,
+                [critical_git_op_map,
                  tree_hash,
                  tree_id_file,
                  absent,
                  serve,
                  native_storage_config,
-                 compat_storage_config,
-                 compat_storage,
-                 local_api,
-                 remote_api,
                  ts,
                  ws_setter,
                  logger](auto const& hashes) {
@@ -188,10 +170,6 @@ void ResolveFilePathTree(
                          absent,
                          serve,
                          native_storage_config,
-                         compat_storage_config,
-                         compat_storage,
-                         local_api,
-                         remote_api,
                          ws_setter,
                          logger](auto const& values) {
                             GitOpValue op_result = *values[0];
@@ -219,11 +197,6 @@ void ResolveFilePathTree(
                                 native_storage_config->GitRoot().string(),
                                 absent,
                                 serve,
-                                native_storage_config,
-                                compat_storage_config,
-                                compat_storage,
-                                local_api,
-                                remote_api,
                                 ws_setter,
                                 logger);
                         },
@@ -251,17 +224,8 @@ void ResolveFilePathTree(
         // tree needs no further processing;
         // if serve endpoint is given, try to ensure it has this tree available
         // to be able to build against it
-        CheckServeAndSetRoot(tree_hash,
-                             repo_root,
-                             absent,
-                             serve,
-                             native_storage_config,
-                             compat_storage_config,
-                             compat_storage,
-                             local_api,
-                             remote_api,
-                             ws_setter,
-                             logger);
+        CheckServeAndSetRoot(
+            tree_hash, repo_root, absent, serve, ws_setter, logger);
     }
 }
 
@@ -274,10 +238,6 @@ auto CreateFilePathGitMap(
     gsl::not_null<ResolveSymlinksMap*> const& resolve_symlinks_map,
     ServeApi const* serve,
     gsl::not_null<StorageConfig const*> const& native_storage_config,
-    StorageConfig const* compat_storage_config,
-    Storage const* compat_storage,
-    IExecutionApi const* local_api,
-    IExecutionApi const* remote_api,
     std::size_t jobs,
     std::string const& multi_repo_tool_name,
     std::string const& build_tool_name) -> FilePathGitMap {
@@ -287,10 +247,6 @@ auto CreateFilePathGitMap(
                        resolve_symlinks_map,
                        serve,
                        native_storage_config,
-                       compat_storage_config,
-                       compat_storage,
-                       local_api,
-                       remote_api,
                        multi_repo_tool_name,
                        build_tool_name](auto ts,
                                         auto setter,
@@ -329,10 +285,6 @@ auto CreateFilePathGitMap(
                  resolve_symlinks_map,
                  serve,
                  native_storage_config,
-                 compat_storage_config,
-                 compat_storage,
-                 local_api,
-                 remote_api,
                  ts,
                  setter,
                  logger](auto const& values) {
@@ -393,10 +345,6 @@ auto CreateFilePathGitMap(
                          resolve_symlinks_map,
                          serve,
                          native_storage_config,
-                         compat_storage_config,
-                         compat_storage,
-                         local_api,
-                         remote_api,
                          ts,
                          setter,
                          logger](auto const& values) {
@@ -419,10 +367,6 @@ auto CreateFilePathGitMap(
                                 resolve_symlinks_map,
                                 serve,
                                 native_storage_config,
-                                compat_storage_config,
-                                compat_storage,
-                                local_api,
-                                remote_api,
                                 ts,
                                 setter,
                                 logger);
@@ -491,10 +435,6 @@ auto CreateFilePathGitMap(
                  resolve_symlinks_map,
                  serve,
                  native_storage_config,
-                 compat_storage_config,
-                 compat_storage,
-                 local_api,
-                 remote_api,
                  ts,
                  setter,
                  logger](auto const& values) {
@@ -520,10 +460,6 @@ auto CreateFilePathGitMap(
                         resolve_symlinks_map,
                         serve,
                         native_storage_config,
-                        compat_storage_config,
-                        compat_storage,
-                        local_api,
-                        remote_api,
                         ts,
                         setter,
                         logger);
