@@ -14,13 +14,22 @@
 
 #include "src/other_tools/root_maps/tree_id_git_map.hpp"
 
+#include <filesystem>
+#include <memory>
+#include <optional>
+#include <string>
+#include <vector>
+
 #include "fmt/core.h"
+#include "src/buildtool/common/artifact.hpp"
 #include "src/buildtool/common/artifact_digest.hpp"
-#include "src/buildtool/common/repository_config.hpp"
-#include "src/buildtool/crypto/hash_info.hpp"
-#include "src/buildtool/execution_api/serve/mr_git_api.hpp"
 #include "src/buildtool/file_system/file_root.hpp"
-#include "src/other_tools/root_maps/root_utils.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
+#include "src/buildtool/multithreading/task_system.hpp"
+#include "src/other_tools/git_operations/git_ops_types.hpp"
+#include "src/other_tools/git_operations/git_repo_remote.hpp"
+#include "src/utils/cpp/expected.hpp"
+#include "src/utils/cpp/tmp_dir.hpp"
 
 namespace {
 
@@ -29,61 +38,21 @@ namespace {
 void UploadToServeAndSetRoot(
     ServeApi const& serve,
     gsl::not_null<StorageConfig const*> const& native_storage_config,
-    StorageConfig const* compat_storage_config,
-    Storage const* compat_storage,
-    std::string const& tree_id,
     ArtifactDigest const& digest,
-    gsl::not_null<IExecutionApi const*> const& local_api,
-    IExecutionApi const& remote_api,
     bool ignore_special,
     TreeIdGitMap::SetterPtr const& setter,
     TreeIdGitMap::LoggerPtr const& logger) {
-    // upload to remote CAS
-    auto repo_config = RepositoryConfig{};
-    if (repo_config.SetGitCAS(native_storage_config->GitRoot())) {
-        auto git_api =
-            MRGitApi{&repo_config,
-                     native_storage_config,
-                     compat_storage_config,
-                     compat_storage,
-                     compat_storage_config != nullptr ? &*local_api : nullptr};
-        if (not git_api.RetrieveToCas(
-                {Artifact::ObjectInfo{.digest = digest,
-                                      .type = ObjectType::Tree}},
-                remote_api)) {
-            (*logger)(fmt::format("Failed to sync tree {} from local Git cache "
-                                  "to remote CAS",
-                                  tree_id),
-                      /*fatal=*/true);
-            return;
-        }
-    }
-    else {
-        (*logger)(fmt::format("Failed to SetGitCAS at {}",
-                              native_storage_config->GitRoot().string()),
-                  /*fatal=*/true);
-        return;
-    }
-    // tell serve to set up the root from the remote CAS tree;
-    // upload can be skipped
-    if (EnsureAbsentRootOnServe(serve,
-                                tree_id,
-                                /*repo_path=*/"",
-                                native_storage_config,
-                                /*compat_storage_config=*/nullptr,
-                                /*compat_storage=*/nullptr,
-                                /*local_api=*/nullptr,
-                                /*remote_api=*/nullptr,
-                                logger,
-                                /*no_sync_is_fatal=*/true)) {
-        // set workspace root as absent
-        auto root = nlohmann::json::array(
-            {ignore_special ? FileRoot::kGitTreeIgnoreSpecialMarker
-                            : FileRoot::kGitTreeMarker,
-             tree_id});
-        (*setter)(std::pair(std::move(root), /*is_cache_hit=*/false));
+    auto uploaded = serve.UploadTree(digest, native_storage_config->GitRoot());
+    if (not uploaded.has_value()) {
+        (*logger)(uploaded.error().Message(), /*fatal=*/true);
         return;
     }
+    // set workspace root as absent
+    auto root = nlohmann::json::array(
+        {ignore_special ? FileRoot::kGitTreeIgnoreSpecialMarker
+                        : FileRoot::kGitTreeMarker,
+         digest.hash()});
+    (*setter)(std::pair(std::move(root), /*is_cache_hit=*/false));
 }
 
 /// \brief Guarantees it terminates by either calling the setter or calling the
@@ -91,13 +60,9 @@ void UploadToServeAndSetRoot(
 void MoveCASTreeToGitAndProcess(
     ServeApi const& serve,
     gsl::not_null<StorageConfig const*> const& native_storage_config,
-    StorageConfig const* compat_storage_config,
-    Storage const* compat_storage,
-    std::string const& tree_id,
     ArtifactDigest const& digest,
     gsl::not_null<ImportToGitMap*> const& import_to_git_map,
     gsl::not_null<IExecutionApi const*> const& local_api,
-    gsl::not_null<IExecutionApi const*> const& remote_api,
     bool ignore_special,
     gsl::not_null<TaskSystem*> const& ts,
     TreeIdGitMap::SetterPtr const& setter,
@@ -116,24 +81,19 @@ void MoveCASTreeToGitAndProcess(
             {Artifact::ObjectInfo{.digest = digest, .type = ObjectType::Tree}},
             {tmp_dir->GetPath()})) {
         (*logger)(fmt::format("Failed to copy git-tree {} to {}",
-                              tree_id,
+                              digest.hash(),
                               tmp_dir->GetPath().string()),
                   true);
         return;
     }
-    CommitInfo c_info{tmp_dir->GetPath(), "tree", tree_id};
+    CommitInfo c_info{tmp_dir->GetPath(), "tree", digest.hash()};
     import_to_git_map->ConsumeAfterKeysReady(
         ts,
         {std::move(c_info)},
         [&serve,
          native_storage_config,
-         compat_storage_config,
-         compat_storage,
          tmp_dir,  // keep tmp_dir alive
-         tree_id,
          digest,
-         local_api,
-         remote_api,
          ignore_special,
          setter,
          logger](auto const& values) {
@@ -146,17 +106,13 @@ void MoveCASTreeToGitAndProcess(
             // the root from the remote CAS tree; set root as absent on success
             UploadToServeAndSetRoot(serve,
                                     native_storage_config,
-                                    compat_storage_config,
-                                    compat_storage,
-                                    tree_id,
                                     digest,
-                                    local_api,
-                                    *remote_api,
                                     ignore_special,
                                     setter,
                                     logger);
         },
-        [logger, tmp_dir, tree_id](auto const& msg, bool fatal) {
+        [logger, tmp_dir, tree_id = digest.hash()](auto const& msg,
+                                                   bool fatal) {
             (*logger)(fmt::format(
                           "While moving git-tree {} from {} to local git:\n{}",
                           tree_id,
@@ -175,8 +131,6 @@ auto CreateTreeIdGitMap(
     bool fetch_absent,
     ServeApi const* serve,
     gsl::not_null<StorageConfig const*> const& native_storage_config,
-    StorageConfig const* compat_storage_config,
-    Storage const* compat_storage,
     gsl::not_null<IExecutionApi const*> const& local_api,
     IExecutionApi const* remote_api,
     std::size_t jobs) -> TreeIdGitMap {
@@ -186,8 +140,6 @@ auto CreateTreeIdGitMap(
                         fetch_absent,
                         serve,
                         native_storage_config,
-                        compat_storage_config,
-                        compat_storage,
                         local_api,
                         remote_api](auto ts,
                                     auto setter,
@@ -201,9 +153,14 @@ auto CreateTreeIdGitMap(
         if (key.absent and not fetch_absent) {
             if (serve != nullptr) {
                 // check serve endpoint
-                auto has_tree = CheckServeHasAbsentRoot(
-                    *serve, key.tree_info.tree_hash.Hash(), logger);
+                auto const has_tree =
+                    serve->CheckRootTree(key.tree_info.tree_hash.Hash());
                 if (not has_tree) {
+                    (*logger)(fmt::format(
+                                  "Checking that the serve endpoint knows tree "
+                                  "{} failed.",
+                                  key.tree_info.tree_hash.Hash()),
+                              /*fatal=*/true);
                     return;
                 }
                 if (*has_tree) {
@@ -228,20 +185,9 @@ auto CreateTreeIdGitMap(
                 }
                 // check if tree in already in remote CAS
                 auto const digest = ArtifactDigest{key.tree_info.tree_hash, 0};
-                if (remote_api->IsAvailable({digest})) {
+                if (remote_api->IsAvailable(digest)) {
                     // tell serve to set up the root from the remote CAS tree;
-                    // upload can be skipped
-                    if (EnsureAbsentRootOnServe(
-                            *serve,
-                            key.tree_info.tree_hash.Hash(),
-                            /*repo_path=*/"",
-                            native_storage_config,
-                            /*compat_storage_config=*/nullptr,
-                            /*compat_storage=*/nullptr,
-                            /*local_api=*/nullptr,
-                            /*remote_api=*/nullptr,
-                            logger,
-                            /*no_sync_is_fatal=*/true)) {
+                    if (serve->GetTreeFromRemote(digest)) {
                         // set workspace root as absent
                         auto root = nlohmann::json::array(
                             {key.ignore_special
@@ -276,12 +222,9 @@ auto CreateTreeIdGitMap(
                     {std::move(op_key)},
                     [serve,
                      native_storage_config,
-                     compat_storage_config,
-                     compat_storage,
                      digest,
                      import_to_git_map,
                      local_api,
-                     remote_api,
                      key,
                      ts,
                      setter,
@@ -326,18 +269,12 @@ auto CreateTreeIdGitMap(
                             // upload tree from Git cache to remote CAS and tell
                             // serve to set up the root from the remote CAS
                             // tree, then set root as absent
-                            UploadToServeAndSetRoot(
-                                *serve,
-                                native_storage_config,
-                                compat_storage_config,
-                                compat_storage,
-                                key.tree_info.tree_hash.Hash(),
-                                digest,
-                                local_api,
-                                *remote_api,
-                                key.ignore_special,
-                                setter,
-                                logger);
+                            UploadToServeAndSetRoot(*serve,
+                                                    native_storage_config,
+                                                    digest,
+                                                    key.ignore_special,
+                                                    setter,
+                                                    logger);
                             // done!
                             return;
                         }
@@ -345,20 +282,15 @@ auto CreateTreeIdGitMap(
                         if (local_api->IsAvailable(digest)) {
                             // Move tree locally from CAS to Git cache, then
                             // continue processing it by UploadToServeAndSetRoot
-                            MoveCASTreeToGitAndProcess(
-                                *serve,
-                                native_storage_config,
-                                compat_storage_config,
-                                compat_storage,
-                                key.tree_info.tree_hash.Hash(),
-                                digest,
-                                import_to_git_map,
-                                local_api,
-                                remote_api,
-                                key.ignore_special,
-                                ts,
-                                setter,
-                                logger);
+                            MoveCASTreeToGitAndProcess(*serve,
+                                                       native_storage_config,
+                                                       digest,
+                                                       import_to_git_map,
+                                                       local_api,
+                                                       key.ignore_special,
+                                                       ts,
+                                                       setter,
+                                                       logger);
                             // done!
                             return;
                         }
