@@ -14,10 +14,19 @@
 
 #include "src/other_tools/utils/parse_archive.hpp"
 
+#include <filesystem>
+#include <functional>
+#include <memory>
+#include <unordered_map>
 #include <utility>  // std::move
+#include <vector>
 
 #include "fmt/core.h"
+#include "src/buildtool/build_engine/expression/expression.hpp"
+#include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/crypto/hash_info.hpp"
+#include "src/buildtool/file_system/symlinks/pragma_special.hpp"
+#include "src/utils/cpp/path.hpp"
 
 auto ParseArchiveContent(ExpressionPtr const& repo_desc,
                          std::string const& origin)
@@ -121,6 +130,13 @@ auto ParseArchiveDescription(ExpressionPtr const& repo_desc,
                                             ? repo_desc_subdir->String()
                                             : "")
                       .lexically_normal();
+    if (not PathIsNonUpwards(subdir)) {
+        (*logger)(fmt::format("ArchiveCheckout: Expected field \"subdir\" to "
+                              "be a non-upwards path, but found {}",
+                              subdir.string()),
+                  /*fatal=*/true);
+        return std::nullopt;
+    }
 
     // check "special" pragma
     auto repo_desc_pragma = repo_desc->At("pragma");
