@@ -17,14 +17,16 @@
 
 #include <atomic>
 #include <condition_variable>
+#include <memory>
+#include <mutex>  //std::unique_lock
 #include <shared_mutex>
-#include <type_traits>
-#include <utility>  // std::move
+#include <type_traits>  // IWYU pragma: keep
+#include <utility>      // IWYU pragma: keep
 
 // Atomic wrapper with notify/wait capabilities.
 // TODO(modernize): Replace any use this class by C++20's std::atomic<T>, once
 // libcxx adds support for notify_*() and wait().
-// [https://libcxx.llvm.org/docs/Cxx2aStatus.html]
+// [https://libcxx.llvm.org/Status/Cxx20.html]
 template <class T>
 class atomic {  // NOLINT(readability-identifier-naming)
   public:
@@ -95,7 +97,7 @@ class atomic {  // NOLINT(readability-identifier-naming)
 // Atomic shared_pointer with notify/wait capabilities.
 // TODO(modernize): Replace any use this class by C++20's
 // std::atomic<std::shared_ptr<T>>, once libcxx adds support for it.
-// [https://libcxx.llvm.org/docs/Cxx2aStatus.html]
+// [https://libcxx.llvm.org/Status/Cxx20.html]
 template <class T>
 class atomic_shared_ptr {  // NOLINT(readability-identifier-naming)
     using ptr_t = std::shared_ptr<T>;
