@@ -14,6 +14,9 @@
 
 #include "src/utils/cpp/tmp_dir.hpp"
 
+#include <cstdlib>
+#include <vector>
+
 #include "src/buildtool/file_system/file_system_manager.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
