@@ -15,16 +15,20 @@
 #include "src/buildtool/build_engine/expression/expression.hpp"
 
 #include <filesystem>
+#include <functional>
 #include <sstream>
 #include <string>
 #include <vector>
 
 #include "catch2/catch_test_macros.hpp"
 #include "catch2/matchers/catch_matchers_all.hpp"
+#include "nlohmann/json.hpp"
 #include "src/buildtool/build_engine/expression/configuration.hpp"
+#include "src/buildtool/build_engine/expression/expression_ptr.hpp"
 #include "src/buildtool/build_engine/expression/function_map.hpp"
+#include "src/buildtool/build_engine/expression/linked_map.hpp"
+#include "src/buildtool/build_engine/expression/target_result.hpp"
 #include "src/buildtool/common/artifact_description.hpp"
-#include "test/utils/container_matchers.hpp"
 
 TEST_CASE("Expression access", "[expression]") {  // NOLINT
     using namespace std::string_literals;
@@ -1832,7 +1836,6 @@ TEST_CASE("Expression hash computation", "[expression]") {
     using namespace std::string_literals;
     using path = std::filesystem::path;
     using number_t = Expression::number_t;
-    using artifact_t = Expression::artifact_t;
     using result_t = Expression::result_t;
     using list_t = Expression::list_t;
     using map_t = Expression::map_t;
@@ -1870,7 +1873,11 @@ TEST_CASE("Expression hash computation", "[expression]") {
 
     CHECK_FALSE(result->ToHash().empty());
     CHECK(result->ToHash() == Expression{result_t{}}.ToHash());
-    CHECK_FALSE(result->ToHash() == Expression{result_t{boolean}}.ToHash());
+    CHECK_FALSE(
+        result->ToHash() ==
+        Expression{
+            result_t{.artifact_stage = boolean, .provides = {}, .runfiles = {}}}
+            .ToHash());
 
     CHECK_FALSE(list->ToHash().empty());
     CHECK(list->ToHash() == Expression{list_t{}}.ToHash());
