@@ -14,12 +14,15 @@
 
 #include "src/buildtool/execution_api/remote/bazel/bazel_api.hpp"
 
-#include <cstdlib>
-#include <string>
+#include <optional>
 
 #include "catch2/catch_test_macros.hpp"
+#include "gsl/gsl"
+#include "src/buildtool/auth/authentication.hpp"
+#include "src/buildtool/common/remote/remote_common.hpp"
 #include "src/buildtool/common/remote/retry_config.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
+#include "src/buildtool/execution_api/common/execution_api.hpp"
 #include "src/buildtool/execution_api/remote/config.hpp"
 #include "test/buildtool/execution_api/common/api_test.hpp"
 #include "test/utils/hermeticity/test_hash_function_type.hpp"
@@ -46,7 +49,7 @@ class FactoryApi final {
                                                &auth_,
                                                &retry_config,
                                                {},
-                                               &hash_function_}};
+                                               hash_function_}};
     }
 
   private:
