@@ -15,20 +15,28 @@
 #include "src/buildtool/execution_api/remote/bazel/bazel_network.hpp"
 
 #include <cstddef>
+#include <functional>
+#include <memory>
 #include <optional>
 #include <string>
+#include <unordered_set>
 #include <vector>
 
+#include <grpc/grpc.h>
+
 #include "catch2/catch_test_macros.hpp"
-#include "src/buildtool/auth/authentication.hpp"
-#include "src/buildtool/common/bazel_digest_factory.hpp"
+#include "gsl/gsl"
+#include "src/buildtool/common/artifact_blob.hpp"
+#include "src/buildtool/common/artifact_digest.hpp"
 #include "src/buildtool/common/protocol_traits.hpp"
+#include "src/buildtool/common/remote/remote_common.hpp"
 #include "src/buildtool/common/remote/retry_config.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
-#include "src/buildtool/execution_api/bazel_msg/bazel_blob_container.hpp"
-#include "src/buildtool/execution_api/remote/bazel/bazel_execution_client.hpp"
+#include "src/buildtool/crypto/hash_info.hpp"
+#include "src/buildtool/execution_api/remote/bazel/bazel_network_reader.hpp"
 #include "src/buildtool/execution_api/remote/config.hpp"
 #include "src/buildtool/file_system/object_type.hpp"
+#include "src/utils/cpp/expected.hpp"
 #include "test/utils/hermeticity/test_hash_function_type.hpp"
 #include "test/utils/remote_execution/test_auth_config.hpp"
 #include "test/utils/remote_execution/test_remote_config.hpp"
@@ -55,44 +63,47 @@ TEST_CASE("Bazel network: write/read blobs", "[execution_api]") {
                                 &*auth_config,
                                 &retry_config,
                                 {},
-                                &hash_function};
+                                hash_function};
 
     std::string content_foo("foo");
     std::string content_bar("bar");
     std::string content_baz(kLargeSize, 'x');  // single larger blob
 
-    BazelBlob foo{BazelDigestFactory::HashDataAs<ObjectType::File>(
-                      hash_function, content_foo),
-                  content_foo,
-                  /*is_exec=*/false};
-    BazelBlob bar{BazelDigestFactory::HashDataAs<ObjectType::File>(
-                      hash_function, content_bar),
-                  content_bar,
-                  /*is_exec=*/false};
-    BazelBlob baz{BazelDigestFactory::HashDataAs<ObjectType::File>(
-                      hash_function, content_baz),
-                  content_baz,
-                  /*is_exec=*/false};
+    auto const foo =
+        ArtifactBlob::FromMemory(hash_function, ObjectType::File, content_foo);
+    REQUIRE(foo.has_value());
+    auto const bar =
+        ArtifactBlob::FromMemory(hash_function, ObjectType::File, content_bar);
+    REQUIRE(bar.has_value());
+    auto const baz =
+        ArtifactBlob::FromMemory(hash_function, ObjectType::File, content_baz);
+    REQUIRE(baz.has_value());
 
     // Search blobs via digest
-    REQUIRE(network.UploadBlobs(BazelBlobContainer{{foo, bar, baz}}));
+    REQUIRE(network.UploadBlobs({*foo, *bar, *baz}));
 
     // Read blobs in order
     auto reader = network.CreateReader();
-    std::vector<bazel_re::Digest> to_read{
-        foo.digest, bar.digest, baz.digest, bar.digest, foo.digest};
+    std::vector<ArtifactDigest> to_read{foo->GetDigest(),
+                                        bar->GetDigest(),
+                                        baz->GetDigest(),
+                                        bar->GetDigest(),
+                                        foo->GetDigest()};
     std::vector<ArtifactBlob> blobs{};
-    for (auto next : reader.ReadIncrementally(to_read)) {
+    for (auto next : reader.ReadIncrementally(&to_read)) {
         blobs.insert(blobs.end(), next.begin(), next.end());
     }
 
     // Check order maintained
     REQUIRE(blobs.size() == 5);
-    CHECK(*blobs[0].data == content_foo);
-    CHECK(*blobs[1].data == content_bar);
-    CHECK(*blobs[2].data == content_baz);
-    CHECK(*blobs[3].data == content_bar);
-    CHECK(*blobs[4].data == content_foo);
+    for (auto const& blob : blobs) {
+        REQUIRE(blob.ReadContent() != nullptr);
+    }
+    CHECK(*blobs[0].ReadContent() == content_foo);
+    CHECK(*blobs[1].ReadContent() == content_bar);
+    CHECK(*blobs[2].ReadContent() == content_baz);
+    CHECK(*blobs[3].ReadContent() == content_bar);
+    CHECK(*blobs[4].ReadContent() == content_foo);
 }
 
 TEST_CASE("Bazel network: read blobs with unknown size", "[execution_api]") {
@@ -119,37 +130,43 @@ TEST_CASE("Bazel network: read blobs with unknown size", "[execution_api]") {
                                 &*auth_config,
                                 &retry_config,
                                 {},
-                                &hash_function};
+                                hash_function};
 
     std::string content_foo("foo");
     std::string content_bar(kLargeSize, 'x');  // single larger blob
 
-    BazelBlob foo{BazelDigestFactory::HashDataAs<ObjectType::File>(
-                      hash_function, content_foo),
-                  content_foo,
-                  /*is_exec=*/false};
-    BazelBlob bar{BazelDigestFactory::HashDataAs<ObjectType::File>(
-                      hash_function, content_bar),
-                  content_bar,
-                  /*is_exec=*/false};
+    // Create and upload blobs:
+    {
+        auto const foo_blob = ArtifactBlob::FromMemory(
+            hash_function, ObjectType::File, content_foo);
+        REQUIRE(foo_blob.has_value());
+        auto const bar_blob = ArtifactBlob::FromMemory(
+            hash_function, ObjectType::File, content_bar);
+        REQUIRE(bar_blob.has_value());
+        REQUIRE(network.UploadBlobs({*foo_blob, *bar_blob}));
+    }
 
-    // Upload blobs
-    REQUIRE(network.UploadBlobs(BazelBlobContainer{{foo, bar}}));
-
-    // Set size to unknown
-    foo.digest.set_size_bytes(0);
-    bar.digest.set_size_bytes(0);
+    // Create digests of unknown size:
+    auto const foo_digest = ArtifactDigest{
+        HashInfo::HashData(hash_function, content_foo, /*is_tree=*/false),
+        /*size_unknown=*/0};
+    auto const bar_digest = ArtifactDigest{
+        HashInfo::HashData(hash_function, content_bar, /*is_tree=*/false),
+        /*size_unknown=*/0};
 
     // Read blobs
     auto reader = network.CreateReader();
-    std::vector<bazel_re::Digest> to_read{foo.digest, bar.digest};
+    std::vector<ArtifactDigest> to_read{foo_digest, bar_digest};
     std::vector<ArtifactBlob> blobs{};
-    for (auto next : reader.ReadIncrementally(to_read)) {
+    for (auto next : reader.ReadIncrementally(&to_read)) {
         blobs.insert(blobs.end(), next.begin(), next.end());
     }
 
     // Check order maintained
     REQUIRE(blobs.size() == 2);
-    CHECK(*blobs[0].data == content_foo);
-    CHECK(*blobs[1].data == content_bar);
+    for (auto const& blob : blobs) {
+        REQUIRE(blob.ReadContent() != nullptr);
+    }
+    CHECK(*blobs[0].ReadContent() == content_foo);
+    CHECK(*blobs[1].ReadContent() == content_bar);
 }
