@@ -15,24 +15,38 @@
 #ifndef INCLUDED_SRC_TEST_BUILDTOOL_EXECUTION_API_COMMON_API_TEST_HPP
 #define INCLUDED_SRC_TEST_BUILDTOOL_EXECUTION_API_COMMON_API_TEST_HPP
 
+#include <algorithm>
 #include <cstdlib>
 #include <filesystem>
 #include <functional>
 #include <map>
+#include <memory>
+#include <optional>
 #include <string>
+#include <unordered_map>
+#include <unordered_set>
+#include <utility>
 #include <vector>
 
 #include "catch2/catch_test_macros.hpp"
+#include "fmt/core.h"
+#include "gsl/gsl"
+#include "src/buildtool/common/artifact.hpp"
+#include "src/buildtool/common/artifact_blob.hpp"
 #include "src/buildtool/common/artifact_description.hpp"
+#include "src/buildtool/common/artifact_digest.hpp"
 #include "src/buildtool/common/artifact_digest_factory.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/execution_api/common/execution_action.hpp"
 #include "src/buildtool/execution_api/common/execution_api.hpp"
 #include "src/buildtool/execution_api/common/execution_response.hpp"
 #include "src/buildtool/execution_api/local/config.hpp"
+#include "src/buildtool/execution_engine/dag/dag.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
+#include "src/utils/cpp/expected.hpp"
 #include "test/utils/hermeticity/test_hash_function_type.hpp"
 
 using ApiFactory = std::function<IExecutionApi::Ptr()>;
@@ -227,12 +241,13 @@ using ExecProps = std::map<std::string, std::string>;
 
     HashFunction const hash_function{TestHashType::ReadFromEnvironment()};
 
-    auto test_digest = ArtifactDigestFactory::HashDataAs<ObjectType::File>(
-        hash_function, test_content);
+    auto const test_blob =
+        ArtifactBlob::FromMemory(hash_function, ObjectType::File, test_content);
+    REQUIRE(test_blob.has_value());
 
-    auto input_artifact_opt =
-        ArtifactDescription::CreateKnown(test_digest, ObjectType::File)
-            .ToArtifact();
+    auto input_artifact_opt = ArtifactDescription::CreateKnown(
+                                  test_blob->GetDigest(), ObjectType::File)
+                                  .ToArtifact();
     auto input_artifact =
         DependencyGraph::ArtifactNode{std::move(input_artifact_opt)};
 
@@ -240,9 +255,7 @@ using ExecProps = std::map<std::string, std::string>;
     std::string output_path{"output_file"};
 
     auto api = api_factory();
-    CHECK(api->Upload(ArtifactBlobContainer{{ArtifactBlob{
-                          test_digest, test_content, /*is_exec=*/false}}},
-                      false));
+    CHECK(api->Upload({*test_blob}, false));
 
     auto action =
         api->CreateAction(*api->UploadTree({{input_path, &input_artifact}}),
@@ -264,7 +277,8 @@ using ExecProps = std::map<std::string, std::string>;
         auto const artifacts = response->Artifacts();
         REQUIRE(artifacts.has_value());
         REQUIRE(artifacts.value()->contains(output_path));
-        CHECK(artifacts.value()->at(output_path).digest == test_digest);
+        CHECK(artifacts.value()->at(output_path).digest ==
+              test_blob->GetDigest());
 
         if (is_hermetic) {
             CHECK(not response->IsCached());
@@ -278,7 +292,8 @@ using ExecProps = std::map<std::string, std::string>;
                 auto const artifacts = response->Artifacts();
                 REQUIRE(artifacts.has_value());
                 REQUIRE(artifacts.value()->contains(output_path));
-                CHECK(artifacts.value()->at(output_path).digest == test_digest);
+                CHECK(artifacts.value()->at(output_path).digest ==
+                      test_blob->GetDigest());
                 CHECK(response->IsCached());
             }
         }
@@ -295,7 +310,8 @@ using ExecProps = std::map<std::string, std::string>;
         auto const artifacts = response->Artifacts();
         REQUIRE(artifacts.has_value());
         REQUIRE(artifacts.value()->contains(output_path));
-        CHECK(artifacts.value()->at(output_path).digest == test_digest);
+        CHECK(artifacts.value()->at(output_path).digest ==
+              test_blob->GetDigest());
         CHECK(not response->IsCached());
 
         SECTION("Rerun execution to verify caching") {
@@ -307,7 +323,8 @@ using ExecProps = std::map<std::string, std::string>;
             auto const artifacts = response->Artifacts();
             REQUIRE(artifacts.has_value());
             REQUIRE(artifacts.value()->contains(output_path));
-            CHECK(artifacts.value()->at(output_path).digest == test_digest);
+            CHECK(artifacts.value()->at(output_path).digest ==
+                  test_blob->GetDigest());
             CHECK(not response->IsCached());
         }
     }
