@@ -15,29 +15,47 @@
 #include "src/buildtool/execution_engine/executor/executor.hpp"
 
 #include <algorithm>
+#include <chrono>
+#include <exception>
 #include <filesystem>
+#include <functional>
 #include <map>
 #include <memory>
 #include <optional>
 #include <string>
 #include <unordered_map>
+#include <unordered_set>
 #include <utility>
+#include <variant>
 #include <vector>
 
 #include "catch2/catch_test_macros.hpp"
 #include "gsl/gsl"
 #include "src/buildtool/auth/authentication.hpp"
+#include "src/buildtool/common/action.hpp"
+#include "src/buildtool/common/action_description.hpp"
+#include "src/buildtool/common/artifact.hpp"
+#include "src/buildtool/common/artifact_blob.hpp"
 #include "src/buildtool/common/artifact_description.hpp"
 #include "src/buildtool/common/artifact_digest.hpp"
 #include "src/buildtool/common/artifact_digest_factory.hpp"
+#include "src/buildtool/common/identifier.hpp"
+#include "src/buildtool/common/remote/remote_common.hpp"
+#include "src/buildtool/common/remote/retry_config.hpp"
 #include "src/buildtool/common/repository_config.hpp"
 #include "src/buildtool/common/statistics.hpp"
 #include "src/buildtool/crypto/hash_function.hpp"
+#include "src/buildtool/execution_api/common/execution_action.hpp"
 #include "src/buildtool/execution_api/common/execution_api.hpp"
+#include "src/buildtool/execution_api/common/execution_response.hpp"
 #include "src/buildtool/execution_api/remote/config.hpp"
 #include "src/buildtool/execution_api/remote/context.hpp"
+#include "src/buildtool/execution_engine/dag/dag.hpp"
 #include "src/buildtool/execution_engine/executor/context.hpp"
+#include "src/buildtool/file_system/file_root.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
+#include "src/buildtool/logging/logger.hpp"
 #include "src/buildtool/progress_reporting/progress.hpp"
 #include "src/utils/cpp/expected.hpp"
 #include "test/utils/executor/test_api_bundle.hpp"
@@ -71,11 +89,6 @@ static auto NamedDigest(std::string const& str) -> ArtifactDigest {
                                                                str);
 }
 
-// forward declarations
-class TestApi;
-class TestAction;
-class TestResponse;
-
 /// \brief Mockup Response, stores only config and action result
 class TestResponse : public IExecutionResponse {
   public:
@@ -165,8 +178,9 @@ class TestAction : public IExecutionAction {
 /// \brief Mockup Api, use config to create action and handle artifact upload
 class TestApi : public IExecutionApi {
   public:
-    explicit TestApi(TestApiConfig config) noexcept
-        : config_{std::move(config)} {}
+    explicit TestApi(TestApiConfig config,
+                     HashFunction::Type hash_type) noexcept
+        : config_{std::move(config)}, hash_type_{hash_type} {}
 
     [[nodiscard]] auto CreateAction(
         ArtifactDigest const& /*unused*/,
@@ -188,7 +202,8 @@ class TestApi : public IExecutionApi {
     [[nodiscard]] auto RetrieveToFds(
         std::vector<Artifact::ObjectInfo> const& /*unused*/,
         std::vector<int> const& /*unused*/,
-        bool /*unused*/) const noexcept -> bool final {
+        bool /*unused*/,
+        IExecutionApi const* /*unused*/) const noexcept -> bool final {
         return false;  // not needed by Executor
     }
     [[nodiscard]] auto RetrieveToCas(
@@ -206,18 +221,21 @@ class TestApi : public IExecutionApi {
         -> std::optional<std::string> override {
         return std::nullopt;  // not needed by Executor
     }
-    [[nodiscard]] auto Upload(ArtifactBlobContainer&& blobs,
+    [[nodiscard]] auto Upload(std::unordered_set<ArtifactBlob>&& blobs,
                               bool /*unused*/) const noexcept -> bool final {
-        auto blob_range = blobs.Blobs();
         return std::all_of(
-            blob_range.begin(), blob_range.end(), [this](auto const& blob) {
+            blobs.begin(), blobs.end(), [this](auto const& blob) {
                 // for local artifacts
-                auto it1 = config_.artifacts.find(*blob.data);
+                auto const content = blob.ReadContent();
+                if (content == nullptr) {
+                    return false;
+                }
+                auto it1 = config_.artifacts.find(*content);
                 if (it1 != config_.artifacts.end() and it1->second.uploads) {
                     return true;
                 }
                 // for known and action artifacts
-                auto it2 = config_.artifacts.find(blob.digest.hash());
+                auto it2 = config_.artifacts.find(blob.GetDigest().hash());
                 return it2 != config_.artifacts.end() and it2->second.uploads;
             });
     }
@@ -234,13 +252,14 @@ class TestApi : public IExecutionApi {
             return false;
         }
     }
-    [[nodiscard]] auto IsAvailable(std::vector<ArtifactDigest> const& digests)
-        const noexcept -> std::vector<ArtifactDigest> final {
-        std::vector<ArtifactDigest> result;
+    [[nodiscard]] auto GetMissingDigests(
+        std::unordered_set<ArtifactDigest> const& digests) const noexcept
+        -> std::unordered_set<ArtifactDigest> final {
+        std::unordered_set<ArtifactDigest> result;
         try {
             for (auto const& digest : digests) {
                 if (not config_.artifacts.at(digest.hash()).available) {
-                    result.push_back(digest);
+                    result.emplace(digest);
                 }
             }
         } catch (std::exception const& /* unused */) {
@@ -249,8 +268,14 @@ class TestApi : public IExecutionApi {
         return result;
     }
 
+    [[nodiscard]] auto GetHashType() const noexcept
+        -> HashFunction::Type final {
+        return hash_type_;
+    }
+
   private:
     TestApiConfig config_{};
+    HashFunction::Type hash_type_;
 };
 
 [[nodiscard]] auto SetupConfig(std::filesystem::path const& ws)
@@ -320,10 +345,10 @@ TEST_CASE("Executor: Process artifact", "[executor]") {
                                        .exec_config = &remote_config};
 
     SECTION("Processing succeeds for valid config") {
-        auto api = std::make_shared<TestApi>(config);
+        auto api = std::make_shared<TestApi>(config, hash_function.GetType());
         Statistics stats{};
         Progress progress{};
-        auto const apis = CreateTestApiBundle(&hash_function, api);
+        auto const apis = CreateTestApiBundle(api);
         ExecutionContext const exec_context{.repo_config = &repo_config,
                                             .apis = &apis,
                                             .remote_context = &remote_context,
@@ -338,10 +363,10 @@ TEST_CASE("Executor: Process artifact", "[executor]") {
     SECTION("Processing fails if uploading local artifact failed") {
         config.artifacts[NamedDigest("local.cpp").hash()].uploads = false;
 
-        auto api = std::make_shared<TestApi>(config);
+        auto api = std::make_shared<TestApi>(config, hash_function.GetType());
         Statistics stats{};
         Progress progress{};
-        auto const apis = CreateTestApiBundle(&hash_function, api);
+        auto const apis = CreateTestApiBundle(api);
         ExecutionContext const exec_context{.repo_config = &repo_config,
                                             .apis = &apis,
                                             .remote_context = &remote_context,
@@ -356,10 +381,10 @@ TEST_CASE("Executor: Process artifact", "[executor]") {
     SECTION("Processing fails if known artifact is not available") {
         config.artifacts[NamedDigest("known.cpp").hash()].available = false;
 
-        auto api = std::make_shared<TestApi>(config);
+        auto api = std::make_shared<TestApi>(config, hash_function.GetType());
         Statistics stats{};
         Progress progress{};
-        auto const apis = CreateTestApiBundle(&hash_function, api);
+        auto const apis = CreateTestApiBundle(api);
         ExecutionContext const exec_context{.repo_config = &repo_config,
                                             .apis = &apis,
                                             .remote_context = &remote_context,
@@ -403,10 +428,10 @@ TEST_CASE("Executor: Process action", "[executor]") {
                                        .exec_config = &remote_config};
 
     SECTION("Processing succeeds for valid config") {
-        auto api = std::make_shared<TestApi>(config);
+        auto api = std::make_shared<TestApi>(config, hash_function.GetType());
         Statistics stats{};
         Progress progress{};
-        auto const apis = CreateTestApiBundle(&hash_function, api);
+        auto const apis = CreateTestApiBundle(api);
         ExecutionContext const exec_context{.repo_config = &repo_config,
                                             .apis = &apis,
                                             .remote_context = &remote_context,
@@ -424,10 +449,10 @@ TEST_CASE("Executor: Process action", "[executor]") {
     SECTION("Processing succeeds even if result was is not cached") {
         config.response.cached = false;
 
-        auto api = std::make_shared<TestApi>(config);
+        auto api = std::make_shared<TestApi>(config, hash_function.GetType());
         Statistics stats{};
         Progress progress{};
-        auto const apis = CreateTestApiBundle(&hash_function, api);
+        auto const apis = CreateTestApiBundle(api);
         ExecutionContext const exec_context{.repo_config = &repo_config,
                                             .apis = &apis,
                                             .remote_context = &remote_context,
@@ -445,10 +470,10 @@ TEST_CASE("Executor: Process action", "[executor]") {
     SECTION("Processing succeeds even if output is not available in CAS") {
         config.artifacts[NamedDigest("output2.exe").hash()].available = false;
 
-        auto api = std::make_shared<TestApi>(config);
+        auto api = std::make_shared<TestApi>(config, hash_function.GetType());
         Statistics stats{};
         Progress progress{};
-        auto const apis = CreateTestApiBundle(&hash_function, api);
+        auto const apis = CreateTestApiBundle(api);
         ExecutionContext const exec_context{.repo_config = &repo_config,
                                             .apis = &apis,
                                             .remote_context = &remote_context,
@@ -469,10 +494,10 @@ TEST_CASE("Executor: Process action", "[executor]") {
     SECTION("Processing fails if execution failed") {
         config.execution.failed = true;
 
-        auto api = std::make_shared<TestApi>(config);
+        auto api = std::make_shared<TestApi>(config, hash_function.GetType());
         Statistics stats{};
         Progress progress{};
-        auto const apis = CreateTestApiBundle(&hash_function, api);
+        auto const apis = CreateTestApiBundle(api);
         ExecutionContext const exec_context{.repo_config = &repo_config,
                                             .apis = &apis,
                                             .remote_context = &remote_context,
@@ -490,10 +515,10 @@ TEST_CASE("Executor: Process action", "[executor]") {
     SECTION("Processing fails if exit code is non-zero") {
         config.response.exit_code = 1;
 
-        auto api = std::make_shared<TestApi>(config);
+        auto api = std::make_shared<TestApi>(config, hash_function.GetType());
         Statistics stats{};
         Progress progress{};
-        auto const apis = CreateTestApiBundle(&hash_function, api);
+        auto const apis = CreateTestApiBundle(api);
         ExecutionContext const exec_context{.repo_config = &repo_config,
                                             .apis = &apis,
                                             .remote_context = &remote_context,
@@ -514,10 +539,10 @@ TEST_CASE("Executor: Process action", "[executor]") {
     SECTION("Processing fails if any output is missing") {
         config.execution.outputs = {"output1.exe" /*, "output2.exe"*/};
 
-        auto api = std::make_shared<TestApi>(config);
+        auto api = std::make_shared<TestApi>(config, hash_function.GetType());
         Statistics stats{};
         Progress progress{};
-        auto const apis = CreateTestApiBundle(&hash_function, api);
+        auto const apis = CreateTestApiBundle(api);
         ExecutionContext const exec_context{.repo_config = &repo_config,
                                             .apis = &apis,
                                             .remote_context = &remote_context,
