@@ -15,11 +15,11 @@
 #ifndef INCLUDED_SRC_TEST_BUILDTOOL_GRAPH_GRAVERSER_GRAPH_TRAVERSER_TEST_HPP
 #define INCLUDED_SRC_TEST_BUILDTOOL_GRAPH_GRAVERSER_GRAPH_TRAVERSER_TEST_HPP
 
+#include <algorithm>
 #include <chrono>
 #include <cstdlib>
 #include <filesystem>
 #include <optional>
-#include <sstream>
 #include <string>
 #include <thread>
 #include <utility>
@@ -29,8 +29,10 @@
 #include "gsl/gsl"
 #include "nlohmann/json.hpp"
 #include "src/buildtool/auth/authentication.hpp"
+#include "src/buildtool/common/cli.hpp"
 #include "src/buildtool/common/protocol_traits.hpp"
 #include "src/buildtool/common/remote/retry_config.hpp"
+#include "src/buildtool/common/repository_config.hpp"
 #include "src/buildtool/common/statistics.hpp"
 #include "src/buildtool/execution_api/common/api_bundle.hpp"
 #include "src/buildtool/execution_api/local/config.hpp"
@@ -38,15 +40,17 @@
 #include "src/buildtool/execution_api/remote/config.hpp"
 #include "src/buildtool/execution_api/remote/context.hpp"
 #include "src/buildtool/execution_engine/executor/context.hpp"
+#include "src/buildtool/file_system/file_root.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
 #include "src/buildtool/file_system/jsonfs.hpp"
+#include "src/buildtool/file_system/object_type.hpp"
 #include "src/buildtool/graph_traverser/graph_traverser.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
 #include "src/buildtool/progress_reporting/progress.hpp"
 #include "src/buildtool/storage/config.hpp"
 #include "src/buildtool/storage/storage.hpp"
-#include "src/utils/cpp/json.hpp"
+#include "src/utils/cpp/expected.hpp"
 #include "test/utils/hermeticity/test_hash_function_type.hpp"
 
 // NOLINTNEXTLINE(google-build-namespaces)
@@ -200,7 +204,7 @@ class TestProject {
                                         .progress = &progress};
 
     GraphTraverser const gt{
-        clargs.gtargs, &exec_context, [](auto done, auto cv) {}};
+        clargs.gtargs, &exec_context, [](auto /*done*/, auto /*cv*/) {}};
     auto const result =
         gt.BuildAndStage(clargs.graph_description, clargs.artifacts);
 
@@ -219,8 +223,9 @@ class TestProject {
 
     SECTION("Executable is retrieved as executable") {
         auto const clargs_exec = p.CmdLineArgs("_entry_points_get_executable");
-        GraphTraverser const gt_get_exec{
-            clargs_exec.gtargs, &exec_context, [](auto done, auto cv) {}};
+        GraphTraverser const gt_get_exec{clargs_exec.gtargs,
+                                         &exec_context,
+                                         [](auto /*done*/, auto /*cv*/) {}};
         auto const exec_result = gt_get_exec.BuildAndStage(
             clargs_exec.graph_description, clargs_exec.artifacts);
 
@@ -275,7 +280,7 @@ class TestProject {
                                         .progress = &progress};
 
     GraphTraverser const gt{
-        clargs.gtargs, &exec_context, [](auto done, auto cv) {}};
+        clargs.gtargs, &exec_context, [](auto /*done*/, auto /*cv*/) {}};
     auto const result =
         gt.BuildAndStage(clargs.graph_description, clargs.artifacts);
 
@@ -325,7 +330,7 @@ class TestProject {
                                         .progress = &progress};
 
     GraphTraverser const gt{
-        clargs.gtargs, &exec_context, [](auto done, auto cv) {}};
+        clargs.gtargs, &exec_context, [](auto /*done*/, auto /*cv*/) {}};
     auto const result =
         gt.BuildAndStage(clargs.graph_description, clargs.artifacts);
 
@@ -334,8 +339,9 @@ class TestProject {
     CHECK(FileSystemManager::IsFile(result->output_paths.at(0)));
 
     auto const clargs_full_build = p.CmdLineArgs("_entry_points_full_build");
-    GraphTraverser const gt_full_build{
-        clargs_full_build.gtargs, &exec_context, [](auto done, auto cv) {}};
+    GraphTraverser const gt_full_build{clargs_full_build.gtargs,
+                                       &exec_context,
+                                       [](auto /*done*/, auto /*cv*/) {}};
     auto const full_build_result = gt_full_build.BuildAndStage(
         clargs_full_build.graph_description, clargs_full_build.artifacts);
 
@@ -389,8 +395,9 @@ class TestProject {
         .statistics = &stats,
         .progress = &progress};
 
-    GraphTraverser const gt_upload{
-        clargs_update_cpp.gtargs, &full_context, [](auto done, auto cv) {}};
+    GraphTraverser const gt_upload{clargs_update_cpp.gtargs,
+                                   &full_context,
+                                   [](auto /*done*/, auto /*cv*/) {}};
     auto const cpp_result = gt_upload.BuildAndStage(
         clargs_update_cpp.graph_description, clargs_update_cpp.artifacts);
 
@@ -417,7 +424,7 @@ class TestProject {
         .statistics = &stats,
         .progress = &progress};
     GraphTraverser const gt{
-        clargs.gtargs, &context_known, [](auto done, auto cv) {}};
+        clargs.gtargs, &context_known, [](auto /*done*/, auto /*cv*/) {}};
     auto const result =
         gt.BuildAndStage(clargs.graph_description, clargs.artifacts);
 
@@ -469,7 +476,8 @@ static void TestBlobsUploadedAndUsed(
                                         .statistics = &stats,
                                         .progress = &progress};
 
-    GraphTraverser gt{clargs.gtargs, &exec_context, [](auto done, auto cv) {}};
+    GraphTraverser gt{
+        clargs.gtargs, &exec_context, [](auto /*done*/, auto /*cv*/) {}};
     auto const result =
         gt.BuildAndStage(clargs.graph_description, clargs.artifacts);
 
@@ -526,7 +534,8 @@ static void TestEnvironmentVariablesSetAndUsed(
                                         .statistics = &stats,
                                         .progress = &progress};
 
-    GraphTraverser gt{clargs.gtargs, &exec_context, [](auto done, auto cv) {}};
+    GraphTraverser gt{
+        clargs.gtargs, &exec_context, [](auto /*done*/, auto /*cv*/) {}};
     auto const result =
         gt.BuildAndStage(clargs.graph_description, clargs.artifacts);
 
@@ -583,7 +592,8 @@ static void TestTreesUsed(
                                         .statistics = &stats,
                                         .progress = &progress};
 
-    GraphTraverser gt{clargs.gtargs, &exec_context, [](auto done, auto cv) {}};
+    GraphTraverser gt{
+        clargs.gtargs, &exec_context, [](auto /*done*/, auto /*cv*/) {}};
     auto const result =
         gt.BuildAndStage(clargs.graph_description, clargs.artifacts);
 
@@ -640,7 +650,8 @@ static void TestNestedTreesUsed(
                                         .statistics = &stats,
                                         .progress = &progress};
 
-    GraphTraverser gt{clargs.gtargs, &exec_context, [](auto done, auto cv) {}};
+    GraphTraverser gt{
+        clargs.gtargs, &exec_context, [](auto /*done*/, auto /*cv*/) {}};
     auto const result =
         gt.BuildAndStage(clargs.graph_description, clargs.artifacts);
 
@@ -699,7 +710,7 @@ static void TestFlakyHelloWorldDetected(
     {
         auto clargs = p.CmdLineArgs("_entry_points_ctimes");
         GraphTraverser const gt{
-            clargs.gtargs, &exec_context, [](auto done, auto cv) {}};
+            clargs.gtargs, &exec_context, [](auto /*done*/, auto /*cv*/) {}};
         auto const result =
             gt.BuildAndStage(clargs.graph_description, clargs.artifacts);
 
@@ -714,7 +725,7 @@ static void TestFlakyHelloWorldDetected(
     auto clargs_output = p.CmdLineArgs();
     clargs_output.gtargs.rebuild = RebuildArguments{};
     GraphTraverser const gt_output{
-        clargs_output.gtargs, &exec_context, [](auto done, auto cv) {}};
+        clargs_output.gtargs, &exec_context, [](auto /*done*/, auto /*cv*/) {}};
     REQUIRE(gt_output.BuildAndStage(clargs_output.graph_description,
                                     clargs_output.artifacts));
     CHECK(stats.ActionsFlakyCounter() == 1);
@@ -725,8 +736,9 @@ static void TestFlakyHelloWorldDetected(
     // make_exe[flaky]->make_output[miss]->strip_time [miss]
     auto clargs_stripped = p.CmdLineArgs("_entry_points_stripped");
     clargs_stripped.gtargs.rebuild = RebuildArguments{};
-    GraphTraverser const gt_stripped{
-        clargs_stripped.gtargs, &exec_context, [](auto done, auto cv) {}};
+    GraphTraverser const gt_stripped{clargs_stripped.gtargs,
+                                     &exec_context,
+                                     [](auto /*done*/, auto /*cv*/) {}};
     REQUIRE(gt_stripped.BuildAndStage(clargs_stripped.graph_description,
                                       clargs_stripped.artifacts));
     CHECK(stats.ActionsFlakyCounter() == 1);
@@ -738,7 +750,7 @@ static void TestFlakyHelloWorldDetected(
     auto clargs_ctimes = p.CmdLineArgs("_entry_points_ctimes");
     clargs_ctimes.gtargs.rebuild = RebuildArguments{};
     GraphTraverser const gt_ctimes{
-        clargs_ctimes.gtargs, &exec_context, [](auto done, auto cv) {}};
+        clargs_ctimes.gtargs, &exec_context, [](auto /*done*/, auto /*cv*/) {}};
     REQUIRE(gt_ctimes.BuildAndStage(clargs_ctimes.graph_description,
                                     clargs_ctimes.artifacts));
     CHECK(stats.ActionsFlakyCounter() == 2);
