@@ -15,21 +15,15 @@
 #ifndef INCLUDED_SRC_TEST_UTILS_EXECUTOR_TEST_API_BUNDLE_HPP
 #define INCLUDED_SRC_TEST_UTILS_EXECUTOR_TEST_API_BUNDLE_HPP
 
-#include <utility>  // std::move
-
 #include "gsl/gsl"
-#include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/execution_api/common/api_bundle.hpp"
 #include "src/buildtool/execution_api/common/execution_api.hpp"
 
 /// \brief Creates an ApiBundle that contains a given IExecutionApi
-/// implementation. As only the hash_function field is actually needed, the
-/// remote_context and repo_config are not needed to be provided.
+/// implementation.
 [[nodiscard]] static auto CreateTestApiBundle(
-    gsl::not_null<HashFunction const*> const& hash_function,
     gsl::not_null<IExecutionApi::Ptr> const& api) noexcept -> ApiBundle {
-    return ApiBundle{
-        .hash_function = *hash_function, .local = api, .remote = api};
+    return ApiBundle{.local = api, .remote = api};
 }
 
 #endif  // INCLUDED_SRC_TEST_UTILS_EXECUTOR_TEST_API_BUNDLE_HPP
